;;
;@brief This file contains most of the number theoretic algs.
;
;  Almost all functions in here strictly use the  ABI class (most of these functions deal with crypto stuff and the  ABI class is taylored for crypto).  If a function deals with any other class, that class
;will be prepended to the name.  Most of these are very basic algs with little to no optimizations.   They were added for trying different things.
;;

INCLUDE GlobalStuff64.inc
INCLUDE AdvancedAlgs64.inc
extrn malloc: PROC

	.data
		
	.code

	;;
	;@brief Pollards Rho alg
	;
	;This alg finds a non trivial factor of _val, except you can change the offset used.  Where as the standard alg consists of _var1=_var1 ^ 2 + 1 mod _val  _var2=_var2 ^ 2 + 1 mod _val  _var2=_var2 ^ 2 + 1 mod _val
	;this alg will consist of _var1=_var1 ^ 2 + _os mod _val   _var2=_var2 ^ 2 + _os mod _val   _var2=_var2 ^ 2 + _os mod _val so that you can change the offset at will.
	;@param rcx - pointer to the the value to work on ( ABI object)
	;@param rdx - pointer to the offset to use ( ABI object)
	;;
	PollardsRho PROC _val: ABI, _os: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBI
		mov		_var1, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBI
		mov		_var2, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBI
		mov		_var3, rax
		call	CreateBI
		mov		_retval, rax
		mov		_t2, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2

MainLoop:

		mov		rcx, _var1
		mov		rdx, _t2
		IPow64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _v2
		IAdd64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _v1
		IMod64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _t2
		IPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _v2
		IAdd64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _v1
		IMod64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _t2
		IPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _v2
		IAdd64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _v1
		IMod64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _var2
		ISub64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var3
		mov		rdx, [rcx]. ABI._num
		cmp		qword ptr[rdx], 1
		ja		MoveOn
		cmp		qword ptr[rdx+8], 0
		je		GoOut

MoveOn:

		mov		rdx, _v1
		call	BIGCD64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var3
		mov		rax, [rcx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		NextCheck
		cmp		qword ptr[rax+8], 1
		je		MainLoop

NextCheck:

		mov		rdx, _v1
		IE64
		cmp		rax, 1
		je		GoOut
		mov		rcx, _retval
		mov		rax, _var3
		mov		_retval, rax
		mov		_var3, rcx

GoOut:

		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _var2
		call	DestroyBI
		mov		rcx, _var3
		call	DestroyBI
		mov		rcx, _t2
		call	DestroyBI
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	PollardsRho endp
	;;
	;@brief Fermats primality test
	;
	;This alg starts by creating _var1==_pr-1, _var2 is the temp variable that will hold a random variable, _var3 is the temp that holds the results of the calculations.  This alg simply calcs
	;a random stored in _var2 ^ _var1 mod _pr  if the result does not equal one then _pr is definately composite.
	;@returns a bool
	;@param rcx - pointer to the the suspected prime ( ABI object)
	;@param rdx - Number of iterations to run (qword)
	;;
	FermatsPrimeTest PROC _pr: ABI, _t:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_retval, 0
		call	CreateBI
		mov		_var2, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		call	CreateBI
		mov		_var3, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rdx, rax
		mov		rcx, _v1
		ISub64
		mov		_var1, rax

MainLoop:

		mov		rax, 1
		cmp		_v2, 0
		je		GoOut
		dec		_v2
		cmp		rax, 0
		cmove	rax, _one
		cmp		rbx, 0
		je		CheckCx
		mul		rbx

CheckCx:

		cmp		rax, 0
		cmove	rax, _one
		cmp		rcx, 0
		je		CheckDx
		mul		rcx

CheckDx:

		cmp		rax, 0
		cmove	rax, _one
		cmp		rdx, 0
		je		CheckDi
		mul		rdx

CheckDi:

		cmp		rax, 0
		cmove	rax, _one
		cmp		rdi, 0
		je		CheckSi
		mul		rdi

CheckSi:

		cmp		rax, 0
		cmove	rax, _one
		cmp		rsi, 0
		je		CheckR8
		mul		rsi

CheckR8:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r8, 0
		je		CheckR9
		mul		r8

CheckR9:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r9, 0
		je		CheckR10
		mul		r9

CheckR10:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r10, 0
		je		CheckR11
		mul		r10

CheckR11:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r11, 0
		je		CheckR12
		mul		r11

CheckR12:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r12, 0
		je		CheckR13
		mul		r12

CheckR13:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r13, 0
		je		CheckR14
		mul		r13

CheckR14:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r14, 0
		je		CheckR15
		mul		r14

CheckR15:

		cmp		rax, 0
		cmove	rax, _one
		cmp		r15, 0
		je		CheckDone
		mul		r15

CheckDone:

		mov		rcx, _var2
		mov		rdx, [rcx]. ABI._num
		mov		qword ptr[rdx+8], rax
		mov		rdx, _var1
		mov		r8, _v1
		IModPow64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var3
		mov		rax, [rcx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		MainLoop
		cmp		qword ptr[rax+8], 1
		cmove	rax, _one
		cmovne	rax, _zero
		je		MainLoop

GoOut:

		mov		_retval, rax
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _var2
		call	DestroyBI
		mov		rcx, _var3
		call	DestroyBI
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		RET

	FermatsPrimeTest endp
	;;
	;@brief Computes the Greatest Common Divisor of 2  ABI objects
	;
	;Standard implementation, starts out by making copies of _val1 and _val2, sets both signs to 0, then sets the larger of the 2 to _v1 and the other to _v2.  Then while both _v1 and _v2 are even
	;_v1=_v1/2  _v2=_v2/2  _var1=_var1*2.  The main loop starts by checking if _v1 is equal to 0, if it is jmp out.  Otherwise while _v1 is even _v1=_v1/2, next while _v2 is even _v2=_v2/2
	;Then _var3=ABS(_v1-_v2)/2  finally if _v1>=_v2 then _v1=_var3, else _v2=_var3
	;@returns a  ABI object
	;@param rcx - pointer to the Val1 ( ABI object)
	;@param rdx - pointer to the Val2 ( ABI object)
	;;
	BinaryGCD64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBIFromBI
		mov		_v1, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, _v2
		call	CreateBIFromBI
		mov		_v2, rax
		mov		[rax]. ABI._sign, 0
		mov		rdi, [rax]. ABI._num
		mov		rcx, _v1
		mov		rsi, [rcx]. ABI._num
		mov		rcx, qword ptr[rdi]
		cmp		rcx, qword ptr[rsi]
		jb		SwitchEm
		ja		MoveOn
		pushf
		std
repe	cmpsq
		jnb		PrepMoveOn
		popf

SwitchEm:

		mov		rax, _v1
		mov		rcx, _v2
		mov		_v2, rax
		mov		_v1, rcx

PrepMoveOn:

		popf

MoveOn:
		
		call	CreateBI
		mov		_var1, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		call	CreateBI
		mov		_var2, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBI
		mov		_var3, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1

PrepLoop:

		mov		rcx, _v1
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		MainLoop
		mov		rcx, _v2
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		MainLoop
		mov		rcx, _v1
		mov		rdx, _var2
		IDivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	DestroyBI
		mov		rcx, _v2
		mov		rdx, _var2
		IDivide64
		mov		rcx, _v2
		mov		_v2, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _var2
		IMultiply64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		jmp		PrepLoop
		
MainLoop:

		mov		rcx, _v1
		mov		rax, [rcx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		XEvenLoop
		cmp		qword ptr[rax+8], 0
		je		GoOut

XEvenLoop:

		mov		rcx, _v1
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 1
		je		YEvenLoop
		mov		rcx, _v1
		mov		rdx, _var2
		IDivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	DestroyBI
		jmp		XEvenLoop

YEvenLoop:

		mov		rcx, _v2
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 1
		je		NextStep
		mov		rcx, _v2
		mov		rdx, _var2
		IDivide64
		mov		rcx, _v2
		mov		_v2, rax
		call	DestroyBI
		jmp		XEvenLoop

NextStep:

		mov		rcx, _v1
		mov		rdx, _v2
		ISub64
		mov		rcx, _var3
		mov		_var3, rax
		mov		[rax]. ABI._sign, 0
		call	DestroyBI
		mov		rcx, _var3
		mov		rdx, _var2
		IDivide64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _v1
		mov		rdi, [rcx]. ABI._num
		mov		rcx, _v2
		mov		rsi, [rcx]. ABI._num
		mov		rcx, qword ptr[rdi]
		cmp		rcx, qword ptr[rsi]
		jb		SetY
		pushf
		std
repe	cmpsq
		jb		PrepSetY
		popf
		mov		rax, _v1
		mov		rcx, _var3
		mov		_var3, rax
		mov		_v1, rcx
		jmp		MainLoop

PrepSetY:

		popf

SetY:

		mov		rax, _v2
		mov		rcx, _var3
		mov		_var3, rax
		mov		_v2, rcx
		jmp		MainLoop

GoOut:

		mov		rcx, _var1
		mov		rdx, _v2
		IMultiply64
		mov		_retval, rax
		mov		rcx, _v1
		call	DestroyBI
		mov		rcx, _v2
		call	DestroyBI
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _var2
		call	DestroyBI
		mov		rcx, _var3
		call	DestroyBI
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BinaryGCD64 endp
	;;
	;@brief Computes the Greatest Common Divisor of 2  ABI objects using the Binary extended GCD alg
	;
	;Standard implementation, starts out by making copies of _val1 and _val2 and sets _var7=1, The while _val1 and _val2 are both even _val1=_val1/2   _val2=_val2/2  _var7=_var7*2
	;next _var5=_val1  and _var6=_val2  _var1=1 _var2=0 _var3=0 and _var4=1
	;next while _var5 is even _var5=_var5/2 and if both _var1 and _var2 are even then _var1=_var1/2  _var2=_var2/2 else _var1=(_var1+_val2)/2  _var2=(_var2-_val1)/2
	;then while _var6 is even _var6=_var6/2 and if both _var3 and _var4 are even then _var3=_var3/2  _var4=_var4/2 else _var3=(_var3+_val2)/2  _var4=(_var4-_val1)/2
	;then if _var5>=_var6 then _var5=_var5-_var6   _var1=_var1-_var3   _var2=_var2-_var4  else  _var6=_var6-_var5   _var3=_var3-_var1   _var4=_var4-_var2 
	;lastly we check if _var5==0 if so we go out (this check it at the head of the main loop).  When going out we create an array and set the first element=_var3, the second=_var4, and the 
	;third=_var7*_var6 (this array represents creation of a ABEGCD object for the return value).  Finally we clean up all our temp variables and exit.
	;@returns a ABEGCD object
	;@param rcx - pointer to the Val1 ( ABI object)
	;@param rdx - pointer to the Val2 ( ABI object)
	;;
	BinaryExtendedGCD64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBIFromBI
		mov		_v1, rax
		mov		rcx, _v2
		call	CreateBIFromBI
		mov		_v2, rax
		call	CreateBI
		mov		_t2, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBI
		mov		_var1, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		call	CreateBI
		mov		_var2, rax
		call	CreateBI
		mov		_var3, rax
		call	CreateBI
		mov		_var4, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		call	CreateBI
		mov		_var7, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1

PrepLoop:

		mov		rcx, _v2
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		PrepMainLoop
		mov		rcx, _v1
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		PrepMainLoop
		mov		rdx, _t2
		IDivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	DestroyBI
		mov		rcx, _v2
		mov		rdx, _t2
		IDivide64
		mov		rcx, _v2
		mov		_v2, rax
		call	DestroyBI
		mov		rcx, _var7
		mov		rdx, _t2
		IMultiply64
		mov		rcx, _var7
		mov		_var7, rax
		call	DestroyBI
		jmp		PrepLoop

PrepMainLoop:

		mov		rcx, _v1
		call	CreateBIFromBI
		mov		_var5, rax
		mov		rcx, _v2
		call	CreateBIFromBI
		mov		_var6, rax

MainLoop:

		mov		rcx, _var5
		mov		rax, [rcx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		ItsGood
		cmp		qword ptr[rax+8], 0
		je		GoOut

ItsGood:

		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		VEven
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var5
		mov		_var5, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		UOtherWay
		mov		rcx, _var1
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		UOtherWay
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		jmp		MainLoop

UOtherWay:

		mov		rdx, _v2
		IAdd64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _v1
		ISub64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		jmp		MainLoop

VEven:
		mov		rcx, _var6
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		NextStep
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var6
		mov		_var6, rax
		call	DestroyBI
		mov		rcx, _var4
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		VOtherWay
		mov		rcx, _var3
		mov		rax, [rcx]. ABI._num
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		VOtherWay
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var4
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var4
		mov		_var4, rax
		call	DestroyBI
		jmp		VEven

VOtherWay:

		mov		rdx, _v2
		IAdd64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var3
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var4
		mov		rdx, _v1
		ISub64
		mov		rcx, _var4
		mov		_var4, rax
		call	DestroyBI
		mov		rcx, _var4
		mov		rdx, _t2
		IDivide64
		mov		rcx, _var4
		mov		_var4, rax
		call	DestroyBI
		jmp		VEven

NextStep:

		mov		rcx, _var5
		mov		rdx, _var6
		IGTOE64
		cmp		rax, 1
		jne		OtherWay
		mov		rcx, _var5
		mov		rdx, _var6
		ISub64
		mov		rcx, _var5
		mov		_var5, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _var3
		ISub64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBI
		mov		rcx, _var2
		mov		rdx, _var4
		ISub64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBI
		jmp		MainLoop

OtherWay:

		mov		rcx, _var6
		mov		rdx, _var5
		ISub64
		mov		rcx, _var6
		mov		_var6, rax
		call	DestroyBI
		mov		rcx, _var3
		mov		rdx, _var1
		ISub64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBI
		mov		rcx, _var4
		mov		rdx, _var2
		ISub64
		mov		rcx, _var4
		mov		_var4, rax
		call	DestroyBI
		jmp		MainLoop

GoOut:

		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		rcx, _var3
		mov		[rax].ABEGCD._a, rcx
		mov		rcx, _var4
		mov		[rax].ABEGCD._b, rcx
		mov		rcx, _var7
		mov		rdx, _var6
		IMultiply64
		mov		rcx, _retval
		mov		[rcx].ABEGCD._gtimesv, rax
		mov		rcx, _v1
		call	DestroyBI
		mov		rcx, _v2
		call	DestroyBI
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _var2
		call	DestroyBI
		mov		rcx, _var5
		call	DestroyBI
		mov		rcx, _var6
		call	DestroyBI
		mov		rcx, _var7
		call	DestroyBI
		mov		rcx, _t2
		call	DestroyBI
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BinaryExtendedGCD64 endp

	end

