;;
;@brief This file contains some of the more advanced math functions to be used by the library.
;
;  Most of the algs in here deal with the trig functions, and most of the are for only the  ABD and  ABF classes.  All of the functions have the class they are meant for prepended to the name.  There is a detailed
;explanation of each function in the documentation.  The functions in here use a combination of the basic functions found in Asm64 and the higher level functions found in BIAsm.
;;

include GlobalStuff64.inc

	.data

	.code

	;;
	;@brief Computes the inverse givin the encryption exponent and the totient
	;
	;Computes the decryption exponent for RSA given the encryption exponent and the totient.
	;@returns a  ABI object
	;@param rcx - pointer to the exponent ( ABI object)
	;@param rdx - pointer to the totient ( ABI object)
	;;
	BIComputeInverse64 PROC _e: ABI, _t: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBIFromBI
		mov		_var9, rax								;_var9==_u3
		mov		rcx, _v2
		call	CreateBIFromBI
		mov		_var3, rax								;_var3==_v3
		call	CreateBI
		mov		_var8, rax								;_var8==_u1
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx+8], 1
		call	CreateBI
		mov		_var1, rax								;_var1==_v1
		call	CreateBI
		mov		_t1, rax								;_t1==_t1
		call	CreateBI
		mov		_t3, rax								;_t3==_t3
		call	CreateBI
		mov		_t4, rax								;_t4==_q
		mov		_t2, 0									;_t2==_iter

MainLoop:

		mov		rdx, _v3
		mov		rax, [rdx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		DoCalc
		cmp		qword ptr[rax+8], 0
		je		HeadOut
		
DoCalc:

		mov		rcx, _var9
		IDivide64
		mov		rcx, _t4
		mov		_t4, rax
		call	DestroyBI
		mov		rcx, _var9
		mov		rdx, _var3
		IMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBI
		mov		rcx, _t4
		mov		rdx, _var1
		IMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBI
		mov		rcx, _var8
		mov		rdx, _t1
		IAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBI
		mov		rcx, _var8
		call	DestroyBI
		mov		rcx, _var1
		call	CreateBIFromBI
		mov		_var8, rax
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _t1
		call	CreateBIFromBI
		mov		_var1, rax
		mov		rcx, _var9
		call	DestroyBI
		mov		rcx, _var3
		call	CreateBIFromBI
		mov		_var9, rax
		mov		rcx, _var3
		call	DestroyBI
		mov		rcx, _t3
		call	CreateBIFromBI
		mov		_var3, rax
		cmp		_t2, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		_t2, rcx
		jmp		MainLoop

HeadOut:

		mov		rcx, _var9
		mov		rax, [rcx]. ABI._num
		cmp		qword ptr[rax], 1
		ja		ItsGood
		cmp		qword ptr[rax+8], 1
		ja		ItsGood
		call	CreateBI
		mov		_retval, rax
		jmp		GoOut

ItsGood:

		cmp		_t2, 0
		je		DontSub
		mov		rcx, _v2
		mov		rdx, _var8
		ISub64
		mov		rcx, _var8
		mov		_var8, rax
		call	DestroyBI

DontSub:

		mov		rax, _var8
		mov		_retval, rax

GoOut:

		mov		rax, _retval
		
		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BIComputeInverse64 ENDP
	;;
	;@brief Computes the Greatest Common Divisor of 2  ABI objects
	;
	;
	;@returns a  ABI object
	;@param rcx - pointer to the Val1 ( ABI object)
	;@param rdx - pointer to the Val2 ( ABI object)
	;;
	BIGCD64 PROC _val1: ABI, _val2: ABI

		;standard greatest common divisor routine

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		_initvals, 1
		je		AllreadyInitialized
		call	InitVals

AllreadyInitialized:

		;set up temp vals

		mov		rcx, _v1
		mov		rcx, [rcx]. ABI._num
		call	CreateBICopy
		mov		rcx, _v1
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABI._sign, rcx
		mov		_v1, rax
		mov		rcx, _v2
		mov		rcx, [rcx]. ABI._num
		call	CreateBICopy
		mov		rcx, _v2
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABI._sign, rcx
		mov		_v2, rax

		;if(_val1>=_val2) then make _a==_val1 and _b==_val2 else reverse them

		mov		rcx, _v1
		mov		rdx, _v2
		IGTOE64
		cmp		rax, 1
		je		MainLoop
		mov		rax, _v1
		mov		rcx, _v2
		mov		_v1, rcx
		mov		_v2, rax

MainLoop:

		;while(_b>0){

		mov		rdx, _v2
		mov		rdx, [rdx]. ABI._num
		cmp		qword ptr[rdx], 1
		ja		DoAlg
		cmp		qword ptr[rdx+8], 0
		je		GoOut

DoAlg:

		;_r=_a%_b;
		;_a=_b;
		;_b=_r;

		mov		rdx, _v2
		mov		rcx, _v1
		IMod64
		mov		rcx, _v1
		mov		rdx, _v2
		mov		_v1, rdx
		mov		_v2, rax
		call	DestroyBI
		jmp		MainLoop

GoOut:

		mov		rcx, _v2
		call	DestroyBI

		;return _a;

		mov		rax, _v1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BIGCD64  ENDP
	;;
	;@brief Computes the Greatest Common Divisor of 2  ABD objects
	;@returns a  ABD object
	;@param rcx - pointer to the Val1 ( ABD object)
	;@param rdx - pointer to the Val2 ( ABD object)
	;;
	BDGCD64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		_v1, rax
		mov		rcx, _v2
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		rcx, _v2
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		_v2, rax
		mov		rcx, _v1
		mov		rdx, _v2
		DGTOE64
		cmp		rax, 1
		je		MainLoop
		mov		rax, _v1
		mov		rcx, _v2
		mov		_v1, rcx
		mov		_v2, rax

MainLoop:

		mov		rdx, _v2
		cmp		[rdx]. ABD._decimalplace, 0
		ja		DoAlg
		mov		rdx, [rdx]. ABD._num
		cmp		qword ptr[rdx], 1
		ja		DoAlg
		cmp		qword ptr[rdx+8], 0
		je		GoOut

DoAlg:

		mov		rdx, _v2
		mov		rcx, _v1
		DMod64
		mov		rcx, _v1
		mov		rdx, _v2
		mov		_v1, rdx
		mov		_v2, rax
		call	DestroyBD
		jmp		MainLoop

GoOut:

		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _v1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDGCD64  ENDP
	;;
	;@brief Computes the Greatest Common Divisor of 2  ABF objects
	;@returns a  ABF object
	;@param rcx - pointer to the Val1 ( ABF object)
	;@param rdx - pointer to the Val2 ( ABF object)
	;;
	BFGCD64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		rcx, _v1
		mov		rcx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		_v1, rax
		mov		rcx, _v2
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		rcx, _v2
		mov		rcx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		_v2, rax
		mov		rcx, _v1
		mov		rdx, _v2
		FGTOE64
		cmp		rax, 1
		je		MainLoop
		mov		rax, _v1
		mov		rcx, _v2
		mov		_v1, rcx
		mov		_v2, rax

MainLoop:

		mov		rdx, _v2
		mov		rcx, [rdx]. ABF._denominator
		cmp		qword ptr[rcx], 1
		ja		DoAlg
		cmp		qword ptr[rcx+8], 1
		ja		DoAlg
		mov		rdx, [rdx]. ABF._numerator
		cmp		qword ptr[rdx], 1
		ja		DoAlg
		cmp		qword ptr[rdx+8], 0
		je		GoOut

DoAlg:

		mov		rdx, _v2
		mov		rcx, _v1
		FMod64
		mov		rcx, _v1
		mov		rdx, _v2
		mov		_v1, rdx
		mov		_v2, rax
		call	DestroyBD
		jmp		MainLoop

GoOut:

		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _v1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFGCD64  ENDP
	;;
	;@brief Computes the Least Common Multiple of 2  ABI objects
	;@returns a  ABI object
	;@param rcx - pointer to the Val1 ( ABI object)
	;@param rdx - pointer to the Val2 ( ABI object)
	;;
	BILCM64 PROC _val1: ABI, _val2: ABI

		;standard least common multiple alg, works by multiplying the values together then dividing by the gcd

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		;_retval=_val1*_val2;

		mov		_v1, rcx
		mov		_v2, rdx
		IMultiply64
		mov		_retval, rax

		;_retval=_retval/GCD(_val1, _val2);

		mov		rcx, _v1
		mov		rdx, _v2
		call	BIGCD64
		mov		_t1, rax
		mov		rdx, rax
		mov		rcx, _retval
		IDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBI
		mov		rcx, _t1
		call	DestroyBI

		;return _retval;

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BILCM64 ENDP
	;;
	;@brief Computes the Least Common Multiple of 2  ABD objects
	;@returns a  ABD object
	;@param rcx - pointer to the Val1 ( ABD object)
	;@param rdx - pointer to the Val2 ( ABD object)
	;;
	BDLCM64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		DMultiply64
		mov		_retval, rax
		mov		rcx, _v1
		mov		rdx, _v2
		call	BDGCD64
		mov		_t1, rax
		mov		rdx, rax
		mov		rcx, _retval
		DDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDLCM64 ENDP
	;;
	;@brief Computes the least common multiple of 2  ABF objects.
	;@returns a  ABF object.
	;@param rcx - The first  ABF object.
	;@param rdx - The second  ABF object.
	;;
	BFLCM64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		FMultiply64
		mov		_retval, rax
		mov		rcx, _v1
		mov		rdx, _v2
		call	BFGCD64
		mov		_t1, rax
		mov		rdx, rax
		mov		rcx, _retval
		FDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFLCM64 ENDP
	;;
	;@brief Computes the factorial of an array.
	;@returns an array.
	;@param rcx - The array to compute.
	;;
	AsmFactorial64 PROC _val:PTR qword

		ENTERFUNC _zero, _zero, _zero, _one, _one, _one, _one, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, 0
		mov		r11, _v1
		cmp		qword ptr[r11], 1
		ja		MoveOn
		cmp		qword ptr[r11+8], 0
		je		ReturnZero
		cmp		qword ptr[r11+8], 1
		je		ReturnOne
		cmp		qword ptr[r11+8], 2
		je		ReturnTwo

MoveOn:

		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r13, rax
		mov		_retval, rax
		mov		r11, _v1
		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		r11, _v1
		mov		rcx, qword ptr[r11]
		mov		rsi, r11
		mov		rdi, rax
		inc		rcx
		pushf
		cld
rep		movsq
		popf

MainLoop:

		mov		rcx, _t1
		mov		rdi, 2
		sub		qword ptr[rcx+8], 1
		jnz		SubOut
		cmp		rdi, qword ptr[rcx]
		ja		SubOut

SubLoop:

		sub		qword ptr[rcx+rdi*8], 1
		jnc		SubOut
		inc		rdi
		cmp		rdi, qword ptr[rcx]
		jna		SubLoop

SubOut:

		mov		rdi, qword ptr[rcx]

CheckZerosLoop:

		cmp		qword ptr[rcx+rdi*8], 0
		ja		ZerosOut
		cmp		rdi, 1
		jna		ZerosOut
		dec		rdi
		mov		qword ptr[rcx], rdi
		jmp		CheckZerosLoop

ZerosOut:

		cmp		qword ptr[rcx], 1
		ja		CarryOn
		cmp		qword ptr[rcx+8], 1
		jna		GetReadyToLeave

CarryOn:

		mov		rdx, _retval
		Multiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64
		jmp		MainLoop

GetReadyToLeave:

		mov		rcx, _t1
		call	Free64
		jmp		OutOfHere

ReturnZero:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_retval, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		jmp		OutOfHere

ReturnOne:

		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		_retval, rax
		jmp		OutOfHere

ReturnTwo:

		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		_retval, rax
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		call	GetMem64
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		_retval, rax
		jmp		OutOfHere

OutOfHere:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _one, _one, _one, _one, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero

		RET

	AsmFactorial64 ENDP
	;;
	;@brief Computes the factorial of a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABI object to compute.
	;;
	BIFactorial64 PROC _val: ABI

		;standard factorial alg uses recursive multiplies

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		;test for bad values and exit if detected

		mov		_v1, rcx
		mov		_t1, 0
		mov		r11, _v1
		mov		rax, [r11]. ABI._num
		cmp		qword ptr[rax], 1
		ja		MoveOn
		cmp		qword ptr[rax+8], 0
		je		ReturnZero
		cmp		qword ptr[rax+8], 1
		je		ReturnOne
		cmp		qword ptr[rax+8], 2
		je		ReturnTwo

MoveOn:

		;create the return val and the count value, initialize them the the value passed in

		mov		rcx, [r11]. ABI._num
		call	CreateBICopy
		mov		_retval, rax
		mov		r11, _v1
		mov		rcx, [r11]. ABI._num
		call	CreateBICopy
		mov		_t1, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, _retval
		mov		[rcx]. ABI._sign, 0
		mov		r11, _v1

MainLoop:

		;decrement the count by 1 and jump out if zero, otherwise multiply the result to the return value

		mov		rcx, _t1
		sub		qword ptr[rcx+8], 1
		jz		GetReadyToLeave
		mov		rdx, _retval
		IMultiply64
		mov		rdx, _retval
		mov		rcx, [rdx]. ABI._num
		mov		[rdx]. ABI._num, rax
		call	Free64
		jmp		MainLoop

GetReadyToLeave:

		;destroy the count value and jump out

		mov		rcx, _t1
		call	DestroyBI
		jmp		OutOfHere

ReturnZero:

		;create a return value initialized to zero and exit

		call	CreateBI
		mov		_retval, rax
		jmp		OutOfHere

ReturnOne:

		;create a return value initialized to one

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		call	CreateBITake
		mov		_retval, rax
		jmp		OutOfHere

ReturnTwo:

		;create a return value initialized to 2

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		call	CreateBITake
		mov		_retval, rax
		jmp		OutOfHere

		;there was an error

		call	CreateBI
		mov		_retval, rax
		jmp		OutOfHere

OutOfHere:

		;exit the alg

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIFactorial64 ENDP
	;;
	;@brief Computes the factorial of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDFactorial64 PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, 0
		mov		r11, _v1
		cmp		[r11]. ABD._decimalplace, 0
		jne		MallocError
		mov		rcx, [r11]. ABD._num
		cmp		qword ptr[rcx], 1
		ja		MoveOn
		cmp		qword ptr[rcx+8], 0
		je		ReturnZero
		cmp		qword ptr[rcx+8], 1
		je		ReturnOne
		cmp		qword ptr[rcx+8], 2
		je		ReturnTwo

MoveOn:

		call	CreateBDCopy
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t1, rax
		mov		[rax]. ABD._sign, 0

MainLoop:

		mov		rcx, _t1
		mov		rdx, [rcx]. ABD._num
		sub		qword ptr[rdx+8], 1
		jz		GetReadyToLeave
		mov		rdx, _retval
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		jmp		MainLoop

GetReadyToLeave:

		mov		rcx, _t1
		call	DestroyBD
		jmp		OutOfHere

ReturnZero:

		call	CreateBD
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		_retval, rax
		jmp		OutOfHere

ReturnOne:

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		_retval, rax
		jmp		OutOfHere

ReturnTwo:

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		_retval, rax
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		OutOfHere
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

OutOfHere:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDFactorial64 ENDP
	;;
	;@brief Computes the factorial of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;;
	BFFactorial64 PROC _val: ABF

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, 0
		mov		r11, _v1
		mov		rax, [r11]. ABF._denominator
		cmp		qword ptr[rax], 1
		jne		MallocError
		cmp		qword ptr[rax+8], 1
		jne		MallocError
		mov		rcx, [r11]. ABF._numerator
		cmp		qword ptr[rcx], 1
		ja		MoveOn
		cmp		qword ptr[rcx+8], 0
		je		ReturnZero
		cmp		qword ptr[rcx+8], 1
		je		ReturnOne
		cmp		qword ptr[rcx+8], 2
		je		ReturnTwo

MoveOn:

		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t1, rax
		mov		[rax]. ABF._sign, 0

MainLoop:

		mov		rcx, _t1
		mov		rdx, [rcx]. ABF._numerator
		sub		qword ptr[rdx+8], 1
		jz		GetReadyToLeave
		mov		rdx, _retval
		FMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		jmp		MainLoop

GetReadyToLeave:

		mov		rcx, _t1
		call	DestroyBD
		jmp		OutOfHere

ReturnZero:

		call	CreateBF
		mov		[rax]. ABD._sign, 0
		mov		_retval, rax
		jmp		OutOfHere

ReturnOne:

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		call	CreateBFTake
		mov		[rax]. ABD._sign, 0
		mov		_retval, rax
		jmp		OutOfHere

ReturnTwo:

		mov		rcx, 32
		call	GetMem64
		mov		rcx, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		call	CreateBFTake
		mov		[rax]. ABD._sign, 0
		mov		_retval, rax
		jmp		OutOfHere

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		OutOfHere
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

OutOfHere:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BFFactorial64 ENDP
	;;
	;@brief Computes the square root of a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABI object to compute.
	;;
	BISqrt64 PROC _val: ABI

		;standard implementation of the square root routine

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		;create the return value and tempoarary values

		mov		_v1, rcx
		mov		r11, _v1
		mov		rcx, [r11]. ABI._num
		call	CreateBICopy
		mov		_retval, rax
		mov		[rax]. ABI._sign, 0
		mov		r11, _v1
		mov		rcx, [r11]. ABI._num
		call	CreateBICopy
		mov		_t1, rax
		mov		[rax]. ABI._sign, 0
		call	CreateBI
		mov		_t2, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBI
		mov		_t3, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		call	CreateBI
		mov		_var1, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBI
		mov		_t5, rax
		mov		[rax]. ABI._sign, 0
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		_v2, 0

MainLoop:

		inc		_v2
		cmp		_v2, 10000
		je		GoOut
		mov		rcx, _t5
		call	DestroyBI
		mov		rcx, _t3
		mov		rcx, [rcx]. ABI._num
		call	CreateBICopy
		mov		_t5, rax
		mov		rcx, _t3
		mov		rcx, _var1
		call	DestroyBI
		mov		rax, _retval
		mov		_var1, rax
		mov		rcx, _t1
		mov		rdx, _t3
		IDivide64
		mov		_retval, rax
		mov		rdx, rax
		mov		rcx, _t3
		IAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBI
		mov		rcx, _retval
		mov		rdx, _t2
		IDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBI
		mov		rcx, _t3
		call	DestroyBI
		mov		rcx, _retval
		mov		rcx, [rcx]. ABI._num
		call	CreateBICopy
		mov		rcx, _retval
		mov		[rax]. ABI._sign, 0
		dec		_var2
		mov		rcx, _t3
		mov		rdx, _t5
		IE64
		cmp		rax, 1
		jne		MainLoop

GoOut:

		mov		rcx, _t5
		call	DestroyBI
		mov		rcx, _t2
		call	DestroyBI
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _retval
		call	DestroyBI
		mov		rcx, _t1
		call	DestroyBI
		mov		rcx, _t3
		call	CreateBITake
		mov		rcx, _t3
		mov		[rax]. ABI._sign, 0

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BISqrt64 ENDP
	;;
	;@brief Computes the square root of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDSqrt64 PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		r11, _v1
		mov		rcx, [r11]. ABD._num						; ABF^ _t1=gcnew  ABF(_val);
		call	CreateBDCopy
		mov		_t1, rax
		mov		[rax]. ABD._sign, 0
		mov		r11, _v1
		mov		rcx, [r11]. ABD._num						; ABF^ _retval=gcnew  ABF(_val);
		call	CreateBDCopy
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		r11, _v1
		mov		rcx, [r11]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rcx
		mov		rdx, _t1
		mov		[rdx]. ABD._decimalplace, rcx
		call	CreateBD					; ABF^ _t2=gcnew  ABF(1);
		mov		_t2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 5
		mov		[rax]. ABD._decimalplace, 1
		mov		[rax]. ABD._sign, 0
		call	CreateBD					; ABF^ _t3=gcnew  ABF(10);
		mov		_t3, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._sign, 0
		call	CreateBD					; ABF^ _var1=gcnew  ABF();
		mov		_var1, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._sign, 0
		call	CreateBD					; ABF^ _var1=gcnew  ABF();
		mov		_t5, rax
		mov		_v2, 0

MainLoop:

		mov		rcx, _t5
		mov		rdx, _t3
		DE64
		cmp		rax, 1
		je		GoOut
		inc		_v2
		cmp		_v2, 10000
		ja		GoOut
		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t5, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx	
		mov		rcx, _var1								;_var1=_retval;
		call	DestroyBD
		mov		rax, _retval
		mov		_var1, rax
		mov		rcx, _t1								;(_t1/_t3);
		mov		rdx, _t3
		DDivide64
		mov		_retval, rax								;_retval=_t3+(_t1/_t3);
		mov		rcx, rax
		mov		rdx, _t3
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval									;_retval=_retval*_t2;
		mov		rdx, _t2
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _retval
		mov		rcx, [rcx]. ABD._num							;_t3=_retval;
		call	CreateBDCopy
		mov		_t3, rax
		mov		rcx, _retval
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		[rax]. ABD._sign, 0
		dec		_var2
		jmp		MainLoop

GoOut:

		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _retval
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _t3

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDSqrt64 ENDP
	;;
	;@brief Computes the square root of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BFSqrt64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_var2, rdx
		mov		_v1, rcx
		mov		r11, _v1
		mov		rcx, [r11]. ABF._numerator						; ABF^ _t1=gcnew  ABF(_val);
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_t1, rax
		mov		[rax]. ABF._sign, 0
		mov		r11, _v1
		mov		rcx, [r11]. ABF._numerator						; ABF^ _retval=gcnew  ABF(_val);
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		r11, _v1
		mov		rdx, _t1
		call	CreateBF					; ABF^ _t2=gcnew  ABF(1);
		mov		_t2, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		mov		[rax]. ABF._sign, 0
		call	CreateBF					; ABF^ _t3=gcnew  ABF(10);
		mov		_t3, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		[rax]. ABF._sign, 0
		call	CreateBF					; ABF^ _var1=gcnew  ABF();
		mov		_var1, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		mov		[rax]. ABD._sign, 0

MainLoop:

		cmp		_var2, 0								;while(_var2<_iter){	;_retval=_t3+(_t1/_t3);
		jna		GoOut
		mov		rcx, _var1								;_var1=_retval;
		call	DestroyBD
		mov		rax, _retval
		mov		_var1, rax
		mov		rcx, _t1								;(_t1/_t3);
		mov		rdx, _t3
		FDivide64
		mov		_retval, rax								;_retval=_t3+(_t1/_t3);
		mov		rcx, rax
		mov		rdx, _t3
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval									;_retval=_retval*_t2;
		mov		rdx, _t2
		FMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator							;_t3=_retval;
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rcx, _retval
		mov		[rax]. ABF._sign, 0
		dec		_var2
		jmp		MainLoop

GoOut:

		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _retval
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _t3

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BFSqrt64 ENDP
	;;
	;@brief Computes the square root of a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to compute.
	;;
	BDISqrt64 PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		call	CreateBD
		mov		_var1, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		mov		rdx, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		DPow64
		mov		_t1, rax
		mov		rdx, _var1
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		DPow64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _t1
		DAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BDSqrt64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABDI._real
		DAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var1
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BDSqrt64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rax, [rcx]. ABDI._im
		cmp		[rax]. ABD._sign, 0
		cmove	rdx, _zero
		cmovne	rdx, _one
		mov		_var2, rdx
		mov		rax, [rax]. ABD._num
		cmp		qword ptr[rax], 1
		ja		MoveOn
		cmp		qword ptr[rax+8], 0
		je		MakeZero

MoveOn:

		call	CreateBDFromBD
		mov		_v2, rax
		mov		rax, [rax]. ABDI._real
		cmp		[rax]. ABD._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABD._sign, rcx
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rcx, [rcx]. ABDI._real
		DPow64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rcx, [rcx]. ABDI._im
		DPow64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, _t2
		DAdd64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		call	BDSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, rax
		mov		rdx, _v2
		mov		rdx, [rdx]. ABDI._real
		DAdd64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _var1
		DDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		call	BDSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		cmp		_var2, 0
		je		DontNeg
		mov		rcx, _var1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		mov		[rcx]. ABD._sign, 1
		mov		rcx, _t2
		mov		rdx, _var1
		DMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD

DontNeg:

		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBDITakeBothBD
		mov		_retval, rax
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBDI
		mov		rcx, _var1
		call	DestroyBD

Finish:

		mov		rcx, _retval
		mov		rcx, [rcx]. ABDI._real
		cmp		[rcx]. ABD._sign, 0
		je		GoOut
		call	CreateBDI
		mov		_v2, rax
		mov		rcx, [rax]. ABDI._real
		mov		[rcx]. ABD._sign, 1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rcx, _v2
		mov		rdx, _retval
		call	BDIMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBDI
		mov		rcx, _v2
		call	DestroyBDI

GoOut:

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero
		
		mov		rax, _retval			

		RET

MakeZero:

		call	CreateBD
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBDITakeBothBD
		mov		_retval, rax
		mov		rcx, _var1
		call	DestroyBD
		jmp		Finish

		RET

	BDISqrt64 ENDP
	;;
	;@brief Computes the square root of a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to compute.
	;;
	BFISqrt64 PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		call	CreateBF
		mov		_var1, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		mov		rdx, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real
		FPow64
		mov		_t1, rax
		mov		rdx, _var1
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		FPow64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _t1
		FAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BFSqrt64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABFI._real
		FAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var1
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BFSqrt64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rax, [rcx]. ABFI._im
		cmp		[rax]. ABF._sign, 0
		cmove	rdx, _zero
		cmovne	rdx, _one
		mov		_var2, rdx
		mov		rax, [rax]. ABF._numerator
		cmp		qword ptr[rax], 1
		ja		MoveOn
		cmp		qword ptr[rax+8], 0
		je		MakeZero

MoveOn:

		mov		rdx, _v1
		mov		rdx, [rdx]. ABFI._im
		mov		rdx, [rdx]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_v2, rax
		mov		rax, [rax]. ABFI._real
		cmp		[rax]. ABF._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABF._sign, rcx
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rcx, [rcx]. ABFI._real
		FPow64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rcx, [rcx]. ABFI._im
		FPow64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, _t2
		FAdd64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _defaultcyclecount
		call	BFSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, rax
		mov		rdx, _v2
		mov		rdx, [rdx]. ABFI._real
		FAdd64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _var1
		FDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _defaultcyclecount
		call	BFSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		cmp		_var2, 0
		je		DontNeg
		mov		rcx, _var1
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 1
		mov		[rcx]. ABF._sign, 1
		mov		rcx, _t2
		mov		rdx, _var1
		FMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD

DontNeg:

		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBFI
		mov		rcx, _var1
		call	DestroyBD

Finish:

		mov		rcx, _retval
		mov		rcx, [rcx]. ABFI._real
		cmp		[rcx]. ABF._sign, 0
		je		GoOut
		call	CreateBFI
		mov		_v2, rax
		mov		rcx, [rax]. ABFI._real
		mov		[rcx]. ABF._sign, 1
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rcx, _v2
		mov		rdx, _retval
		call	BFIMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBFI
		mov		rcx, _v2
		call	DestroyBFI

GoOut:

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero
		
		mov		rax, _retval			

		RET

MakeZero:

		call	CreateBF
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		rcx, _var1
		call	DestroyBD
		jmp		Finish

		RET

	BFISqrt64 ENDP
	;;
	;@brief Computes the nth root of a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABI object to compute.
	;@param rdx - The root to compute.
	;@param r8 - The number of cycles to run.
	;;
	BINthRoot64 PROC _val: ABI, _nthroot: ABI, _cycles:qword

		;standard imp[lementation of the nthroot alg, this alg will run till two successive runs produce the same result

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_t3, r8
		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBI
		mov		_retval, rax
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, _v2
		mov		rdx, _retval
		ISub64
		mov		_var2, rax
		mov		rcx, _retval
		mov		rdx, _v2
		IDivide64
		mov		_var1, rax
		mov		rcx, _retval
		mov		rcx, [rcx]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		rcx, _var2
		mov		rdx, _retval
		IMultiply64
		mov		_t1, rax
		mov		rcx, _retval
		mov		rdx, _var2
		IPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _v1
		IDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBI
		mov		rcx, _t1
		mov		rdx, _t2
		IAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _t1
		IMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBI

MainLoop:

		cmp		_t3, 0
		jna		GoOut
		mov		rcx, _var2
		mov		rdx, _retval
		IMultiply64
		mov		_t1, rax
		mov		rcx, _retval
		mov		rdx, _var2
		IPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _v1
		IDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBI
		mov		rcx, _t1
		mov		rdx, _t2
		IAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBI
		mov		rcx, _var1
		mov		rdx, _t1
		IMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBI
		dec		_t3
		jmp		MainLoop

GoOut:

		mov		rcx, _t1
		call	DestroyBI
		mov		rcx, _t2
		call	DestroyBI
		mov		rcx, _var1
		call	DestroyBI
		mov		rcx, _var2
		call	DestroyBI
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BINthRoot64 ENDP
	;;
	;@brief Computes the nth root of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The root to compute (a  ABD object).
	;;
	BDNthRoot64 PROC _v1al: ABD, _nthroot: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_var3, rcx
		mov		_var4, rdx
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBD
		mov		_t2, rax
		call	CreateBD
		mov		_retval, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		call	CreateBD
		mov		_t1, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, _t1
		mov		rdx, _var4
		DDivide64
		mov		_var1, rax
		mov		rcx, _var4
		mov		rdx, _t1
		DSub64
		mov		_var2, rax
		call	CreateBD
		mov		_t5, rax
		mov		_v2, 0

MainLoop:

		mov		rcx, _t5
		mov		rdx, _retval
		DE64
		cmp		rax, 1
		je		GoOut
		inc		_v2
		cmp		_v2, 10000
		ja		GoOut
		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _retval
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t5, rax
		mov		rcx, _retval
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rdx, _var2
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var3
		mov		rdx, _t1
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _retval
		DMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _t1
		DAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var1
		mov		rdx, _t1
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _retval
		mov		rax, [rcx]. ABD._decimalplace
		cmp		rax, [rdx]. ABD._decimalplace
		jne		Next
		mov		rcx, [rcx]. ABD._num
		mov		rdx, [rdx]. ABD._num
		mov		rdi, qword ptr[rcx]
		cmp		rdi, qword ptr[rdx]
		jne		Next

CheckLoop:

		mov		rax, qword ptr[rcx+rdi*8]
		cmp		rax, qword ptr[rdx+rdi*8]
		jne		Next
		sub		rdi, 1
		jnz		CheckLoop
		mov		rax, _t1
		mov		rcx, _retval					;_retval=_t1
		mov		_retval, rax					;_t1=_retval
		mov		_t1, rcx
		jmp		GoOut

Next:

		mov		rax, _t1						;_retval=_t1
		mov		rcx, _retval					;_t1=_retval
		mov		_t1, rcx						
		mov		_retval, rax
		jmp		MainLoop

GoOut:

		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _var2
		call	DestroyBD

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDNthRoot64 ENDP
	;;
	;@brief Computes the nth root of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The root to compute (a  ABF object).
	;@param r8 - The number of cycles to run.
	;;
	BFNthRoot64 PROC _v1al: ABF, _nthroot: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_t3, r8
		mov		_var3, rcx
		mov		_var4, rdx
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBF
		mov		_t2, rax
		call	CreateBF
		mov		_retval, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 10
		mov		[rax]. ABF._sign, 0
		call	CreateBF
		mov		_t1, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABF._sign, 0
		mov		rcx, _t1
		mov		rdx, _var4
		FDivide64
		mov		_var1, rax
		mov		rcx, _var4
		mov		rdx, _t1
		FSub64
		mov		_var2, rax
		mov		_var5, 0

MainLoop:

		mov		rcx, _retval
		mov		rdx, _var2
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var3
		mov		rdx, _t1
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _retval
		FMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _t1
		FAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var1
		mov		rdx, _t1
		FMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _retval
		mov		rcx, [rcx]. ABF._denominator
		mov		rdx, [rdx]. ABF._denominator
		mov		rdi, qword ptr[rcx]
		cmp		rdi, qword ptr[rdx]
		jne		Next

CheckLoopd:

		mov		rax, qword ptr[rcx+rdi*8]
		cmp		rax, qword ptr[rdx+rdi*8]
		jne		Next
		sub		rdi, 1
		jnz		CheckLoopd
		mov		rcx, _t1
		mov		rdx, _retval
		mov		rcx, [rcx]. ABF._numerator
		mov		rdx, [rdx]. ABF._numerator
		mov		rdi, qword ptr[rcx]
		cmp		rdi, qword ptr[rdx]
		jne		Next

CheckLoop:

		mov		rax, qword ptr[rcx+rdi*8]
		cmp		rax, qword ptr[rdx+rdi*8]
		jne		Next
		sub		rdi, 1
		jnz		CheckLoop
		mov		rax, _t1
		mov		rcx, _retval					;_retval=_t1
		mov		_retval, rax					;_t1=_retval
		mov		_t1, rcx
		jmp		GoOut

Next:

		mov		rax, _t1						;_retval=_t1
		mov		rcx, _retval					;_t1=_retval
		mov		_t1, rcx						
		mov		_retval, rax
		inc		_var5
		mov		rax, _t3
		cmp		_var5, rax
		jna		MainLoop

GoOut:

		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _var2
		call	DestroyBD

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFNthRoot64 ENDP
	;;
	;@brief Computes the nth root of a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to compute.
	;@param rdx - The root to compute (a  ABDI object).
	;@param r8 - The number of cycles to run.
	;;
	BDINthRoot64 PROC _val: ABDI, _root: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		cmp		[rcx]. ABDI._polarform, 1
		je		MoveOn
		call	BDIToPolar
		mov		_var1, rax

MoveOn:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._real
		call	CreateBDFromBD
		mov		_t1, rax
		mov		[rax]. ABD._sign, 0
		mov		rcx, rax
		mov		rdx, _v2
		call	BDNthRoot64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		call	CreateBD
		mov		_t2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		mov		rcx, _t2
		mov		rdx, _bdpi
		DMultiply64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, _var1
		mov		rdx, [rdx]. ABDI._im
		DAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rax, [rdx]. ABD._num
		mov		rcx, _v2
		mov		rcx, [rcx]. ABD._num
		mov		rdi, 0

CopyLoop:

		mov		rdx, qword ptr[rcx+rdi*8]
		mov		qword ptr[rax+rdi*8], rdx
		inc		rdi
		cmp		rdi, qword ptr[rcx]
		jna		CopyLoop
		mov		rdx, _t2
		mov		rcx, _t3
		DDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBDITakeBothBD
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 1
		mov		rcx, _var1
		cmp		rcx, _v1
		je		GoOut
		call	DestroyBDI
		mov		rcx, _retval
		call	BDIFromPolar
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBDI

GoOut:

		mov		rax, _retval
		
		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDINthRoot64 ENDP
	;;
	;@brief Computes the nth root of a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to compute.
	;@param rdx - The root to compute (a  ABFI object).
	;@param r8 - The number of cycles to run.
	;;
	BFINthRoot64 PROC _val: ABFI, _root: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		cmp		[rcx]. ABFI._polarform, 1
		je		MoveOn
		call	BFIToPolar
		mov		_var1, rax

MoveOn:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABFI._real
		call	CreateBFFromBF
		mov		_t1, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, rax
		mov		rdx, _v2
		call	BFNthRoot64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		call	CreateBF
		mov		_t2, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		mov		rcx, _t2
		mov		rdx, _bfpi
		FMultiply64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, _var1
		mov		rdx, [rdx]. ABFI._im
		FAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rax, [rdx]. ABF._numerator
		mov		rcx, _v2
		mov		rcx, [rcx]. ABF._numerator
		mov		rdi, 0

CopyLoop:

		mov		rdx, qword ptr[rcx+rdi*8]
		mov		qword ptr[rax+rdi*8], rdx
		inc		rdi
		cmp		rdi, qword ptr[rcx]
		jna		CopyLoop
		mov		rdx, _t2
		mov		rcx, _t3
		FDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 1
		mov		rcx, _var1
		cmp		rcx, _v1
		je		GoOut
		call	DestroyBFI
		mov		rcx, _retval
		call	BFIFromPolar
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBFI

GoOut:

		mov		rax, _retval
		
		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFINthRoot64 ENDP
	;;
	;@brief Computes the exponential of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BDExp64 PROC _val: ABD, _cycles:qword

		;exponential function

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		[rcx]. ABD._decimalplace, 0
		je		AllGood
		call	CreateBDCopy
		mov		_t1, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		call	CreateBD
		mov		_var4, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		rdx, _v1
		mov		rdx, [rdx]. ABD._decimalplace
		mov		qword ptr[rcx+8], rdx
		mov		rdx, rax
		mov		rcx, _bdten
		DPow64
		mov		_var3, rax
		mov		rdx, rax
		mov		rcx, _t1
		DMod64
		mov		_var1, rax
		mov		rdx, rax
		mov		rcx, _t1
		DSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var3
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _var1
		mov		rcx, [rcx]. ABD._num
		mov		rdx, 10
		COUNTTHEDIGITS
		mov		_dp, rax
		mov		rcx, _v1
		cmp		rax, [rcx]. ABD._decimalplace
		jb		DontAdjust
		mov		rcx, _var1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t2, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0

GetLeadingDigitLoop:

		mov		rcx, _t2
		mov		rax, [rcx]. ABD._num
		cmp		qword ptr[rax], 1
		ja		CarryOn
		cmp		qword ptr[rax+8], 0
		je		FoundDigit

CarryOn:

		mov		rdx, _bdten
		DMod64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _var3
		DSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _bdten
		DDivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		jmp		GetLeadingDigitLoop

FoundDigit:

		mov		rcx, _var3
		mov		rax, [rcx]. ABD._num
		cmp		qword ptr[rax+8], 5
		jb		DontAdjust
		mov		rcx, _t1
		mov		rdx, _bdone
		DAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _bdone
		mov		rdx, _var1
		mov		rax, _dp
		mov		[rdx]. ABD._decimalplace, rax
		DSub64
		mov		rcx, _var1
		mov		_var1, rax
		mov		[rax]. ABD._decimalplace, 0
		call	DestroyBD

DontAdjust:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rdx, rax
		mov		rcx, _bdone
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		[rcx]. ABD._sign, 0
		mov		[rcx]. ABD._decimalplace, 0
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2

MainLoop:

		mov		rcx, _var1
		mov		rdx, _t2
		DPow64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _t2
		call	BDFactorial64
		mov		rcx, _var4
		mov		_var4, rax
		call	DestroyBD
		mov		rcx, _var3
		mov		rdx, _var4
		DDivide64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var3
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rax, [rcx]. ABD._num
		add		qword ptr[rax+8], 1
		mov		rdx, _v2
		cmp		qword ptr[rax+8], rdx
		jna		MainLoop
		mov		rcx, _bde
		mov		rdx, _t1
		DPow64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var3
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _var4
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		jmp		GoOut

AllGood:

		mov		rcx, _bde
		mov		rdx, _v1
		DPow64
		mov		_retval, rax

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDExp64 ENDP
	;;
	;@brief Computes the exponential of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BFExp64 PROC _v1al: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_t2, rcx
		mov		_v2, rdx
		call	BFtoBD
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BDExp64
		mov		_v2, rax
		mov		rcx, rax
		call	BDtoBF
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _t2

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFExp64 ENDP
	;;
	;@brief Computes the natural logarithm of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BDln64 PROC _val: ABD, _cycles:qword
	
		;natural logarithm alg runs until it gets the same result on 2 successive runs

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABD._sign, 1
		je		NegInputError
		mov		_v1, rcx
		mov		_var3, rdx
		mov		_var2, 0
		mov		_var5, 0
		call	CreateBD
		mov		_t5, rax
		call	CreateBD
		mov		_t3, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_var1, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rcx, [rcx]. ABD._num
		mov		rdx, 10
		COUNTTHEDIGITS
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._decimalplace
		dec		rax
		mov		rdx, _var1
		mov		[rdx]. ABD._decimalplace, rax
		cmp		rax, rcx
		jb		OtherWay
		sub		rax, rcx
		jmp		PastOtherWay

OtherWay:

		sub		rcx, rax
		mov		rax, rcx
		mov		_var2, 1

PastOtherWay:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		[rcx]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._num
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], rax
		call	CreateBD
		mov		_var7, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		rcx, _var1
		mov		rdx, _var7
		call	BDNthRoot64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBD
		call	CreateBD
		mov		_var4, rax
		mov		rdx, [rax]. ABD._num
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 3
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		call	CreateBD
		mov		_t1, rax
		mov		rdx, rax
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		mov		rcx, _var1
		DSub64
		mov		_t2, rax
		mov		rcx, _var1
		mov		rdx, _t1
		DAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _t1
		DDivide64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _var1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx

MainLoop:

		mov		rcx, _retval
		mov		rdx, _t5
		DE64
		cmp		rax, 1
		je		MainOut
		inc		_var5
		cmp		_var5, 10000
		ja		MainOut
		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _retval
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t5, rax
		mov		rcx, _retval
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _var1
		mov		rdx, _var4
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var4
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _var4
		mov		rax, [rax]. ABD._num
		add		qword ptr[rax+8], 2
		jmp		MainLoop

MainOut:

		mov		rcx, _var1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 20
		mov		[rcx]. ABD._sign, 0
		mov		[rcx]. ABD._decimalplace, 0
		mov		rdx, _retval
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _var1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 23025851
		mov		[rcx]. ABD._decimalplace, 7
		mov		rdx, _t3
		DMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t3
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _var1
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _retval

NegInputError:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDln64 ENDP			
	;;
	;@brief Computes the natural logarithm of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BFln64 PROC _val: ABD, _cycles:qword
		
		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_var4, rdx	
		call	BFtoBD
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _var4
		call	BDln64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BDtoBF
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rax, _t1	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFln64 ENDP
	;;
	;@brief Computes the base 10 logarithm of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BDLog64 PROC _val: ABD, _cycles:qword

		;base 10 log alg

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABD._sign, 1
		je		NegInputError
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var5, 0
		call	CreateBD
		mov		_t5, rax
		call	CreateBD
		mov		_t4, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		call	CreateBD
		mov		_var2, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rcx, [rcx]. ABD._num
		mov		rdx, 10
		COUNTTHEDIGITS
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._decimalplace
		dec		rax
		mov		rdx, _t3
		mov		[rdx]. ABD._decimalplace, rax
		cmp		rax, rcx
		jb		OtherWay
		sub		rax, rcx
		jmp		PastOtherWay

OtherWay:

		sub		rcx, rax
		mov		rax, rcx
		mov		_var1, 1

PastOtherWay:

		mov		rcx, _var2
		mov		rdx, _var1
		mov		[rcx]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._num
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], rax
		mov		rcx, _t3
		mov		rdx, _t4
		call	BDNthRoot64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _v2
		call	BDln64
		mov		_retval, rax
		mov		rcx, _t3
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		[rcx]. ABD._decimalplace, 0
		mov		[rcx]. ABD._sign, 0
		mov		rdx, _v2
		call	BDln64
		mov		_t1, rax
		mov		rcx, _t3
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		mov		rdx, _t1
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 2
		mov		rdx, _t1
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var2
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

NegInputError:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDLog64 ENDP
	;;
	;@brief Computes the base 10 logarithm of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The number of cycles to run.
	;;
	BFLog64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		call	BFtoBD
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BDLog64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval
		call	BDtoBF
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

MallocError:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFLog64 ENDP
	;;
	;@brief Computes the power function of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The base  ABD object to compute.
	;@param rdx - The exponent  ABD object to compute.
	;@param r8 - The number of cycles to run.
	;;
	BDPower64 PROC _base: ABD, _exp: ABD, _cycles:qword

		;power function

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, rdx
		mov		_t2, r8
		call	BDln64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _t1
		DMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t2
		call	BDExp64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDPower64 ENDP
	;;
	;@brief Computes the power function of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The base  ABF object to compute.
	;@param rdx - The exponent  ABF object to compute.
	;@param r8 - The number of cycles to run.
	;;
	BFPower64 PROC _base: ABF, _t1xp: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, rdx
		mov		_t2, r8
		call	BFtoBD
		mov		_v1, rax
		mov		rcx, _t1
		call	BFtoBD
		mov		_t1, rax
		mov		rdx, rax
		mov		rcx, _v1
		mov		r8, _t2
		call	BDPower64
		mov		_retval, rax
		mov		rcx, rax
		call	BDtoBF
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _v1
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFPower64 ENDP
	;;
	;@brief Reduces the  ABD object representing the radians for the sin and cos functions.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The offset.
	;;
	BDRadReduceIt64 PROC _val: ABD, _offset:qword

		;this is going to be the main part for the sin and cos algs, they both go through the same methods to reduce the variable
		;the offset sent in will control the path this alg takes

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v3, rdx
		call	CreateBDFromBD
		mov		_v1, rax
		mov		rcx, rax
		mov		rdx, _bdtorad
		DDivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	DestroyBD
		mov		rax, [rcx]. ABD._num
		mov		_var1, 0
		mov		_var2, 0
		mov		_t1, 0
		mov		_t3, 0
		mov		_t2, 0
		mov		_var8, 1
		mov		_var6, 0
		mov		_var7, 0
		mov		_var5, 0
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rdx, _v1
		mov		rdi, [rdx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdi
		mov		rdi, [rdx]. ABD._sign
		mov		[rax]. ABD._sign, rdi
		call	CreateBD
		mov		_t1, rax
		call	CreateBD
		mov		_t2, rax
		call	CreateBD
		mov		_var6, rax
		mov		[rax]. ABD._sign, 0
		call	CreateBD
		mov		_var7, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 4
		call	CreateBD
		mov		_var5, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBD
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBD
		mov		_var2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 360
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rdx, rax
		mov		rcx, _t3
		DLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		DMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		cmp		[rcx]. ABD._sign, 0
		je		Skip2
		mov		rdx, _var2
		DAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 270
		mov		rcx, _t3;
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		rdx, _var2
		mov		rcx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 0
		je		SkipIt
		mov		_var1, 1
		jmp		SkipIt

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 180
		DLT64
		cmp		rax, 1
		je		Skip4
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1
		jmp		SkipIt

Skip4:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 90
		DLT64
		cmp		rax, 1
		je		SkipIt
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 1
		je		SkipIt
		mov		_var1, 1

SkipIt:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 45
		DLT64
		cmp		rax, 1
		je		Prep
		cmp		_v3, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v3, rax
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Prep:

		mov		_var4, 0
		mov		rcx, _t3
		mov		rdx, _bdtorad
		DMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		call	CreateBD
		mov		_t5, rax
		call	CreateBD
		mov		_var2, rax
		call	CreateBD
		mov		_var3, rax
		call	CreateBD
		mov		_t1, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		rcx, _v3
		add		qword ptr[rax+8], rcx
		cmp		_v3, 1
		je		CreateX
		call	CreateBD
		mov		_retval, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		jmp		Start

CreateX:

		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx

Start:

		mov		rcx, _t5
		mov		rdx, _retval
		DE64
		cmp		rax, 1
		je		GoOut
		inc		_var4
		cmp		_var4, 10000
		ja		GoOut
		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _retval
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t5, rax
		mov		rcx, _retval
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _t3
		mov		rdx, _t1
		DPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BDFactorial64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _var3
		DDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rax, _t1
		mov		rax, [rax]. ABD._num
		mov		rax, qword ptr[rax+8]
		sub		rax, _v3
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoAdd
		mov		rcx, _retval
		mov		rdx, _var2
		DSub64
		jmp		PastAdd

DoAdd:

		mov		rcx, _retval
		mov		rdx, _var2
		DAdd64

PastAdd:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rax, [rcx]. ABD._num
		add		qword ptr[rax+8], 2
		jmp		Start

GoOut:

		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _var2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		cmp		_var1, 0
		je		Noneg
		mov		rax, _retval
		mov		[rax]. ABD._sign, 1

NoNeg:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDRadReduceIt64 ENDP
	;;
	;@brief Reduces the  ABF object representing the radians for the sin and cos functions.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The offset.
	;;
	BFRadReduceIt64 PROC _val: ABF, _offset:qword, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v3, rdx
		mov		_v2, r8
		call	CreateBFFromBF
		mov		_v1, rax
		mov		rcx, rax
		mov		rdx, _bftorad
		FDivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	DestroyBD
		mov		_var1, 0
		mov		_var2, 0
		mov		_t1, 0
		mov		_t3, 0
		mov		_t2, 0
		mov		_var8, 1
		mov		_var6, 0
		mov		_var7, 0
		mov		_var5, 0
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rdx, _v1
		mov		rdi, [rdx]. ABF._sign
		mov		[rax]. ABF._sign, rdi
		call	CreateBF
		mov		_t1, rax
		call	CreateBF
		mov		_t2, rax
		call	CreateBF
		mov		_var6, rax
		mov		[rax]. ABD._sign, 0
		call	CreateBF
		mov		_var7, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 4
		call	CreateBF
		mov		_var5, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBF
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBF
		mov		_var2, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 360
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rdx, rax
		mov		rcx, _t3
		FLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		FMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		cmp		[rcx]. ABF._sign, 0
		je		Skip2
		mov		rdx, _var2
		FAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 270
		mov		rcx, _t3;
		FLT64
		cmp		rax, 1
		je		Skip3
		mov		rdx, _var2
		mov		rcx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 0
		je		SkipIt
		mov		_var1, 1
		jmp		SkipIt

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 180
		FLT64
		cmp		rax, 1
		je		Skip4
		mov		rcx, _t3
		mov		rdx, _var2
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1
		jmp		SkipIt

Skip4:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		FLT64
		cmp		rax, 1
		je		SkipIt
		mov		rcx, _t3
		mov		rdx, _var2
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 1
		je		SkipIt
		mov		_var1, 1

SkipIt:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 45
		FLT64
		cmp		rax, 1
		je		Prep
		cmp		_v3, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v3, rax
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Prep:

		mov		_var4, 0
		mov		rcx, _t3
		mov		rdx, _bftorad
		FMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		call	CreateBF
		mov		_var2, rax
		call	CreateBF
		mov		_var3, rax
		call	CreateBF
		mov		_t1, rax
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		rcx, _v3
		add		qword ptr[rax+8], rcx
		cmp		_v3, 1
		je		CreateX
		call	CreateBF
		mov		_retval, rax
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		jmp		Start

CreateX:

		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3

Start:

		inc		_var4
		mov		rax, _v2
		cmp		_var4, rax
		ja		GoOut
		mov		rcx, _retval
		mov		rcx, _t3
		mov		rdx, _t1
		FPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BFFactorial64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _var3
		FDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rax, _t1
		mov		rax, [rax]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		sub		rax, _v3
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoAdd
		mov		rcx, _retval
		mov		rdx, _var2
		FSub64
		jmp		PastAdd

DoAdd:

		mov		rcx, _retval
		mov		rdx, _var2
		FAdd64

PastAdd:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rax, [rcx]. ABF._numerator
		add		qword ptr[rax+8], 2
		jmp		Start

GoOut:

		mov		rcx, _var2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		cmp		_var1, 0
		je		Noneg
		mov		rax, _retval
		mov		[rax]. ABF._sign, 1

NoNeg:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFRadReduceIt64 ENDP
	;;
	;@brief Computes the sin of the  ABD object given in radians.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDRadSin64 PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, 1
		call	BDReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDRadSin64 ENDP
	;;
	;@brief Computes the sin of the  ABF object given in radians.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg.
	;;
	BFRadSin64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r8, rdx
		mov		rdx, 1
		call	BFReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFRadSin64 ENDP
	;;
	;@brief Computes the cos of the  ABD object given in radians.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDRadCos64 PROC _v1al: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, 0
		call	BDReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDRadCos64 ENDP
	;;
	;@brief Computes the cos of the  ABF object given in radians.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg.
	;;
	BFRadCos64 PROC _v1al: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r8, rdx
		mov		rdx, 0
		call	BFReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFRadCos64 ENDP
	;;
	;@brief Reduces a  ABD object representing degrees for the sin and cos functions.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The offset.
	;;
	BDReduceIt64 PROC _val: ABD, _offset:qword

		;this is going to be the main part for the sin and cos algs, they both go through the same methods to reduce the variable
		;the offset sent in will control the path this alg takes

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, [rcx]. ABD._num
		mov		_v1, rcx
		mov		_v3, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t1, 0
		mov		_t3, 0
		mov		_t2, 0
		mov		_var8, 1
		mov		_var6, 0
		mov		_var7, 0
		mov		_var5, 0
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rdx, _v1
		mov		rdi, [rdx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdi
		mov		rdi, [rdx]. ABD._sign
		mov		[rax]. ABD._sign, rdi
		call	CreateBD
		mov		_t1, rax
		call	CreateBD
		mov		_t2, rax
		call	CreateBD
		mov		_var6, rax
		mov		[rax]. ABD._sign, 0
		call	CreateBD
		mov		_var7, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 4
		call	CreateBD
		mov		_var5, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBD
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBD
		mov		_var2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 360
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rdx, rax
		mov		rcx, _t3
		DLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		DMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		cmp		[rcx]. ABD._sign, 0
		je		Skip2
		mov		rdx, _var2
		DAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 270
		mov		rcx, _t3;
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		rdx, _var2
		mov		rcx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 0
		je		SkipIt
		mov		_var1, 1
		jmp		SkipIt

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 180
		DLT64
		cmp		rax, 1
		je		Skip4
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1
		jmp		SkipIt

Skip4:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 90
		DLT64
		cmp		rax, 1
		je		SkipIt
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 1
		je		SkipIt
		mov		_var1, 1

SkipIt:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 45
		DLT64
		cmp		rax, 1
		je		Prep
		cmp		_v3, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v3, rax
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Prep:

		mov		_var4, 0
		mov		rcx, _t3
		mov		rdx, _bdtorad
		DMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		call	CreateBD
		mov		_t5, rax
		call	CreateBD
		mov		_var2, rax
		call	CreateBD
		mov		_var3, rax
		call	CreateBD
		mov		_t1, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		rcx, _v3
		add		qword ptr[rax+8], rcx
		cmp		_v3, 1
		je		CreateX
		call	CreateBD
		mov		_retval, rax
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		jmp		Start

CreateX:

		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx

Start:

		mov		rcx, _t5
		mov		rdx, _retval
		DE64
		cmp		rax, 1
		je		GoOut
		inc		_var4
		cmp		_var4, 10000
		ja		GoOut
		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _retval
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t5, rax
		mov		rcx, _retval
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _t3
		mov		rdx, _t1
		DPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BDFactorial64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _var3
		DDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rax, _t1
		mov		rax, [rax]. ABD._num
		mov		rax, qword ptr[rax+8]
		sub		rax, _v3
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoAdd
		mov		rcx, _retval
		mov		rdx, _var2
		DSub64
		jmp		PastAdd

DoAdd:

		mov		rcx, _retval
		mov		rdx, _var2
		DAdd64

PastAdd:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rax, [rcx]. ABD._num
		add		qword ptr[rax+8], 2
		jmp		Start

GoOut:

		mov		rcx, _t5
		call	DestroyBD
		mov		rcx, _var2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		cmp		_var1, 0
		je		Noneg
		mov		rax, _retval
		mov		[rax]. ABD._sign, 1

NoNeg:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDReduceIt64 ENDP
	;;
	;@brief Reduces a  ABF object representing degrees for the sin and cos functions.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The offset.
	;;
	BFReduceIt64 PROC _val: ABF, _offset:qword, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v3, rdx
		mov		_v2, r8
		mov		_var1, 0
		mov		_var2, 0
		mov		_t1, 0
		mov		_t3, 0
		mov		_t2, 0
		mov		_var8, 1
		mov		_var6, 0
		mov		_var7, 0
		mov		_var5, 0
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rdx, _v1
		mov		rdi, [rdx]. ABF._sign
		mov		[rax]. ABF._sign, rdi
		call	CreateBF
		mov		_t1, rax
		call	CreateBF
		mov		_t2, rax
		call	CreateBF
		mov		_var6, rax
		mov		[rax]. ABD._sign, 0
		call	CreateBF
		mov		_var7, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 4
		call	CreateBF
		mov		_var5, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBF
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBF
		mov		_var2, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 360
		mov		[rax]. ABF._sign, 0
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		rdx, rax
		mov		rcx, _t3
		FLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		FMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		cmp		[rcx]. ABF._sign, 0
		je		Skip2
		mov		rdx, _var2
		FAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 270
		mov		rcx, _t3;
		FLT64
		cmp		rax, 1
		je		Skip3
		mov		rdx, _var2
		mov		rcx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 0
		je		SkipIt
		mov		_var1, 1
		jmp		SkipIt

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 180
		FLT64
		cmp		rax, 1
		je		Skip4
		mov		rcx, _t3
		mov		rdx, _var2
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1
		jmp		SkipIt

Skip4:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		FLT64
		cmp		rax, 1
		je		SkipIt
		mov		rcx, _t3
		mov		rdx, _var2
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		cmp		_v3, 1
		je		SkipIt
		mov		_var1, 1

SkipIt:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 45
		FLT64
		cmp		rax, 1
		je		Prep
		cmp		_v3, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v3, rax
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Prep:

		mov		_var4, 0
		mov		rcx, _t3
		mov		rdx, _bftorad
		FMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		call	CreateBF
		mov		_var2, rax
		call	CreateBF
		mov		_var3, rax
		call	CreateBF
		mov		_t1, rax
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		rcx, _v3
		add		qword ptr[rax+8], rcx
		cmp		_v3, 1
		je		CreateX
		call	CreateBF
		mov		_retval, rax
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		jmp		Start

CreateX:

		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3

Start:

		inc		_var4
		mov		rax, _v2
		cmp		_var4, rax
		ja		GoOut
		mov		rcx, _retval
		mov		rcx, _t3
		mov		rdx, _t1
		FPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rcx, _t1
		call	BFFactorial64
		mov		rcx, _var3
		mov		_var3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _var3
		FDivide64
		mov		rcx, _var2
		mov		_var2, rax
		call	DestroyBD
		mov		rax, _t1
		mov		rax, [rax]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		sub		rax, _v3
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoAdd
		mov		rcx, _retval
		mov		rdx, _var2
		FSub64
		jmp		PastAdd

DoAdd:

		mov		rcx, _retval
		mov		rdx, _var2
		FAdd64

PastAdd:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rax, [rcx]. ABF._numerator
		add		qword ptr[rax+8], 2
		jmp		Start

GoOut:

		mov		rcx, _var2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t3
		call	DestroyBD
		cmp		_var1, 0
		je		Noneg
		mov		rax, _retval
		mov		[rax]. ABF._sign, 1

NoNeg:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFReduceIt64 ENDP
	;;
	;@brief Computes the sin of a  ABD object given in degrees.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDSin64 PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, 1
		call	BDReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDSin64 ENDP
	;;
	;@brief Computes the sin of a  ABF object given in degrees.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BFSin64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r8, rdx
		mov		rdx, 1
		call	BFReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFSin64 ENDP
	;;
	;@brief Computes the cos of a  ABD object given in degrees.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDCos64 PROC _v1al: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, 0
		call	BDReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDCos64 ENDP
	;;
	;@brief Computes the cos of a  ABF object given in degrees.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BFCos64 PROC _v1al: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r8, rdx
		mov		rdx, 0
		call	BFReduceIt64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFCos64 ENDP

	BDCos642 PROC _val: ABD, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_var3, rdx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_var2, 0
		mov		_t1, 0
		mov		_t3, 0
		mov		_t2, 0
		mov		_var8, 1
		mov		_var6, 0
		mov		_var7, 0
		mov		_var5, 0
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rdx, _v1
		mov		rdi, [rdx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdi
		mov		rdi, [rdx]. ABD._sign
		mov		[rax]. ABD._sign, rdi
		call	CreateBD
		mov		_t1, rax
		call	CreateBD
		mov		_t2, rax
		call	CreateBD
		mov		_var6, rax
		mov		[rax]. ABD._sign, 0
		call	CreateBD
		mov		_var7, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 4
		call	CreateBD
		mov		_var5, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 2
		call	CreateBD
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0
		call	CreateBD
		mov		_var2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 360
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rdx, rax
		mov		rcx, _t3
		DLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		DMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		cmp		[rcx]. ABD._sign, 0
		je		Skip2
		mov		rdx, _var2
		DAdd64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 270
		mov		rcx, _t3;
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		rdx, _var2
		mov		rcx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		jmp		Prep

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 180
		DLT64
		cmp		rax, 1
		je		Skip4
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1
		jmp		Prep

Skip4:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 90
		DLT64
		cmp		rax, 1
		je		Prep
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1

Prep:

		mov		rcx, _t3
		mov		rdx, _bdtorad
		DMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip5:

		mov		rcx, _bdpi
		mov		rdx, _var7
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _t1
		DLT64
		cmp		rax, 1
		je		Next
		mov		rax, _var8
		mov		_var9, rax
		mov		rcx, _var6
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		[rcx]. ABD._sign, 0
		mov		[rcx]. ABD._decimalplace, 0

InnerLoop:

		mov		rcx, _var6
		mov		rdx, _var5
		DAdd64
		mov		rcx, _var6
		mov		_var6, rax
		call	DestroyBD
		mov		rcx, _var6
		mov		rdx, _var3
		call	BDSqrt64
		mov		rcx, _var6
		mov		_var6, rax
		sub		_var9, 1
		jnz		InnerLoop
		mov		rcx, _var6
		mov		rdx, _var5
		DDivide64
		mov		rcx, _var6
		mov		_var6, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var6
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _t1
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		cmp		qword ptr[rcx], 1
		ja		Next
		cmp		qword ptr[rcx+8], 0
		je		GoOut

Next:

		inc		_var8
		mov		rcx, _var7
		mov		rdx, _var5
		DMultiply64
		mov		rcx, _var7
		mov		_var7, rax
		call	DestroyBD
		sub		_v2, 1
		jnz		Skip5

GoOut:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _var2
		call	DestroyBD

Skip7:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		cmp		_t3, 0
		je		Skip10
		mov		rcx, _t3
		call	DestroyBD

Skip10:

		cmp		_t1, 0
		je		Skip11
		mov		rcx, _t1
		call	DestroyBD

Skip11:

		cmp		_var6, 0
		je		Skip12
		mov		rcx, _var6
		call	DestroyBD

Skip12:

		cmp		_var7, 0
		je		Skip13
		mov		rcx, _var7
		call	DestroyBD

Skip13:

		cmp		_var5, 0
		je		Skip14
		mov		rcx, _var5
		call	DestroyBD

Skip14:

		mov		rax, _retval
		mov		rcx, _var1
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

Return45:

	BDCos642 ENDP
	;;
	;@brief Computes the tan of a  ABD object given in degrees.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;;
	BDTan64 PROC _val:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_t3, 0
		mov		_t1, 0
		mov		_t2, 0
		mov		_var2, 0
		mov		_var3, 0
		mov		_var4, 0
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		call	CreateBD
		mov		_t1, rax
		call	CreateBD
		mov		_t2, rax
		call	CreateBD
		mov		_var2, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 180
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rdx, rax
		mov		rcx, _t3
		DLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		DMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 90
		DLT64
		cmp		rax, 1
		je		Skip2
		mov		rcx, _t3
		mov		rdx, _var2
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1

Skip2:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 45
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		_var3, 1
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		DSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 225
		mov		[rdx]. ABD._decimalplace, 1
		DLT64
		cmp		rax, 1
		je		Skip4
		mov		_var4, 1
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABD._decimalplace, 0
		DDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip4:

		mov		rcx, _t3
		mov		rdx, _bdtorad
		DMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 3
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 5
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 2
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 15
		DDivide64
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 7
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 17
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 315
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		cmp		_var4, 0
		je		Skip5
		mov		rcx, _retval
		mov		rdx, _var2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax+8], 2
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var2
		DPow64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		mov		rdx, _t2
		DSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		DDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var3, 0
		je		Skip6
		mov		rcx, _var2
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		mov		rdx, _retval
		DDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _var2
		call	DestroyBD

Skip7:

		cmp		_t1, 0
		je		Skip8
		mov		rcx, _t1		
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		cmp		_t3, 0
		je		GoOut
		mov		rcx, _t3
		call	DestroyBD
		mov		rax, _retval
		mov		rcx, _var1
		mov		[rax]. ABD._sign, rcx

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDTan64 ENDP
	;;
	;@brief Computes the tan of a  ABF object given in degrees.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;;
	BFTan64 PROC _v1al: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, 0
		mov		_t3, 0
		mov		_t1, 0
		mov		_t2, 0
		mov		_var2, 0
		mov		_var3, 0
		mov		_var4, 0
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		call	CreateBF
		mov		_t1, rax
		call	CreateBF
		mov		_t2, rax
		call	CreateBF
		mov		_var2, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 180
		mov		[rax]. ABF._sign, 0
		mov		rdx, rax
		mov		rcx, _t3
		FLT64
		cmp		rax, 1
		je		Skip1
		mov		rcx, _t3
		mov		rdx, _var2
		FMod64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip1:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		FLT64
		cmp		rax, 1
		je		Skip2
		mov		rcx, _t3
		mov		rdx, _var2
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		_var1, 1

Skip2:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 45
		FLT64
		cmp		rax, 1
		je		Skip3
		mov		_var3, 1
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 90
		mov		rdx, _t3
		FSub64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 225
		mov		rax, [rdx]. ABF._denominator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		FLT64
		mov		rcx, _var2
		mov		rcx, [rcx]. ABF._denominator
		mov		qword ptr[rcx+8], 1
		cmp		rax, 1
		je		Skip4
		mov		_var4, 1
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 2
		FDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip4:

		mov		rcx, _t3
		mov		rdx, _bftorad
		FMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 3
		mov		[rdx]. ABF._sign, 0
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 5
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 2
		FMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 15
		FDivide64
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t3
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 7
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 17
		FMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 315
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		cmp		_var4, 0
		je		Skip5
		mov		rcx, _retval
		mov		rdx, _var2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax+8], 2
		FMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var2
		FPow64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 1
		mov		rdx, _t2
		FSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		FDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var3, 0
		je		Skip6
		mov		rcx, _var2
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 1
		mov		rdx, _retval
		FDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _var2
		call	DestroyBD

Skip7:

		cmp		_t1, 0
		je		Skip8
		mov		rcx, _t1		
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		cmp		_t3, 0
		je		GoOut
		mov		rcx, _t3
		call	DestroyBD
		mov		rax, _retval
		mov		rcx, _var1
		mov		[rax]. ABF._sign, rcx

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFTan64 ENDP
	;;
	;@brief Computes the Arcsin of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BDASin64 PROC _val: ABD, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v2, rdx
		mov		_v1, rcx
		call	CreateBD
		mov		_t1, rax
		mov		rdx, rax
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABD._decimalplace, 0
		mov		[rdx]. ABD._sign, 0
		mov		rcx, _v1
		DPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		DSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v2
		call	BDSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rdx, _t2
		mov		rcx, _v1
		DDivide64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BDATan64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDASin64 ENDP
	;;
	;@brief Computes the Arcsin of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BFASin64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v2, rdx
		mov		_v1, rcx
		call	CreateBF
		mov		_t1, rax
		mov		rdx, rax
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _v1
		FPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 1
		FSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v2
		call	BFSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rdx, _t2
		mov		rcx, _v1
		FDivide64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BFATan64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFASin64 ENDP
	;;
	;@brief Computes the Arccos of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BDACos64 PROC _val: ABD, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v2, rdx
		mov		_v1, rcx
		call	CreateBD
		mov		_t1, rax
		mov		rdx, rax
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABD._decimalplace, 0
		mov		[rdx]. ABD._sign, 0
		mov		rcx, _v1
		DPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax+8], 1
		DSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v2
		call	BDSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v1
		DDivide64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BDATan64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDACos64 ENDP
	;;
	;@brief Computes the Arccos of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BFACos64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v2, rdx
		mov		_v1, rcx
		call	CreateBF
		mov		_t1, rax
		mov		rdx, rax
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _v1
		FPow64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax+8], 1
		FSub64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v2
		call	BFSqrt64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t2
		mov		rdx, _v1
		FDivide64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v2
		call	BFATan64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFACos64 ENDP
	;;
	;@brief Computes the Arctan of a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABD object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BDATan64 PROC _val: ABD, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_var2, 0
		mov		_v1, rcx
		mov		_var1, rdx
		mov		_v2, 0
		mov		_var4, 0
		mov		_var5, 0
		mov		_var6, 0
		mov		_var3, 0
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBD
		mov		_var3, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		cmp		[rax]. ABD._sign, 0
		je		Skip1
		mov		_var2, 1
		mov		[rax]. ABD._sign, 0

Skip1:

		mov		rcx, _t3
		mov		rdx, _var3
		DLT64
		cmp		rax, 1
		je		Skip2
		mov		_var5, 1
		mov		rcx, _var3
		mov		rdx, _t3
		DDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rcx, _t3
		mov		rdx, _bd2sqr3
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		_var6, 1
		mov		rcx, _bdsqr3
		mov		rdx, _t3
		DMultiply64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _var3
		DSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _bdsqr3
		mov		rdx, _t3
		DAdd64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		DDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _var3
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 3
		mov		[rcx]. ABD._sign, 0
		mov		[rcx]. ABD._decimalplace, 0
		mov		_v2, rcx
		mov		_var3, 0
		cmp		_t1, 0
		jne		Skip4
		call	CreateBD
		mov		_t1, rax

Skip4:

		cmp		_t2, 0
		jne		MainLoop
		call	CreateBD
		mov		_t2, rax

MainLoop:

		mov		rcx, _t3
		mov		rdx, _v2
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _v2
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v2
		mov		rax, [rcx]. ABD._num
		mov		rax, qword ptr[rax+8]
		inc		rax
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoSub
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		jmp		PastSub

DoSub:

		mov		rcx, _retval
		mov		rdx, _t1
		DSub64

PastSub:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rax, [rcx]. ABD._num
		add		qword ptr[rax+8], 2
		cmp		qword ptr[rax+8], rdx
		jna		MainLoop
		cmp		_var6, 0
		je		Skip5
		mov		rdx, _v2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 6
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		mov		rcx, _bdpi
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var5, 0
		je		Skip6
		mov		rdx, _v2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		mov		rcx, _bdpi
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DSub64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _retval
		cmp		[rcx]. ABD._sign, 0
		cmove	rax, _one
		cmovne	rax, _zero
		mov		[rcx]. ABD._sign, rax
		cmp		_v2, 0
		je		Skip7
		mov		rcx, _v2
		call	DestroyBD

Skip7:

		cmp		_t3, 0
		je		Skip8
		mov		rcx, _t3
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		mov		rcx, _retval
		mov		rdx, _bdtorad
		DDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval
	
		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		RET

	BDATan64 ENDP
	;;
	;@brief Computes the Arctan of a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABF object to compute.
	;@param rdx - The cycles to run the alg through.
	;;
	BFATan64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_var2, 0
		mov		_v1, rcx
		mov		_var1, rdx
		mov		_v2, 0
		mov		_var4, 0
		mov		_var5, 0
		mov		_var6, 0
		mov		_var3, 0
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBF
		mov		_var3, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABF._sign, 0
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		cmp		[rax]. ABF._sign, 0
		je		Skip1
		mov		_var2, 1
		mov		[rax]. ABF._sign, 0

Skip1:

		mov		rcx, _t3
		mov		rdx, _var3
		FLT64
		cmp		rax, 1
		je		Skip2
		mov		_var5, 1
		mov		rcx, _var3
		mov		rdx, _t3
		FDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rcx, _t3
		mov		rdx, _bf2sqr3
		FLT64
		cmp		rax, 1
		je		Skip3
		mov		_var6, 1
		mov		rcx, _bfsqr3
		mov		rdx, _t3
		FMultiply64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _var3
		FSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _bfsqr3
		mov		rdx, _t3
		FAdd64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		FDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		mov		rcx, _var3
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 3
		mov		[rcx]. ABF._sign, 0
		mov		_v2, rcx
		mov		_var3, 0
		cmp		_t1, 0
		jne		Skip4
		call	CreateBF
		mov		_t1, rax

Skip4:

		cmp		_t2, 0
		jne		MainLoop
		call	CreateBF
		mov		_t2, rax

MainLoop:

		mov		rcx, _t3
		mov		rdx, _v2
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _v2
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v2
		mov		rax, [rcx]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		inc		rax
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoSub
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		jmp		PastSub

DoSub:

		mov		rcx, _retval
		mov		rdx, _t1
		FSub64

PastSub:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rax, [rcx]. ABF._numerator
		add		qword ptr[rax+8], 2
		cmp		qword ptr[rax+8], rdx
		jna		MainLoop
		cmp		_var6, 0
		je		Skip5
		mov		rdx, _v2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 6
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _bfpi
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var5, 0
		je		Skip6
		mov		rdx, _v2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _bfpi
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FSub64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _retval
		cmp		[rcx]. ABF._sign, 0
		cmove	rax, _one
		cmovne	rax, _zero
		mov		[rcx]. ABF._sign, rax
		cmp		_v2, 0
		je		Skip7
		mov		rcx, _v2
		call	DestroyBD

Skip7:

		cmp		_t3, 0
		je		Skip8
		mov		rcx, _t3
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		mov		rcx, _retval
		mov		rdx, _bftorad
		FDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		RET

	BFATan64 ENDP

	BDRadATan64 PROC _val: ABD, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_var2, 0
		mov		_v1, rcx
		mov		_var1, rdx
		mov		_v2, 0
		mov		_var4, 0
		mov		_var5, 0
		mov		_var6, 0
		mov		_var3, 0
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBD
		mov		_var3, rax
		mov		rcx, [rax]. ABD._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		cmp		[rax]. ABD._sign, 0
		je		Skip1
		mov		_var2, 1
		mov		[rax]. ABD._sign, 0

Skip1:

		mov		rcx, _t3
		mov		rdx, _var3
		DLT64
		cmp		rax, 1
		je		Skip2
		mov		_var5, 1
		mov		rcx, _var3
		mov		rdx, _t3
		DDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rcx, _t3
		mov		rdx, _bd2sqr3
		DLT64
		cmp		rax, 1
		je		Skip3
		mov		_var6, 1
		mov		rcx, _bdsqr3
		mov		rdx, _t3
		DMultiply64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _var3
		DSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _bdsqr3
		mov		rdx, _t3
		DAdd64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		DDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rcx, [rcx]. ABD._num
		call	CreateBDCopy
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rcx, _var3
		mov		rax, [rcx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 3
		mov		[rcx]. ABD._sign, 0
		mov		[rcx]. ABD._decimalplace, 0
		mov		_v2, rcx
		mov		_var3, 0
		cmp		_t1, 0
		jne		Skip4
		call	CreateBD
		mov		_t1, rax

Skip4:

		cmp		_t2, 0
		jne		MainLoop
		call	CreateBD
		mov		_t2, rax

MainLoop:

		mov		rcx, _t3
		mov		rdx, _v2
		DPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _v2
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v2
		mov		rax, [rcx]. ABD._num
		mov		rax, qword ptr[rax+8]
		inc		rax
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoSub
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		jmp		PastSub

DoSub:

		mov		rcx, _retval
		mov		rdx, _t1
		DSub64

PastSub:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rax, [rcx]. ABD._num
		add		qword ptr[rax+8], 2
		cmp		qword ptr[rax+8], rdx
		jna		MainLoop
		cmp		_var6, 0
		je		Skip5
		mov		rdx, _v2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 6
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		mov		rcx, _bdpi
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var5, 0
		je		Skip6
		mov		rdx, _v2
		mov		rax, [rdx]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABD._sign, 0
		mov		[rdx]. ABD._decimalplace, 0
		mov		rcx, _bdpi
		DDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		DSub64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _retval
		cmp		[rcx]. ABD._sign, 0
		cmove	rax, _one
		cmovne	rax, _zero
		mov		[rcx]. ABD._sign, rax
		cmp		_v2, 0
		je		Skip7
		mov		rcx, _v2
		call	DestroyBD

Skip7:

		cmp		_t3, 0
		je		Skip8
		mov		rcx, _t3
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		mov		rax, _retval
	
		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		RET

	BDRadATan64 ENDP

	BFRadATan64 PROC _val: ABF, _cycles:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_var2, 0
		mov		_v1, rcx
		mov		_var1, rdx
		mov		_v2, 0
		mov		_var4, 0
		mov		_var5, 0
		mov		_var6, 0
		mov		_var3, 0
		mov		_t1, 0
		mov		_t2, 0
		call	CreateBF
		mov		_var3, rax
		mov		rcx, [rax]. ABF._numerator
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 1
		mov		[rax]. ABF._sign, 0
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_t3, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		cmp		[rax]. ABF._sign, 0
		je		Skip1
		mov		_var2, 1
		mov		[rax]. ABF._sign, 0

Skip1:

		mov		rcx, _t3
		mov		rdx, _var3
		FLT64
		cmp		rax, 1
		je		Skip2
		mov		_var5, 1
		mov		rcx, _var3
		mov		rdx, _t3
		FDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip2:

		mov		rcx, _t3
		mov		rdx, _bf2sqr3
		FLT64
		cmp		rax, 1
		je		Skip3
		mov		_var6, 1
		mov		rcx, _bfsqr3
		mov		rdx, _t3
		FMultiply64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _var3
		FSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _bfsqr3
		mov		rdx, _t3
		FAdd64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		FDivide64
		mov		rcx, _t3
		mov		_t3, rax
		call	DestroyBD

Skip3:

		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._denominator
		mov		rcx, [rcx]. ABF._numerator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3
		mov		rdx, [rcx]. ABF._sign
		mov		[rax]. ABF._sign, rdx
		mov		rcx, _var3
		mov		rax, [rcx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 3
		mov		[rcx]. ABF._sign, 0
		mov		_v2, rcx
		mov		_var3, 0
		cmp		_t1, 0
		jne		Skip4
		call	CreateBF
		mov		_t1, rax

Skip4:

		cmp		_t2, 0
		jne		MainLoop
		call	CreateBF
		mov		_t2, rax

MainLoop:

		mov		rcx, _t3
		mov		rdx, _v2
		FPow64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _v2
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v2
		mov		rax, [rcx]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		inc		rax
		xor		rdx, rdx
		div		_four
		cmp		rdx, 0
		je		DoSub
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		jmp		PastSub

DoSub:

		mov		rcx, _retval
		mov		rdx, _t1
		FSub64

PastSub:

		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rdx, _var1
		mov		rcx, _v2
		mov		rax, [rcx]. ABF._numerator
		add		qword ptr[rax+8], 2
		cmp		qword ptr[rax+8], rdx
		jna		MainLoop
		cmp		_var6, 0
		je		Skip5
		mov		rdx, _v2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 6
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _bfpi
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip5:

		cmp		_var5, 0
		je		Skip6
		mov		rdx, _v2
		mov		rax, [rdx]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		[rdx]. ABF._sign, 0
		mov		rcx, _bfpi
		FDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _t1
		FSub64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

Skip6:

		cmp		_var2, 0
		je		Skip7
		mov		rcx, _retval
		cmp		[rcx]. ABF._sign, 0
		cmove	rax, _one
		cmovne	rax, _zero
		mov		[rcx]. ABF._sign, rax
		cmp		_v2, 0
		je		Skip7
		mov		rcx, _v2
		call	DestroyBD

Skip7:

		cmp		_t3, 0
		je		Skip8
		mov		rcx, _t3
		call	DestroyBD

Skip8:

		cmp		_t2, 0
		je		Skip9
		mov		rcx, _t2
		call	DestroyBD

Skip9:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		RET

	BFRadATan64 ENDP

	QWRand64 PROC

		;this routine serves as the basis for all the other random routines so needs to be red_one correctly

		;Right now the routine simply gets to arrays and uses whatever happens to be in those memory locations
		;and does a few xors between them

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rcx, 24
		call	GetMem64
		push	rax
		mov		rcx, 24
		call	GetMem64
		mov		r8, rax
		pop		r9
		mov		rax, qword ptr[r8]
		xor		rax, qword ptr[r9+8]
		xor		rax, qword ptr[r8+16]
		xor		rax, qword ptr[r9+16]
		push	rax
		push	r8
		mov		rcx, r9
		call	Free64
		pop		rcx
		call	Free64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		pop		rax

		RET

	QWRand64  ENDP

	QWRandMax64 PROC _tmax:qword

		;This routine simply gets a random qword generated by the QWRand64 routine and divides it by _max and takes the remainder

		ENTERFUNC _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		call	QWRand64
		xor		rdx, rdx
		pop		rcx
		div		rcx
		mov		rax, rdx

		EXITFUNC _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	QWRandMax64  ENDP

	QWRandMinMax64 PROC _tmin:qword, _tmax:qword

		;this routine takes a number generated by QWRand64 divides it by _max then compares the remainder to _min if it is smaller it adds _min toit

		ENTERFUNC _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		call	QWRand64
		pop		rcx
		xor		rdx, rdx
		div		rcx
		mov		rax, rdx
		pop		rcx
		cmp		rax, rcx
		jnb		GoOut
		add		rax, rcx

GoOut:

		EXITFUNC _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	QWRandMinMax64  ENDP

	QWPtrRand64 PROC _bytes:qword

		;This routine takes numbers generated by QWRand64 and supplies them into the return array

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_t1, rcx
		mov		rax, rcx
		add		rax, 2
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_retval, rax
		mov		rcx, _t1
		mov		qword ptr[rax], rcx

MainLoop:

		cmp		_t1, 0
		je		GoOut
		call	QWRand64
		mov		rdi, _t1
		mov		r8, _retval
		mov		qword ptr[r8+rdi*8], rax
		dec		rdi
		mov		_t1, rdi
		jmp		MainLoop

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	QWPtrRand64  ENDP

	QWPtrRandMax64 PROC _tmax:PTR qword

		;This routine takes an array generated by the QWPtrRand64 divides it by max and returns the Remainder

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rcx, qword ptr[rcx]
		call	QWPtrRand64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v1
		call	AMod64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	QWPtrRandMax64  ENDP

	QWPtrRandMinMax64 PROC _tmin:PTR qword, _tmax:PTR qword

		;This routine takes an array generated by QWPtrRand64 divides it by _max, compares it to _min and if it is smaller adds _min to it

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, rdx
		call	QWPtrRandMax64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v1
		call	AGTOE64
		cmp		rax, 0
		ja		GoOut
		mov		rcx, _retval
		mov		rdx, _v1
		call	AAdd64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	QWPtrRandMinMax64 ENDP

	AGCD64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx

	; ABI^ _a;

		mov		rax, qword ptr[rcx]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, _v1
		mov		_var9, rax
		mov		rsi, r8
		mov		rdi, rax
		mov		rcx, qword ptr[r8]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _var9
		mov		_v1, rax

	; ABI^ _b;

		mov		r8, _v2
		mov		rax, qword ptr[r8]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, _v2
		mov		_var9, rax
		mov		rsi, r8
		mov		rdi, rax
		mov		rcx, qword ptr[r8]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _var9
		mov		_v2, rax

	;if(_val1>=_val2){

		;_a=gcnew  ABI(_val1);
		;_b=gcnew  ABI(_val2);

	;else{

		;_a=gcnew  ABI(_val2);
		;_b=gcnew  ABI(_val1);

		mov		rdx, rax
		mov		rcx, _v1
		call	AGTOE64
		mov		rcx, _v1
		mov		rdx, _v2
		cmp		rax, 1
		cmovne	rcx, _v2
		cmovne	rdx, _v1
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 24
		call	GetMem64
		mov		_retval, rax

MainLoop:

	;while(_b>0){

		mov		rdx, _v2
		cmp		qword ptr[rdx], 1
		ja		DoAlg
		cmp		qword ptr[rdx+8], 0
		je		GoOut

DoAlg:

	;_retval=_a%_b;
	;_a=_b;
	;_b=_retval;

		mov		rcx, _v1
		call	AMod64
		mov		rcx, _v1
		mov		rdx, _v2
		mov		_v1, rdx
		mov		_v2, rax
		call	Free64
		jmp		MainLoop

GoOut:

		mov		rcx, _v2
		call	Free64
		mov		rcx, _retval
		call	Free64

	;return _a;

		mov		rax, _v1

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	AGCD64  ENDP

	ALCM64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

	;_retvaletval=_val1*_val2;

		mov		_v1, rcx
		mov		_v2, rdx
		call	AMultiply64
		mov		_retval, rax

	;_retvaletval=_retvaletval/GCD(_val1, _val2);

		mov		rcx, _v1
		mov		rdx, _v2
		call	AGCD64
		push	rax
		mov		rdx, rax
		mov		rcx, _retval
		call	ADivide64
		pop		rcx
		push	rax
		call	Free64
		mov		rcx, _retval
		call	Free64

	;return _retvaletval;

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		pop		rax
		RET

	ALCM64 ENDP

	AMillerRabbin64 PROC _pn:PTR qword, _cert:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

	;isprime=true;
	;blnd_one=false;
	;found=false;

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var6, 1
		mov		_var7, 0
		mov		_var8, 0

	; ABI^ temp=gcnew  ABI(1);

		mov		rcx, 24
		call	GetMem64
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1

	; ABI^ primenumber=gcnew  ABI(_posibleprime);
	; ABI^ nminus_one=gcnew  ABI(1);

		mov		rax, _v1
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_var5, rax
		mov		r8, rax
		mov		r11, _v1
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf

	;nminus_one=primenumber-1;

		mov		rcx, _var5
		mov		rdx, _t2
		call	ASub64
		mov		rcx, _var5
		mov		_var5, rax
		call	Free64

	; ABI^ _var3=gcnew  ABI(MAX)

		mov		rcx, 32
		call	GetMem64
		mov		_var3, rax
		mov		qword ptr[rax], 2
		mov		qword ptr[rax+8], 0
		mov		qword ptr[rax+16], 1

	; ABI^ _var4=gcnew  ABI();

		mov		rcx, 32
		call	GetMem64
		mov		_var4, rax
		mov		qword ptr[rax], 2
		mov		qword ptr[rax+8], 0
		mov		qword ptr[rax+16], 1
		mov		rax, _var3
		mov		_t1, 1

FindRLoop:

	;while (blnd_one==false){

	;if (nminus_one%_var4==0){

		mov		rcx, _var5
		mov		rdx, _var4
		call	AMod64
		cmp		qword ptr[rax], 1
		jne		RNoGood
		cmp		qword ptr[rax+8], 0
		je		RGood

RNoGood:

		mov		rcx, rax
		call	Free64

NextExp:

	;s=s+1;
	;_var4= ABI::Pow(_var3, s);

		mov		rax, _t1
		inc		rax
		mov		_t1, rax
		mov		rcx, _var4
		mov		rdx, _var3
		call	AMultiply64
		mov		rcx, _var4
		mov		_var4, rax
		call	Free64

	;check _var5<_var4

		mov		rcx, _var5
		mov		rdx, _var4
		call	ALT64
		cmp		rax, 0
		je		ROut
		jmp		FindRLoop

RGood:

	;r=nminus_one/_var4;

		mov		rcx, rax
		call	Free64
		mov		rcx, _var5
		mov		rdx, _var4
		call	ADivide64
		mov		_retval, rax

	;if(r%2==1){

		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 1
		je		FoundR
		mov		rcx, _retval
		call	Free64
		jmp		NextExp

FoundR:

	;found=true;

		mov		_var8, 1
		
ROut:

	; ABI^ r=gcnew  ABI(1);

		cmp		_var8, 0
		jne		RAllGood
		mov		rcx, _var5
		mov		rdx, _t2
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 2
		call	ADivide64
		mov		_retval, rax
		mov		_t1, 1

RAllGood:

	; ABI^ s=gcnew  ABI();

		mov		rcx, 32
		call	GetMem64
		mov		_var1, rax
		mov		qword ptr[rax], 1
		mov		rdx, _t1
		mov		qword ptr[rax+8], rdx

	; ABI^ sminus_one=gcnew  ABI(1);
	;sminus_one=s-1;

		mov		rcx, 32
		call	GetMem64
		mov		_t3, rax
		mov		qword ptr[rax], 1
		mov		rdx, _var1
		mov		rdx, qword ptr[rdx+8]
		dec		rdx
		mov		qword ptr[rax+8], rdx

	; ABI^ y=gcnew  ABI(1);

		mov		rcx, 32
		call	GetMem64
		mov		_var2, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		_t1, 0

MainLoop:


	;for (i=1;i<=_cert;i++){

		mov		rax, _t1
		inc		rax
		cmp		rax, _v2
		jnb		GoOut
		mov		_t1, rax

	;y= ABI::Random(primenumber-1);

		mov		rcx, _var5
		call	QWPtrRandMax64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64

	;y= ABI::ModPow(y,r,primenumber);

		mov		rcx, _var2
		mov		rdx, _retval
		mov		r8, _v1
		call	AModPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64


	;if (y!=1 && y!=nminus_one){

		mov		rcx, _var2
		cmp		qword ptr[rcx], 1
		jne		NextCheck
		cmp		qword ptr[rcx+8], 1
		je		MainLoop

NextCheck:

		mov		rdx, _var5
		call	AE64
		cmp		rax, 1
		je		MainLoop
		mov		_var9, 0
		mov		rax, _t2
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 2
		mov		_var9, 0

InnerLoop:

	;while (j<=sminus_one && y!=nminus_one){

		mov		rax, _var9
		inc		rax
		mov		rdx, _t3
		cmp		qword ptr[rdx+8], rax
		ja		MainLoop

	;y= ABI::ModPow(y,2,primenumber);

		mov		rcx, _var2
		mov		rdx, _t2
		mov		r8, _v1
		call	AModPow64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64

	;if (y==1){

		mov		rcx, _var2
		cmp		qword ptr[rcx], 1
		jne		NeverMind
		cmp		qword ptr[rcx+8], 1
		jne		NeverMind

	;isprime=false;

		mov		_var6, 0
		jmp		GoOut

NeverMind:

		mov		rdx, _var5
		call	AE64
		cmp		rax, 1
		je		MainLoop
		jmp		InnerLoop

GoOut:

		mov		rcx, _v1
		call	Free64
		mov		rcx, _var1
		call	Free64
		mov		rcx, _retval
		call	Free64
		mov		rcx, _var2
		call	Free64
		mov		rcx, _var5
		call	Free64
		mov		rcx, _t3
		call	Free64
		mov		rcx, _var3
		call	Free64
		mov		rcx, _var4
		call	Free64
		mov		rcx, _t2
		call	Free64
		mov		rax, _var6

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	AMillerRabbin64 ENDP

	AEuclideanAlg64 PROC _val1:PTR qword, _val2:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

	; ABI^ _x=gcnew  ABI(_ta);
	; ABI^ _y=gcnew  ABI(_tb);
	; ABI^ _g=gcnew  ABI(1);

		mov		_t1, rcx
		mov		_t2, rdx
		mov		rax, qword ptr[rcx]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, rax
		mov		r11, _t1
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		_t1, r8
		mov		r8, _t2
		mov		rax, qword ptr[r8]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, rax
		mov		r11, _t2
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		_t2, r8
		mov		rcx, 32
		call	GetMem64
		mov		_t3, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1

	;while(_t1->IsEven && _t2->IsEven){

InitLoop:

		mov		r11, _t2
		mov		rax, qword ptr[r11+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		InitLoopOut
		mov		rcx, _t1
		mov		rax, qword ptr[rcx+8]
		div		_two
		cmp		rdx, 0
		jne		InitLoopOut

	;_t1=_t1/2;
	;_t2=_t2/2;
	;_t3=_t3*2;

		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	Free64
		mov		rcx, _t2
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t2
		mov		_t2, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t9
		call	AMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		jmp		InitLoop

	; ABI^ _t4=gcnew  ABI(_t1);
	; ABI^ _v1=gcnew  ABI(_t2);
	; ABI^ _t5=gcnew  ABI(1);
	; ABI^ _t6=gcnew  ABI();
	; ABI^ _t7=gcnew  ABI();
	; ABI^ _t8=gcnew  ABI(1);

InitLoopOut:

		mov		r8, _t1
		mov		rax, qword ptr[r8]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_t4, rax
		mov		r8, rax
		mov		r11, _t1
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r8, _t2
		mov		rax, qword ptr[r8]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_v1, rax
		mov		r8, rax
		mov		r11, _t2
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, 32
		call	GetMem64
		mov		_t5, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rcx, 32
		call	GetMem64
		mov		_t8, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rcx, 32
		call	GetMem64
		mov		_t6, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rcx, 32
		call	GetMem64
		mov		_t7, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

	;while(_t4->IsEven){

UEven:

		mov		rcx, _t4
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		VEven

	;_t4=_t4/2;

		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t4
		mov		_t4, rax
		call	Free64

	;if(_t5->IsEven && _t6->IsEven){

		mov		rcx, _t6
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		DoOtherWay1
		mov		rcx, _t5
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		DoOtherWay1

	;_t5=_t5/2;
	;_t6=_t6/2;

		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		mov		rcx, _t6
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64

DoOtherWay1:

	;_t5=(_t5+_t2)/2;

		mov		rax, _v2
		cmp		rax, _var9
		jne		SignsNotEqual1
		mov		rcx, _t5
		mov		rdx, _t2
		call	AAdd64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		mov		rcx, _t5
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		jmp		NextVal1

SignsNotEqual1:

		mov		rcx, _t5
		mov		rdx, _t2
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValBigger1
		mov		rcx, _t5
		mov		rdx, _t2
		call	ASub64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		mov		rcx, _t5
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		jmp		NextVal1

OtherValBigger1:

		cmp		_v2, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v2, rax
		mov		rcx, _t2
		mov		rdx, _t5
		call	ASub64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		mov		rcx, _t5
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
	
NextVal1:
			
	;_t6=(_t6-_t1)/2;

		mov		rax, _var9
		cmp		rax, _var2
		je		SignsEqual1
		mov		rcx, _t6
		mov		rdx, _t1
		call	AAdd64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		mov		rcx, _t6
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		UEven

SignsEqual1:

		mov		rcx, _t6
		mov		rdx, _t1
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValBigger2
		mov		rcx, _t6
		mov		rdx, _t1
		call	ASub64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		mov		rcx, _t6
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		UEven

OtherValBigger2:

		cmp		_var9, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var9, rax
		mov		rcx, _t1
		mov		rdx, _t6
		call	ASub64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		mov		rcx, _t6
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		UEven

	;while(_v1->IsEven){

VEven:

		mov		rcx, _v1
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		NextCheck

	;_v1=_v1/2;

		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _v1
		mov		_v1, rax
		call	Free64

	;if(_t7->IsEven && _t8->IsEven){

		mov		rcx, _t8
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		DoOtherWay2
		mov		rcx, _t7
		mov		rax, qword ptr[rcx+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		DoOtherWay2

	;_t7=_t7/2;
	;_t8=_t8/2;

		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		mov		rcx, _t8
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64

DoOtherWay2:

	;_t7=(_t7+_t2)/2;

		mov		rax, _var8
		cmp		rax, _var6
		jne		SignsNotEqual2
		mov		rcx, _t7
		mov		rdx, _t2
		call	AAdd64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		mov		rcx, _t7
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		jmp		NextVal2

SignsNotEqual2:

		mov		rcx, _t7
		mov		rdx, _t2
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValBigger3
		mov		rcx, _t7
		mov		rdx, _t2
		call	ASub64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		mov		rcx, _t7
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		jmp		NextVal2

OtherValBigger3:

		cmp		_var8, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var8, rax
		mov		rcx, _t2
		mov		rdx, _t7
		call	ASub64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		mov		rcx, _t7
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
	
NextVal2:
			
	;_t8=(_t8-_t1)/2;

		mov		rax, _var6
		cmp		rax, _var2
		je		SignsEqual2
		mov		rcx, _t6
		mov		rdx, _t1
		call	AAdd64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		mov		rcx, _t6
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		VEven

SignsEqual2:

		mov		rcx, _t8
		mov		rdx, _t1
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValBigger4
		mov		rcx, _t8
		mov		rdx, _t1
		call	ASub64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		mov		rcx, _t8
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		jmp		VEven

OtherValBigger4:

		cmp		_var6, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var6, rax
		mov		rcx, _t1
		mov		rdx, _t8
		call	ASub64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		mov		rcx, _t8
		mov		rdx, _t9
		call	ADivide64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		jmp		VEven

NextCheck:

	;if(_t4>=_v1){

		mov		rcx, _t4
		mov		rdx, _v1
		call	AGTOE64
		cmp		rax, 1
		jne		VIsBigger

	;_t4=_t4-_v1;

		mov		rax, _var4
		cmp		rax, _var3
		jne		SignsEqual3
		mov		rcx, _t4
		mov		rdx, _v1
		call	AAdd64
		mov		rcx, _t4
		mov		_t4, rax
		call	Free64
		jmp		NextVal3

SignsEqual3:

		mov		rcx, _t4
		mov		rdx, _v1
		call	ASub64
		mov		rcx, _t4
		mov		_t4, rax
		call	Free64

NextVal3:

	;_t5=_t5-_t7;

		mov		rax, _v2
		cmp		rax, _var8
		je		SignsEqual4
		mov		rcx, _t5
		mov		rdx, _t7
		call	AAdd64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		jmp		NextVal4

SignsEqual4:

		mov		rcx, _t5
		mov		rdx, _t7
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValBigger5
		mov		rcx, _t5
		mov		rdx, _t7
		call	ASub64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64
		jmp		NextVal4

OtherValBigger5:

		cmp		_v2, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_v2, rax
		mov		rcx, _t7
		mov		rdx, _t5
		call	ASub64
		mov		rcx, _t5
		mov		_t5, rax
		call	Free64

NextVal4:

	;_t6=_t6-_t8;

		mov		rax, _var9
		cmp		rax, _var6
		je		SignsEqual5
		mov		rcx, _t6
		mov		rdx, _t8
		call	AAdd64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		DoLastCheck

SignsEqual5:

		mov		rcx, _t6
		mov		rdx, _t8
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValIsBigger6
		mov		rcx, _t6
		mov		rdx, _t8
		call	ASub64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		DoLastCheck

OtherValIsBigger6:

		cmp		_var9, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var9, rax
		mov		rcx, _t8
		mov		rdx, _t6
		call	ASub64
		mov		rcx, _t6
		mov		_t6, rax
		call	Free64
		jmp		NextVal5

VIsBigger:

	;_v1=_v1-_t4;

		mov		rax, _var3
		cmp		rax, _var4
		je		SignsEqual6
		mov		rcx, _v1
		mov		rdx, _t4
		call	AAdd64
		mov		rcx, _v1
		mov		_v1, rax
		call	Free64
		jmp		NextVal5

SignsEqual6:

		mov		rcx, _v1
		mov		rdx, _t4
		call	ASub64
		mov		rcx, _v1
		mov		_v1, rax
		call	Free64

NextVal5:

	;_t7=_t7-_t5;

		mov		rax, _var8
		cmp		rax, _v2
		je		SignsEqual7
		mov		rcx, _t7
		mov		rdx, _t5
		call	AAdd64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		jmp		NextVal6

SignsEqual7:

		mov		rcx, _t7
		mov		rdx, _t5
		call	AGTOE64
		cmp		rax, 1
		je		OtherValBigger7
		mov		rcx, _t7
		mov		rdx, _t5
		call	ASub64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64
		jmp		NextVal6

OtherValBigger7:

		cmp		_var8, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var8, rax
		mov		rcx, _t5
		mov		rdx, _t7
		call	ASub64
		mov		rcx, _t7
		mov		_t7, rax
		call	Free64

NextVal6:

	;_t8=_t8-_t6;

		mov		rax, _var6
		cmp		rax, _var9
		je		SignsEqual8
		mov		rcx, _t8
		mov		rdx, _t6
		call	AAdd64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		jmp		DoLastCheck

SignsEqual8:

		mov		rcx, _t8
		mov		rdx, _t6
		call	AGTOE64
		cmp		rax, 1
		jne		OtherValIsBigger8
		mov		rcx, _t8
		mov		rdx, _t6
		call	ASub64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64
		jmp		DoLastCheck

OtherValIsBigger8:

		cmp		_var6, 1
		cmove	rax, _zero
		cmovne	rax, _one
		mov		_var6, rax
		mov		rcx, _t6
		mov		rdx, _t8
		call	ASub64
		mov		rcx, _t8
		mov		_t8, rax
		call	Free64

DoLastCheck:

	;if(_t4==0){

		mov		rcx, _t4
		cmp		qword ptr[rcx], 1
		jne		UEven
		cmp		qword ptr[rcx+8], 0
		jne		UEven

	;_retvaletval->_t1=_t7;
	;_retvaletval->_t2=_t8;
	;_retvaletval->_t8=_t3*_v1;

		mov		rcx, _t3
		mov		rdx, _v1
		call	AMultiply64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, 48
		call	GetMem64
		mov		qword ptr[rax], 3
		mov		rcx, _t7
		mov		qword ptr[rax+8], rcx
		mov		rcx, _t8
		mov		qword ptr[rax+16], rcx
		mov		rcx, _t3
		mov		qword ptr[rax+24], rcx

	;return _retvaletval;

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	AEuclideanAlg64 ENDP

	AConvertFromString64 PROC val:PTR byte, radix:qword
	
		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		cmp		_initialized, 0
		jne		AllreadyInitialized
		;call	AADoInit

AllreadyInitialized:

		mov		r9, 0
		cmp		radix, 2
		cmove	r9, _tobase2
		cmove	r10, _one
		cmp		radix, 8
		cmove	r9, _tobase8
		cmove	r10, _seven
		cmp		radix, 10
		cmove	r9, _tobase10
		cmove	r10, _nine
		cmp		radix, 16
		cmove	r9, _tobase16
		cmove	r10, _fifteen
		cmp		radix, 32
		cmove	r9, _tobase32
		cmove	r10, _thirtyone
		cmp		r9, 64
		cmove	r9, _tobase64
		cmove	r10, _sixtythree
		cmp		r9, 0
		je		ErrorOut
		mov		r11, rcx
		mov		rdi, 0
		mov		rax, 1
		xor		rdx, rdx

FindBaseLoop:

		mul		_two
		inc		rdi
		cmp		rdi, r10
		jna		FindBaseLoop
		cmp		rdx, 0
		cmova	rcx, _thirtytwo
		cmovna	rcx, _twentyfour
		push	r9
		push	r10
		push	r11
		push	rax
		call	GetMem64
		mov		r12, rax
		pop		rax
		pop		r11
		pop		r10
		pop		r9
		mov		qword ptr[r12], 1
		mov		qword ptr[r12+8], rax
		cmp		rax, 0
		je		KeepIt
		mov	qword ptr[r12], 2
		mov	qword ptr[r12+16], 1

KeepIt:

		push	r9
		push	r10
		push	r11
		push	r12
		mov		rcx, 24
		call	GetMem64
		mov		r13, rax
		mov		qword ptr[r13], 1
		mov		qword ptr[r13+8], 0
		push	r13
		mov		rcx, 24
		call	GetMem64
		mov		r8, rax
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		pop		r13
		pop		r12
		pop		r11
		pop		r10
		pop		r9
		mov		rdi, 0
		xor		rax, rax

MainLoop:

		push	r10							;r10
		push	r12							;r12, r10
		push	r13							;r13, r12, r10
		push	r8							;r8, r13, r12, r10
		push	r9							;r9, r8, r13, r12, r10
		push	r11							;r11, r9, r8, r13, r12, r10
		push	rdi							;rdi, r11, r9, r8, r13, r12, r10
		mov		qword ptr[r13+8], rdi
		mov		rdx, r13
		mov		rcx, r12
		call	APow64
		pop		rdi							;r11, r9, r8, r13, r12, r10
		pop		r11							;r9, r8, r13, r12, r10
		pop		r9							;r8, r13, r12, r10
		pop		r8							;r13, r12, r10
		pop		r13							;r12, r10
		push	r13							;r13, r12, r10
		push	r9							;r9, r13, r12, r10
		push	r11							;r11, r9, r13, r12, r10
		push	rdi							;rdi, r11, r9, r13, r12, r10
		push	rax							;pow, rdi, r11, r9, r13, r12, r10
		push	r8							;r8, pow, rdi, r11, r9, r13, r12, r10
		mov		rcx, rax
		xor		rax, rax
		mov		al, byte ptr[r11+rdi]
		mov		rsi, 0

FindValLoop:

		cmp		rax, qword ptr[r9+rsi*8]
		je		FoundIt
		inc		rsi
		cmp		rsi, r10
		jna		FindValLoop

FoundIt:

		mov		qword ptr[r13+8], rsi
		mov		rdx, r13
		call	AMultiply64
		mov		rcx, rax
		pop		r8							;pow, rdi, r11, r9, r13, r12, r10
		mov		rdx, r8
		push	rax							;mul, pow, rdi, r11, r9, r13, r12, r10
		push	r8							;r8, mul, pow, rdi, r11, r9, r13, r12, r10
		call	AAdd64
		mov		r8, rax
		pop		rcx							;mul, pow, rdi, r11, r9, r13, r12, r10
		push	r8							;r8, mul, pow, rdi, r11, r9, r13, r12, r10
		call	Free64
		pop		r8							;mul, pow, rdi, r11, r9, r13, r12, r10
		pop		rcx							;pow, rdi, r11, r9, r13, r12, r10
		push	r8							;r8, pow, rdi, r11, r9, r13, r12, r10
		call	Free64
		pop		r8							;pow, rdi, r11, r9, r13, r12, r10
		pop		rcx							;rdi, r11, r9, r13, r12, r10
		push	r8							;r8, rdi, r11, r9, r13, r12, r10
		call	Free64
		pop		r8							;rdi, r11, r9, r13, r12, r10
		pop		rdi							;r11, r9, r13, r12, r10
		pop		r11							;r9, r13, r12, r10
		pop		r9							;r13, r12, r10
		pop		r13							;r12, r10
		pop		r12							;r10
		pop		r10							;
		inc		rdi
		mov		rax, rdi
		cmp		al, byte ptr[r11+rdi]
		jna		MainLoop
		push	r8
		push	r12
		mov		rcx, r13
		call	Free64
		pop		rcx
		call	Free64

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		RET

ErrorOut:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero

		RET
		
	AConvertFromString64 ENDP

	AConvertToString64	PROC val:PTR QWORD, radix:qword
		
		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero

		cmp		_initialized, 0
		jne		AllreadyInitialized
		;call	AADoInit

AllreadyInitialized:

		mov		r11, rcx
		mov		r15, rdx
		mov		rax, qword ptr[r11]
		add		rax, 2
		mul		_eight
		push	r11
		push	r15
		mov		rcx, rax
		call	GetMem64
		mov		r8, rax
		pop		r15
		pop		r11
		mov		rsi, r11
		mov		rdi, r8
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r11, r8
		mov		_v1, r8
		mov		rcx, 24
		push	r11
		push	r15
		call	GetMem64
		mov		r9, rax
		mov		_t2,rax
		pop		r15
		pop		r11
		mov		qword ptr[r9], 1
		mov		rax, 1
		xor		rdi, rdi
		dec		r15
		mov		_var1, r15

FindBaseLoop:

		mul		_two
		inc		rdi
		cmp		rdi, r15
		jna		FindBaseLoop
		mov		qword ptr[r9+8], rax
		mov		qword ptr[r9+16], rdx
		mov		rcx, 24
		push	r15
		push	r11
		push	r9
		call	GetMem64
		mov		r10, rax
		mov		_t3, rax
		mov		qword ptr[r10], 1
		push	r10
		xor		rdi, rdi

FindLengthLoop:

		mov		rcx, _t2
		mov		rdx, _t3
		push	rdi
		mov		qword ptr[rdx+8], rdi
		call	APow64
		mov		rcx, rax
		mov		rdx, _v1
		call	AGT64
		pop		rdi
		cmp		rax, 1
		je		FoundLength
		inc		rdi
		jmp		FindLengthLoop


FoundLength:

		inc		rdi
		mov		rcx, rdi
		add		rcx, 10
		push	rdi
		call	GetMem64
		mov		r8, rax
		mov		_retval, rax
		pop		rax
		mov		byte ptr[r8], al
		dec		rax
		pop		r10
		pop		r9
		pop		r11
		pop		r15
		push	rax

MainLoop:

		pop		rdi
		mov		rcx, _t2
		mov		rdx, _t3
		mov		qword ptr[rdx+8], rdi
		push	rdi
		call	APow64
		mov		_var2, rax
		mov		rsi, _var1
		mov		_t1, rsi

FindMultLoop:

		mov		rcx, _var2
		mov		rdx, _t3
		mov		rsi, _t1
		mov		qword ptr[rdx+8], rsi
		call	AMultiply64
		mov		_var3, rax
		mov		rcx, rax
		mov		rdx, _v1
		call	ALTOE64
		cmp		rax, 1
		je		FoundMult
		mov		rsi, _t1
		cmp		rsi, 0
		je		FoundMult
		dec		rsi
		mov		_t1, rsi
		jmp		FindMultLoop

FoundMult:

		mov		r8, _retval
		pop		rdi
		push	rdi
		inc		rdi
		mov		rax, _t1
		mov		byte ptr[r8+rdi], al
		mov		rcx, _v1
		mov		rdx, _var3
		call	ASub64
		mov		rcx, _v1
		mov		_v1, rax
		call	Free64
		mov		rcx, _var2
		call	Free64
		mov		rcx, _var3
		call	Free64
		pop		rdi
		dec		rdi
		cmp		rdi, 0
		je		GoOut
		dec		rdi
		push	rdi
		jmp		MainLoop

GoOut:

		mov		r8, _retval
		xor		rdx, rdx
		xor		rdi, rdi
		mov		dl, byte ptr[r8]
		mov		rdi, rdx


CheckZerosLoop:

		cmp		rdi, 0
		jna		OutOfHere
		cmp		byte ptr[r8+rdi], 0
		ja		OutOfhere
		dec		rdi
		mov		rdx, rdi
		mov		byte ptr[r8], dl
		jmp		CheckZerosLoop

MallocError:

		;if there was an error requesting memory try to get the error code, store it in the return value and send it back to managed code

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		OutOfHere
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

OutOfHere:

		mov		rax, r8

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero

		cld
		RET
		
	AConvertToString64 ENDP

	END