;;
;@brief This file contains the basic math functions for the  ABI  ABD  ABF  ABDI and  ABFI classes
;
;  These algs are the ones to call from native c++, all the classes here are simply prepended with a C to represent the native versions and can take those that were created in native c++ just as easily
;as those created in an assembly alg, and native c++ can handle the versions of the class created in the assembly routines just fine.  If you try inhereting the classes and you implement a destructor,
;the arrays need to be freed using the internal memory manager.  The class name is prepended to all the math alg, but it is appended to the create and destroy routines.
;;

.list
include GlobalStuff64.inc
include AdvancedAlgs64.inc

	.data
	
			_howtohandlerounding	qword		0
			_decimallimit			qword		50
			_radix					qword		0
			_workingdecimallimit	qword		60

			_suboffset				qword		0

			_sub			qword		0
			_add			qword		0
			_norm			qword		0
			_red			qword		0

	.code
	;;
	;@brief Creates a  ABI object equal to 0 
	;@returns a  ABI object
	;;
	CreateBI PROC

		ENTERFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		rcx, 24
		call	malloc
		mov		_retval, rax
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABI._num, rax
		mov		[rcx]. ABI._sign, 0
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rax, _retval
		
		EXITFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBI endp
	;;
	;@brief Creates a  ABI object given an array for the value.
	;@returns a  ABI object
	;@param rcx - pointer to the array.
	;;
	CreateBICopy PROC _val:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 24
		call	malloc
		mov		_retval, rax
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABI._num, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval
		
		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBICopy endp
	;;
	;@brief Creates a  ABI object given an array for the value (takes posession of the array.
	;@returns a  ABI object
	;@param rcx - pointer to the array.
	;;
	CreateBITake PROC _val:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 24
		call	malloc
		mov		_retval, rax
		mov		rcx, _v1
		mov		[rax]. ABI._num, rcx
		mov		[rax]. ABI._sign, 0
		
		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBITake endp
	;;
	;@brief Creates a  ABI object copying the values of another  ABI object.
	;@returns a  ABI object
	;@param rcx - pointer to the  ABI object to copy.
	;;
	CreateBIFromBI PROC _val: ABI

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 24
		call	malloc
		mov		_retval, rax
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABI._num, rax
		mov		rsi, _v1
		mov		rdx, [rsi]. ABI._sign
		mov		[rcx]. ABI._sign, rdx
		mov		rdi, rax
		mov		rsi, [rsi]. ABI._num
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval
		
		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBIFromBI endp
	;;
	;@brief Creates a  ABI object given an array for the value and a qword for the sign value.
	;@returns a  ABI object
	;@param rcx - pointer to the array.
	;@param rdx - value to be used for the sign.
	;;
	CreateBICopyWithSign PROC _val:PTR qword, _si:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 24
		call	malloc
		mov		_retval, rax
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABI._num, rax
		mov		rdx, _v2
		mov		[rcx]. ABI._sign, rdx
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval
		
		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBICopyWithSign endp
	;;
	;@brief Destroys a  ABI object.
	;@param rcx - pointer to the  ABI object to be destroyed.
	;;
	DestroyBI PROC _val: ABI

		ENTERFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, [rcx]. ABI._num
		call	Free64
		mov		rcx, _v1
		call	free
		
		EXITFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	DestroyBI endp

	;the equality algorithms work under the assumption that all may be worked from implementing just 2 of them to make it easier to maintain
	;so we write out algorithms for the Equals and the LessThan all other algs are worked through these
	;AE uses the Equals, ANE uses equals then nots the result, LT uses LessThan, GT uses LessThan with the inputs reversed, 
	;LTOE uses the LessThan with the inputs reversed and not the result and GTOE uses the LessThan and not the result

	;;
	;@brief Checks for equality between 2  ABI objects.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIE64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		r11, [rcx]. ABI._sign
		cmp		r11, [rdx]. ABI._sign
		je		SignsEQ
		mov		rax, 0

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

SignsEQ:

		mov		r11, [rcx]. ABI._num
		mov		r12, [rdx]. ABI._num
		mov		rcx, qword ptr[r11]
		inc		rcx
		mov		rdi, r11
		mov		rsi, r12
		pushf
		cld
repe	cmpsq
		jne		eNotEqual
		popf
		mov		rax, 1
		jmp		eGoOut
		
eNotEqual:

		popf
		mov		rax, 0

eGoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BIE64 ENDP
	;;
	;@brief Checks for inequality between 2  ABI objects.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BINE64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		IE64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BINE64 ENDP	
	;;
	;@brief Checks if the first value is less than the second value.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BILT64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		r11, [rcx]. ABI._sign
		cmp		r11, [rdx]. ABI._sign
		ja		ltLessThan
		jb		ltNotLessThan
		mov		_t1, r11
		mov		r11, [rcx]. ABI._num
		mov		r12, [rdx]. ABI._num
		mov		rcx, qword ptr[r11]
		cmp		rcx, qword ptr[r12]
		ja		ltNotLessThan
		jb		ltLessThan
		mov		rax, rcx
		mul		_eight
		mov		rsi, r11
		add		rsi, rax
		mov		rdi, r12
		add		rdi, rax
		pushf
		std
repe	cmpsq
		jb		PrepItsLess
		popf
		mov		rax, 0
		jmp		ltGoOut

ltNotLessThan:

		mov		rax, 0
		jmp		ltGoOut

PrepItsLess:

		popf

ltLessThan:

		mov		rax, 1

ltGoOut:

		cmp		_t1, 0
		je		LeaveIt
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one

LeaveIt:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET
			
	BILT64 ENDP	
	;;
	;@brief Checks if the first value is less than or equal to the second value.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BILTOE64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		r11, [rcx]. ABI._sign
		cmp		r11, [rdx]. ABI._sign
		ja		ItsLessThan
		jb		ItsNotLessThan
		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		ILT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		jmp		GoOut

ItsLessThan:

		mov		rax, 1
		jmp		GoOut

ItsNotLessThan:

		mov		rax, 0

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET
		
	BILTOE64 ENDP
	;;
	;@brief Checks if the first value is greater than the second value.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIGT64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		ILT64		

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BIGT64 ENDP
	;;
	;@brief Checks if the first value is greater than or equal to the second value.
	;@returns a bool.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIGTOE64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		ILT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
gtoeAllreadySet:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BIGTOE64 ENDP
	;;
	;@brief Adds 2  ABI objects together.
	;
	;This function starts by comparing the _signs of the 2 instances.  If they are equal, it send the 2 arrays to the add function in Asm64.  If they are not equal, it compares
	;the 2 arrays to find which one has the higher absolute value, sets that instance to _val1, the other to _val2, and sends the arrays to the subtract function in Asm64 and sets the _sign accordingly.
	;@returns a  ABI object.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIAdd64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		cmp		r11, [rdx]. ABI._sign
		jne		NotEqual
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Add64

		mov		rcx, rax
		call	CreateBITake
		mov		rdx, _sgn
		mov		[rax]. ABI._sign, rdx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

PrepLarger:

		popf
		mov		rcx, _v1
		mov		rdx, _v2
		jmp		Larger

PrepSmaller:

		popf
		mov		rcx, _v1
		mov		rdx, _v2
		jmp		Smaller

NotEqual:

		mov		r11, [rcx]. ABI._num
		mov		r12, [rdx]. ABI._num
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		Larger
		jb		Smaller
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rdi, r12
		mov		rsi, r11
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		ja		PrepLarger
		jb		PrepSmaller
		popf
		call	CreateBI

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Larger:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		rbx, [rcx]. ABI._sign
		mov		_sgn, rbx
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Sub64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Smaller:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		rbx, [rdx]. ABI._sign
		mov		_sgn, rbx
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rax, [rcx]. ABI._num
		mov		rcx, [rdx]. ABI._num
		mov		rdx, rax
		Sub64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BIAdd64 ENDP 
	;;
	;@brief Subtracts 2  ABI objects together.
	;
	;This function starts by comparing the _signs of the 2 instances, if they DO NOT equal, the 2 arrays are sent to the add function in Asm64.  If they do equal
	;the 2 arrays are compared to find the one with the higher absolute value, that value is set to _val1, the other to _val2 and they are sent to the subtract routine in Asm64 and the _sign set accordingly.
	;@returns a  ABI object.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BISub64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		cmp		r11, [rdx]. ABI._sign
		je		TheyEqual
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Add64

		mov		rcx, rax
		call	CreateBITake
		mov		rdx, _sgn
		mov		[rax]. ABI._sign, rdx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

PrepLarger:

		popf
		mov		rcx, _v1
		mov		rdx, _v2
		jmp		Larger

PrepSmaller:

		popf
		mov		rcx, _v1
		mov		rdx, _v2
		jmp		Smaller

TheyEqual:

		mov		r11, [rcx]. ABI._num
		mov		r12, [rdx]. ABI._num
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		Larger
		jb		Smaller
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rdi, r12
		mov		rsi, r11
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		ja		PrepLarger
		jb		PrepSmaller
		popf
		call	CreateBI
		mov		rcx, [rax]. ABI._num
		mov		qword ptr[rcx], 1
		mov		qword ptr[rcx+8], 0

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Larger:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		rbx, [rcx]. ABI._sign
		mov		_sgn, rbx
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Sub64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Smaller:

		mov		rcx, _v1
		mov		rdx, _v2
		cmp		[rdx]. ABI._sign, 0
		cmove	rbx, _one
		cmovne	rbx,_zero
		mov		_sgn, rbx
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rax, [rcx]. ABI._num
		mov		rcx, [rdx]. ABI._num
		mov		rdx, rax
		Sub64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BISub64 ENDP
	;;
	;@brief Multiplies 2  ABI objects together.
	;
	;This function simply sends the 2 arrays to the multiply function in Asm64 and sets the _sign to the exclusive or of the 2 instances _signs.
	;@returns a  ABI object.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIMultiply64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABI._sign
		xor		r11, [rdx]. ABI._sign
		mov		_sgn, r11
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Multiply64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIMultiply64 ENDP
	;;
	;@brief Squares a  ABI object.
	;
	;This function just send the array to the Square function in the Asm64 and sets the _sign to 0.
	;@returns a  ABI object.
	;@param rcx -  ABI object to be squared.
	;;
	BISquare64 PROC _val1: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rcx, [rcx]. ABI._num
		Square64

		mov		rcx, rax
		call	CreateBITake
		mov		[rax]. ABI._sign, 0

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BISquare64 ENDP
	;;
	;@brief Divides one  ABI object from another.
	;
	;This function sends the 2 arrays to the Divide function in Asm64 and sets the _sign to the exclusive or of the 2 instnaces _sign.
	;@returns a  ABI object.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIDivide64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABI._sign
		xor		r11, [rdx]. ABI._sign
		mov		_sgn, r11
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Divide64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIDivide64 ENDP
	;;
	;@brief Modular reduction of a  ABI object.
	;
	;This function sends the 2 arrays to the Mod function in the Asm64 file and sets the _sign equal to the _sign in _val1
	;@returns a  ABI object.
	;@param rcx - first  ABI object.
	;@param rdx - Second  ABI oobject.
	;;
	BIMod64 PROC _val1: ABI, _val2: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [rdx]. ABI._num
		Mod64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIMod64 ENDP
	;;
	;@brief Exponentiation of a  ABI object 2  ABI objects together.
	;
	;This function send the 2 arrays to the Pow function in the Asm64 then checks the _sign of val1.  If its negative, it checks if the Exponent is even or odd and sets the _sign accordingly.
	;@returns a  ABI object.
	;@param rcx - Base  ABI object.
	;@param rdx - Exponent  ABI oobject.
	;;
	BIPow64 PROC _base: ABI, _exp: ABI

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r12, [rdx]. ABI._num
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		cmp		rax, 0
		je		LeaveIt
		mov		rax, qword ptr[r12+8]
		mov		r12, rdx
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [r12]. ABI._num
		Pow64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET
	
	BIPow64	ENDP
	;;
	;@brief Modular exponentiation of a  ABI object.
	;
	;This function sends the 3 arrays to the ModPow function in Asm64, then checks the _sign of _val1, if its negative it checks if the exponent is even or odd and sets the _sign accordingly.
	;@returns a  ABI object.
	;@param rcx - Base  ABI object.
	;@param rdx - Exponent  ABI oobject.
	;@param r8 - Modulus  ABI oobject.
	;;
	BIModPow64 PROC _base: ABI, _exp: ABI, _mod: ABI	

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_v3, r8
		mov		r8, _v3
		mov		r12, [rdx]. ABI._num
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		cmp		rax, 0
		je		LeaveIt
		mov		rax, qword ptr[r12+8]
		mov		r12, rdx
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		r8, _v3
		mov		rcx, [rcx]. ABI._num
		mov		rdx, [r12]. ABI._num
		mov		r8, [r8]. ABI._num
		ModPow64

		mov		rcx, rax
		call	CreateBITake
		mov		rbx, _sgn
		mov		[rax]. ABI._sign, rbx

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET	

	BIModPow64	ENDP
	;;
	;@brief Bit shift left of a  ABI object.
	;@returns a  ABI object.
	;@param rcx -  ABI object.
	;@param rdx - amount to shift  ABI oobject.
	;;
	BIBitShiftLeft64 PROC _val: ABI, _count:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r12, rdx
		mov		_v1, rcx
		mov		_t3, 0
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		mov		r11, [rcx]. ABI._num
		mov		rax, r12
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		cmp		rdx, 0
		ja		DoBits
		mov		rdx, rax
		call	BIByteShiftLeft64
		mov		_retval, rax
		mov		rcx, _sgn
		mov		[rax]. ABI._sign, rcx

NotCreated2:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoBits:

		call	CreateBI
		mov		_retval, rax
		mov		rcx, _sgn
		mov		[rax]. ABI._sign, rcx
		mov		_t2, r12
		add		rax, qword ptr[r11]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r8, _retval
		mov		rcx, [r8]. ABI._num
		mov		[r8]. ABI._num, rax
		call	Free64
		mov		r8, [r8]. ABI._num
		mov		rax, qword ptr[r11]
		add		rax, _t1
		inc		rax
		mov		qword ptr[r8], rax
		mov		rdi, 1

ZeroOutReturn:

		mov		qword ptr[r8+rdi*8], 0
		inc		rdi
		cmp		rdi, qword ptr[r8]
		jna		ZeroOutReturn
		mov		rdi, 1
		mov		rsi, 1
		add		rsi, _t1
		mov		rbx, 64
		sub		rbx, _t2
		mov		rcx, _t2

MainLoop:

		mov		rax, qword ptr[r11+rdi*8]
		shl		rax, cl
		add		qword ptr[r8+rsi*8], rax
		mov		rax, qword ptr[r11+rdi*8]
		xchg	rbx, rcx
		shr		rax, cl
		inc		rdi
		inc		rsi
		mov		qword ptr[r8+rsi*8], rax
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r11]
		jna		MainLoop
		mov		rdi, qword ptr[r8]

CheckForZerosLoop:

		cmp		qword ptr[r8+rdi*8], 0
		ja		GoOut
		cmp		rdi, 1
		jna		GoOut
		dec		rdi
		mov		qword ptr[r8], rdi
		jmp		CheckForZerosLoop

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BIBitShiftLeft64 ENDP
	;;
	;@brief Bit shift right of a  ABI object.
	;@returns a  ABI object.
	;@param rcx -  ABI object.
	;@param rdx - amount to shift  ABI oobject.
	;;
	BIBitShiftRight64 PROC _val: ABI, _shift:qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r12, rdx
		mov		_v1, rcx
		mov		_t3, 0
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		mov		r11, [rcx]. ABI._num
		call	CreateBI
		mov		_retval, rax
		mov		rdx, r12
		mov		rax, rdx
		xor		rdx, rdx
		div		_sixtyfour
		mov		_t1, rax
		mov		_t2, rdx
		cmp		rax, qword ptr[r11]
		jnb		ReturnZero
		mov		rax, qword ptr[r11]
		sub		rax, _t1
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, _retval
		mov		rcx, [r8]. ABI._num
		mov		[r8]. ABI._num, rax
		call	Free64
		mov		rax, _sgn
		mov		[r8]. ABI._sign, rax
		mov		r8, [r8]. ABI._num
		mov		rbx, 1
		mov		rsi, qword ptr[r11]
		mov		rdi, rsi
		sub		rdi, _t1
		mov		qword ptr[r8], rdi
		mov		rsi, _t1
		mov		rdi, 1

CopyVal1Loop:

		mov		rax, qword ptr[r11+rsi*8]
		mov		qword ptr[r8+rdi*8], rax
		inc		rdi
		inc		rsi
		cmp		rsi, qword ptr[r11]
		jna		CopyVal1Loop
		cmp		_t2, 0
		je		GoOut
		mov		rbx, 64
		mov		rcx, _t2
		sub		rbx, rcx
		mov		rsi, 1
		mov		rdi, 2

ShiftLoop:

		shr		qword ptr[r8+rsi*8], cl
		xchg	rbx, rcx
		cmp		rdi, qword ptr[r8]
		ja		GoOut
		mov		rax, qword ptr[r8+rdi*8]
		shl		rax, cl
		add		qword ptr[r8+rsi*8], rax
		xchg	rbx, rcx
		inc		rdi
		inc		rsi
		jmp		ShiftLoop

ReturnZero:

		call	CreateBI
		mov		_retval, rax
		mov		rcx, 32
		call	GetMem64
		mov		r8, _retval
		mov		rcx, [r8]. ABI._num
		mov		[r8]. ABI._num, rax
		call	Free64
		mov		r8, [r8]. ABI._num
		mov		qword ptr[r8], 1
		mov		qword ptr[r8+8], 0
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BIBitShiftRight64 ENDP
	;;
	;@brief Byte shift left of a  ABI object.
	;@returns a  ABI object.
	;@param rcx -  ABI object.
	;@param rdx - amount to shift  ABI oobject.
	;;
	BIByteShiftLeft64 PROC _val: ABI, _shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rdx
		mov		_v1, rcx
		mov		_t2, 0
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		mov		r11, [rcx]. ABI._num
		mov		rdx, _t1
		call	CreateBI
		mov		_retval, rax
		mov		rax, qword ptr[r11]
		add		rax, _t1
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, _retval
		mov		rcx, [r8]. ABI._num
		mov		[r8]. ABI._num, rax
		call	Free64
		mov		rax, _sgn
		mov		[r8]. ABI._sign, rax
		mov		r8, [r8]. ABI._num
		mov		rax, qword ptr[r11]
		mov		rdi, rax
		add		rax, _t1
		mov		qword ptr[r8], rax
		mov		rsi, rax

CopyVal1Loop:

		mov		rax, qword ptr[r11+rdi*8]
		mov		qword ptr[r8+rsi*8], rax
		dec		rsi
		dec		rdi
		cmp		rdi, 0
		ja		CopyVal1Loop
		cmp		rsi, 0
		je		GoOut

ZeroLoop:

		mov		qword ptr[r8+rsi*8], 0
		dec		rsi
		cmp		rsi, 0
		ja		ZeroLoop
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIByteShiftLeft64 ENDP
	;;
	;@brief Byte shift right of a  ABI object.
	;@returns a  ABI object.
	;@param rcx -  ABI object.
	;@param rdx - amount to shift  ABI oobject.
	;;
	BIByteShiftRight64 PROC _val: ABI, _shift:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rdx
		mov		_t2, 0
		mov		_v1, rcx
		mov		r11, [rcx]. ABI._sign
		mov		_sgn, r11
		mov		r11, [rcx]. ABI._num
		mov		rdx, _t1
		call	CreateBI
		mov		_retval, rax
		mov		rax, qword ptr[r11]
		cmp		rax, _t1
		jna		ReturnZero
		sub		rax, _t1
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r8, _retval
		mov		rcx, [r8]. ABI._num
		mov		[r8]. ABI._num, rax
		call	Free64
		mov		rax, _sgn
		mov		[r8]. ABI._sign, rax
		mov		r8, [r8]. ABI._num
		cmp		r8, 0
		je		MallocError
		mov		rax, qword ptr[r11]
		sub		rax, _t1
		mov		qword ptr[r8], rax
		mov		rsi, 1
		mov		rdi, _t1
		inc		rdi

CopyVal1Loop:

		mov		rax, qword ptr[r11+rdi*8]
		mov		qword ptr[r8+rsi*8], rax
		inc		rsi
		inc		rdi
		cmp		rdi, qword ptr[r11]
		jna		CopyVal1Loop
		jmp		GoOut

ReturnZero:

		call	CreateBI
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BIByteShiftRight64 ENDP
	;;
	;@brief To string routine of a  ABI object.
	;
	;This function sends the array and radix to the ToString function in Asm64 then appends a - to the begginging of the returned value if the instance is negative.
	;@returns a char array.
	;@param rcx -  ABI object.
	;@param rdx - radix to use.
	;;
	BIToString64 PROC _val: ABI, _trad:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, [rcx]. ABI._num
		call	AToString64
		mov		_retval, rax
		mov		_t1, rcx
		mov		rcx, _v1
		cmp		[rcx]. ABI._sign, 0
		je		GoOut
		mov		rcx, _t1
		add		rcx, 10
		call	GetMem64
		mov		_v2, rax
		mov		rax, 0
		mov		rcx, _t1
		mov		rdi, _v2
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _v2
		mov		rsi, _retval
		mov		byte ptr[rdi], '-'
		inc		rdi
		mov		rcx, _t1
		pushf
		cld
rep		movsb
		popf
		mov		rcx, _retval
		mov		rax, _v2
		mov		_retval, rax
		call	Free64

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BIToString64 ENDP



















	;;
	;@brief Creates a  ABD object equal to zero.
	;@returns a  ABD object.
	;;
	CreateBD PROC

		ENTERFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._adj, 0
		mov		rcx, 16
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABD._num, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rax, _retval

		EXITFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBD ENDP
	;;
	;@brief Creates a  ABD object copying from a  ABD object.
	;@returns a  ABD object.
	;@param rcx -  ABD object to copy.
	;;
	CreateBDFromBD PROC _val: ABD

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABD._sign
		mov		[rax]. ABD._sign, rdx
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rdx
		mov		rdx, [rcx]. ABD._adj
		mov		[rax]. ABD._adj, rdx
		mov		rax, [rcx]. ABD._num
		mov		rax, qword ptr[rax]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABD._num, rax
		mov		rdi, rax
		mov		rcx, _v1
		mov		rsi, [rcx]. ABD._num
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval
				
		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDFromBD ENDP
	;;
	;@brief Creates a  ABD object equal to the value of the array passed in.
	;@returns a  ABD object.
	;@param rcx - array to copy.
	;;
	CreateBDCopy PROC _val:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._adj, 0
		mov		rax, _v1

		mov		rax, qword ptr[rax]
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABD._num, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDCopy ENDP
	;;
	;@brief Creates a  ABD object takes possession of the array passed in.
	;@returns a  ABD object.
	;@param rcx - array to take possession of.
	;;
	CreateBDTake PROC _val:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._adj, 0
		mov		rcx, _v1
		mov		[rax]. ABD._num, rcx
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDTake ENDP
	;;
	;@brief Destroys a  ABD object.
	;@param rcx -  ABD object to delete.
	;;
	DestroyBD PROC _val: ABD

		ENTERFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rcx, [rcx]. ABD._num
		call	Free64
		mov		rcx, _v1
		call	free

		EXITFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	DestroyBD ENDP
	;;
	;@brief Aligns the decimal places for the functions that need them aligned.
	;
	;This function compares the _decimalplace variable of the 2 instances, if they are equal it just returns the same variables in the rcx and rdx registers.
	;If they are not equal, the function will multiply the instance with a smaller decimalplace by 10^(the difference of the larger decimalplace from the smaller decimalplace) thereby aligning them and
	;placing the value in a new instance of  ABD.  This function allways returns 2  ABD objects in rcx and rdx.  At least one of the instances will be the same as what was passed in.
	;If the value in rcx was changed the new instance will be returned in rcx, likewise if the value in rdx was changed.
	;@returns two  ABD objects one in rcx representing the value passed in rcx and one in rdx representing the value passed in rdx.
	;@param rcx -  ABD object.
	;@param rdx -  ABD object.
	;;
	BDAlignDecimalPlaces PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _one, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		mov		_t1, 0
		mov		_v3, 0
		mov		_t2, 0
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._decimalplace
		cmp		rax, [r12]. ABD._decimalplace
		je		JustGoOut
		mov		rdi, 2
		mov		rax, [r12]. ABD._sign
		mov		_sgn, rax
		mov		rsi, rax
		mov		rax, [r11]. ABD._decimalplace
		mov		rbx, [r12]. ABD._decimalplace
		mov		rcx, [r11]. ABD._num
		mov		r14, [r12]. ABD._num
		mov		r13, 0
		cmp		rax, rbx
		cmovna	rsi, [r11]. ABD._sign
		cmovna	rax, [r12]. ABD._decimalplace
		cmovna	rbx, [r11]. ABD._decimalplace
		cmovna	rcx, [r12]. ABD._num
		cmovna	r14, [r11]. ABD._num
		cmovna	r13, _one
		cmovna	rdi, _one
		mov		_sgn, rsi
		mov		_t3, rdi
		mov		r15, r14
		mov		_t1, r13
		mov		_dp, rax
		sub		rax, rbx
		cmp		rax, 0
		je		JustGoOut
		mov		_v3, rax
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r13, rax
		mov		qword ptr[r13], 1
		mov		rax, _v3
		mov		qword ptr[r13+8], rax
		mov		_v3, r13
		mov		rcx, _bdten
		mov		rcx, [rcx]. ABD._num
		mov		rdx, r13
		Pow64	
		mov		rcx, rax
		mov		_t2, rax
		mov		rdx, r15
		Multiply64
		mov		rcx, rax
		call	CreateBDTake
		mov		rdx, _sgn
		mov		[rax]. ABD._sign, rdx
		mov		rdx, _dp
		mov		[rax]. ABD._decimalplace, rdx
		mov		rdx, _var9
		mov		[rax]. ABD._adj, rdx
		mov		r11, _v1
		cmp		_t3, 1
		je		Switch1
		mov		_v2, rax
		jmp		JustGoOut

Switch1:

		mov		_v1, rax

JustGoOut:

		mov		rcx, r11
		mov		rdx, rax
		cmp		qword ptr[rbp+40], 0				;cmp		_t1, 0
		je		CheckV3
		mov		rcx, rax
		mov		rdx, r12

CheckV3:

		cmp		_v3, 0
		je		CheckPow
		mov		rcx, r13
		call	Free64

CheckPow:

		cmp		_t2, 0
		je		GoOut
		mov		rcx, _t2
		call	Free64
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		cmp		rcx, 0
		jne		KeepV1
		mov		rcx, r11

KeepV1:

		cmp		rdx, 0
		jne		KeepV2
		mov		rdx, r12

KeepV2:


		mov		rcx, _v1
		mov		rdx, _v2

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _one, _one, _one, _one, _zero

		RET

	BDAlignDecimalPlaces ENDP
	;;
	;@brief Fixes the end of the value passed in.  Gets rid of any trailing zeros to the right of the decimal point and makes sure the decimal limit isnt exceeded.
	;
	;This function fixes the end of a  ABD instance.  It starts by checking the _decimalplace value of the instance, if it is > _decimallimit then it drops all the digits past _decimallimit+1,
	;then finds how it is supposed to handle rounding and deals with the last digit accordingly.  After the _decimalplace is corrected, it checks if the value has trailing zeros to the right
	;of the _decimalpoint.  If it does it will chop the trailing zeros up to the first non zero digit or the _decimalpoint, whichever comes first.
	;@returns a  ABD object.
	;@param rcx -  ABD object to be fixed.
	;;
	BDFixTheEnd PROC _val1: ABD

		ENTERFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, [rcx]. ABD._adj
		add		rdx, _decimallimit
		mov		_v2, rdx
		mov		_t1, 0
		mov		_t2, 0
		mov		_t5, 0
		mov		rax, [rcx]. ABD._num
		mov		_t3, rax
		mov		rax, [rcx]. ABD._sign
		mov		_sgn, rax
		mov		rax, [rcx]. ABD._decimalplace
		mov		_dp, rax
		cmp		rax, _v2
		jna		CheckForZs
		sub		rax, _v2
		cmp		rax, 1
		jna		DoRounding
		dec		rax
		mov		_jfunc, rax
		mov		rax, _v2
		inc		rax
		mov		_dp, rax
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		rbx, _jfunc
		mov		qword ptr[rax+8], rbx
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		rcx, _t2
		mov		rdx, _t1
		Pow64
		mov		rcx, _t2
		mov		_t2, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Mod64
		mov		rcx, _t1
		mov		_t1, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t1
		Sub64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64

DoRounding:

		cmp		_t2, 0
		jne		AllreadyHere
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

AllreadyHere:

		mov		rdx, _t2
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 10
		mov		rcx, _t3
		Mod64
		mov		_t4, rax
		mov		rdx, rax
		mov		rcx, _t3
		Sub64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t4
		dec		_dp
		cmp		qword ptr[rcx+8], 5
		jb		KeepIt
		mov		rdx, _t2
		mov		qword ptr[rdx+8], 1
		mov		rcx, _t3
		Add64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64

KeepIt:

		mov		rcx, _t4
		call	Free64

CheckForZs:

		mov		rcx, _t3
		cmp		qword ptr[rcx], 1
		ja		ItsGood
		cmp		qword ptr[rcx+8], 0
		ja		ItsGood
		mov		_dp, 0
		mov		_sgn, 0
		jmp		GoOut

ItsGood:

		cmp		_t2, 0
		jne		DontCreate
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

DontCreate:

		mov		rdx, _t2
		mov		qword ptr[rdx+8], 10

CheckForZerosLoop:

		cmp		_dp, 0
		je		GoOut
		mov		rcx, _t3
		mov		rdx, _t2
		Mod64
		mov		rcx, _t1
		mov		_t1, rax
		call	Free64
		mov		rdx, _t1
		cmp		qword ptr[rdx+8], 0
		jne		GoOut
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		dec		_dp
		jmp		CheckForZerosLoop

GoOut:

		cmp		_t1, 0
		je		CheckMod
		mov		rcx, _t1
		call	Free64

CheckMod:

		cmp		_t2, 0
		je		D_one
		mov		rcx, _t2
		call	Free64

D_one:

		mov		r11, _t3
		mov		rdi, qword ptr[r11]
		cmp		rdi, 1
		jna		ZerosDone

CheckZerosLoop:

		cmp		qword ptr[r11+rdi*8], 0
		ja		ZerosDone
		cmp		rdi, 1
		jna		ZerosDone
		dec		rdi
		mov		qword ptr[r11], rdi
		jmp		CheckZerosLoop

ZerosDone:

		mov		rax, _v1
		mov		rbx, _t3
		mov		[rax]. ABD._num, rbx
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx

		EXITFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		EXITFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDFixTheEnd ENDP
	;;
	;@brief Does the same thing as fix the end with the decimal limit adjusted (deprecated for the _adj variable in  ABD).
	;@sa BDFixTheEnd
	;@returns a  ABD object.
	;@param rcx -  ABD object to be fixed.
	;;
	BDWorkingFixTheEnd PROC _val1: ABD

		ENTERFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, 0
		mov		_t2, 0
		mov		_t5, 0
		mov		rax, [rcx]. ABD._num
		mov		_t3, rax
		mov		rax, [rcx]. ABD._sign
		mov		_sgn, rax
		mov		rax, [rcx]. ABD._decimalplace
		mov		_dp, rax
		cmp		rax, _workingdecimallimit
		jna		CheckForZs
		sub		rax, _workingdecimallimit
		cmp		rax, 1
		jna		DoRounding
		dec		rax
		mov		_jfunc, rax
		mov		rax, _workingdecimallimit
		inc		rax
		mov		_dp, rax
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		rbx, _jfunc
		mov		qword ptr[rax+8], rbx
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		rcx, _t2
		mov		rdx, _t1
		Pow64
		mov		rcx, _t2
		mov		_t2, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Mod64
		mov		rcx, _t1
		mov		_t1, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t1
		Sub64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64

DoRounding:

		cmp		_t2, 0
		jne		AllreadyHere
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

AllreadyHere:

		mov		rdx, _t2
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 10
		mov		rcx, _t3
		Mod64
		mov		_t4, rax
		mov		rdx, rax
		mov		rcx, _t3
		Sub64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		mov		rcx, _t4
		dec		_dp
		cmp		qword ptr[rcx+8], 5
		jb		KeepIt
		mov		rdx, _t2
		mov		qword ptr[rdx+8], 1
		mov		rcx, _t3
		Add64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64

KeepIt:

		mov		rcx, _t4
		call	Free64

CheckForZs:

		mov		rcx, _t3
		cmp		qword ptr[rcx], 1
		ja		ItsGood
		cmp		qword ptr[rcx+8], 0
		ja		ItsGood
		mov		_dp, 0
		mov		_sgn, 0
		jmp		GoOut

ItsGood:

		cmp		_t2, 0
		jne		DontCreate
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t2, rax
		mov		qword ptr[rax], 1
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_t1, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0

DontCreate:

		mov		rdx, _t2
		mov		qword ptr[rdx+8], 10

CheckForZerosLoop:

		cmp		_dp, 0
		je		GoOut
		mov		rcx, _t3
		mov		rdx, _t2
		Mod64
		mov		rcx, _t1
		mov		_t1, rax
		call	Free64
		mov		rdx, _t1
		cmp		qword ptr[rdx+8], 0
		jne		GoOut
		mov		rcx, _t3
		mov		rdx, _t2
		Divide64
		mov		rcx, _t3
		mov		_t3, rax
		call	Free64
		dec		_dp
		jmp		CheckForZerosLoop

GoOut:

		cmp		_t1, 0
		je		CheckMod
		mov		rcx, _t1
		call	Free64

CheckMod:

		cmp		_t2, 0
		je		D_one
		mov		rcx, _t2
		call	Free64

D_one:

		mov		r11, _t3
		mov		rdi, qword ptr[r11]
		cmp		rdi, 1
		jna		ZerosDone

CheckZerosLoop:

		cmp		qword ptr[r11+rdi*8], 0
		ja		ZerosDone
		cmp		rdi, 1
		jna		ZerosDone
		dec		rdi
		mov		qword ptr[r11], rdi
		jmp		CheckZerosLoop

ZerosDone:

		mov		rax, _v1
		mov		rbx, _t3
		mov		[rax]. ABD._num, rbx
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx

		EXITFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		EXITFUNC _zero, _one, _zero, _one, _one, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDWorkingFixTheEnd ENDP
	;;
	;@brief Checks for equality between 2  ABD objects.
	;
	;This algorithm works by aligning the decimalpoints then checking for equality.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDE64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rcx
		mov		_t2, rdx
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._sign
		cmp		rax, [rdx]. ABD._sign
		je		SignsEQ
		mov		rax, 0

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

SignsEQ:

		mov		r11, [rcx]. ABD._num
		mov		r12, [rdx]. ABD._num
		mov		rcx, qword ptr[r11]
		inc		rcx
		mov		rdi, r11
		mov		rsi, r12
		pushf
		cld
repe	cmpsq
		jne		PrepeNotEqual
		popf
		mov		rax, 1
		jmp		eGoOut

PrepeNotEqual:

		popf
		
eNotEqual:

		mov		rax, 0

eGoOut:

		mov		rcx, _v1
		cmp		rcx, _t1
		je		DoVal2
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc
		jmp		GoOut

DoVal2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		GoOut
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDE64 ENDP
	;;
	;@brief Checks for inequality between 2  ABD objects.
	;
	;this function simply sends the arrays to the BDE function then nots the return value.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDNE64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		DE64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BDNE64 ENDP	
	;;
	;@brief Checks if the first value is less than the second value.
	;
	;This function aligns the _decimalpoint, then compares the signs, if they are equal it checks to see if the absolute value of the first value is less than the absolute value of the second value.
	;if the _signs are negative it nots the result.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDLT64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t2, rcx
		mov		_t3, rdx
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._sign
		cmp		rax, [rdx]. ABD._sign
		ja		ltLessThan
		jb		ltNotLessThan
		mov		_t1, rax
		mov		r11, [rcx]. ABD._num
		mov		r12, [rdx]. ABD._num
		mov		rcx, qword ptr[r11]
		cmp		rcx, qword ptr[r12]
		ja		ltNotLessThan
		jb		ltLessThan
		mov		rax, rcx
		mul		_eight
		mov		rsi, r11
		mov		rdi, r12
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		jb		PrepltLessThan
		popf	
		mov		rax, 0
		jmp		ltGoOut

ltNotLessThan:

		mov		rax, 0
		jmp		ltGoOut

PrepltLessThan:

		popf

ltLessThan:

		mov		rax, 1

ltGoOut:

		cmp		_t1, 0
		je		LeaveIt
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one

LeaveIt:

		mov		rcx, _v1
		cmp		rcx, _t2
		je		CheckV2
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc
		jmp		GoOut

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t3
		je		GoOut
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET
			
	BDLT64 ENDP	
	;;
	;@brief Checks if the first value is less than or equal to the second value.
	;
	;This function switches the positions of the variables, sends them to the less than routine, then nots the result.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDLTOE64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._sign
		cmp		rax, [rdx]. ABD._sign
		ja		ItsLessThan
		jb		ItsNotLessThan
		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		DLT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		jmp		GoOut

ItsLessThan:

		mov		rax, 1
		jmp		GoOut

ItsNotLessThan:

		mov		rax, 0

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BDLTOE64 ENDP
	;;
	;@brief Checks if the first value is greater than the second value.
	;
	;This function reverses the positions of the variables and sends them to the less than routine.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDGT64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		DLT64		

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BDGT64 ENDP
	;;
	;@brief Checks if the first value is greater than or equal to the second value.
	;
	;This function sends the values to the less than function than nots the result.
	;@returns a bool.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDGTOE64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		DLT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
gtoeAllreadySet:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BDGTOE64 ENDP
	;;
	;@brief Adds 2  ABD objects.
	;
	;This function aligns the decimal points, then compares the signs, if they are equal it sends the arrays to the add function in Asm64.  If they are not equal, it compares the 2 arrays
	;and sets the larger to _val1, the other to _val2, sends them to the subtract routiine in Asm64, then creates a  ABD instance from the return value and sets the sign accordingly.
	;finally, it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDAdd64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rcx
		mov		_t2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._decimalplace
		mov		_dp, rax
		mov		rax, [r11]. ABD._sign
		mov		_sgn, rax
		cmp		rax, [r12]. ABD._sign
		je		Same
		mov		r8, [r11]. ABD._num
		mov		r9, [r12]. ABD._num
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		ja		Above
		jb		Below
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rdi, r8
		mov		rsi, r9
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		jb		PrepAbove
		ja		PrepBelow
		popf
		jmp		ReturnZero

PrepAbove:

		popf

Above:

		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Sub64
		jmp		GoOut
		
PrepBelow:

		popf

Below:

		mov		rax, [r12]. ABD._sign
		mov		_sgn, rax
		mov		rax, [r11]. ABD._num
		mov		rcx, [r12]. ABD._num
		mov		rdx, rax
		Sub64
		jmp		GoOut

Same:

		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num	
		Add64
		
GoOut:

		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd

DontDelete:

		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		KeepIt
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

KeepIt:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		jmp		KeepIt

ReturnZero:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rcx, rax
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		jmp		KeepIt

	BDAdd64 ENDP 
	;;
	;@brief Subtracts 2  ABD objects.
	;
	;This function aligns the decimal points, then compares the signs of the 2 instances.  If they DO NOT equal, it sends the 2 arrays to the add function in Asm64.  If they do equal
	; it compares the 2 arrays and sets the larger to _val1, the other to _val2 and send the to the subatract routine in Asm64, then creates a  ABD object from the return value
	; and sets the sign accordingly.  Finally, it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDSub64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rcx
		mov		_t2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._decimalplace
		mov		_dp, rax
		mov		rax, [r11]. ABD._sign
		mov		_sgn, rax
		cmp		rax, [r12]. ABD._sign
		jne		NotSame
		mov		r8, [r11]. ABD._num
		mov		r9, [r12]. ABD._num
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		ja		Above
		jb		Below
		mov		rcx, rdi
		mov		rax, rdi
		mul		_eight
		mov		rdi, r8
		mov		rsi, r9
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		jb		PrepAbove
		ja		PrepBelow
		popf
		jmp		ReturnZero

PrepAbove:

		popf

Above:

		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Sub64
		jmp		GoOut

PrepBelow:

		popf

Below:

		mov		rax, [r12]. ABD._sign
		cmp		rax, 0
		cmove	rax, _one
		cmovne	rax, _zero
		mov		_sgn, rax
		mov		rax, [r11]. ABD._num
		mov		rcx, [r12]. ABD._num
		mov		rdx, rax
		Sub64
		jmp		GoOut

NotSame:

		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Add64

GoOut:

		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd

DontDelete:

		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		KeepIt
		mov		_jfunc, rax
		call	DestroyBD
		mov		rax, _jfunc

KeepIt:

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		ErrorOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

ErrorOut:

		jmp		KeepIt

ReturnZero:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rcx, rax
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		jmp		KeepIt

	BDSub64 ENDP
	;;
	;@brief Multiplies 2  ABD objects.
	;
	;This function sends the 2 arrays to the multiply function in Asm64, creates a  ABD object from the result.  Sets the sign to the exclusive or of the sign of the 2 instances.
	;  Then sets the _decimalpoint to the sum of the 2 instances decimalpoints.  Then it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDMultiply64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		rax, 65537
		cmovne	rax, _zero
		mov		_t1, rax
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._sign
		xor		rax, [r12]. ABD._sign
		mov		_sgn, rax
		mov		rax, [r11]. ABD._decimalplace
		add		rax, [r12]. ABD._decimalplace
		mov		_dp, rax
		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Multiply64
		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rdx, _dp
		mov		[rax]. ABD._decimalplace, rdx
		mov		rdx, _var9
		mov		[rax]. ABD._adj, rdx
		mov		rcx, rax
		DFixTheEnd
		jmp		OutOfHere1

OtherOne:

		DWorkingFixTheEnd

OutOfHere1:

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDMultiply64 ENDP
	;;
	;@brief Squares a  ABD object.
	;
	;This function sends the array to the Square function in Asm64, creates a  ABD object from the result, sets the sign to 0, and the decimal point to the sum of the 2 instances decimal points.  The it sends
	;the result to the fix the end alg.
	;@returns a  ABD object.
	;@param rcx -  ABD object to be squared.
	;;
	BDSquare64 PROC _val1: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		r11, rcx
		mov		rcx, [r11]. ABD._num
		Square64
		mov		rcx, rax
		mov		_jfunc, rax
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		r8, [r11]. ABD._decimalplace
		add		r8, r8
		mov		[rax]. ABD._decimalplace, r8
		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._adj
		mov		[rax]. ABD._adj, rcx
		mov		rcx, _jfunc
		mov		_jfunc, rax
		call	Free64
		mov		rax, _jfunc
		mov		rcx, rax
		DFixTheEnd

OutOfHere1:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDSquare64 ENDP
	;;
	;@brief Divides one  ABD object from another.
	;
	;This algorithm starts by aligning the decimalpoints, then computes the modulus of the 2 instances, then divides the 2 instances and the result stored in _retval.  If the modulus is not zero, then the modulus is
	;mutliplied by 10, the result is divided by _val2, the _retval is multiplied by 10, the previous result is then added to the _retval, and the decimal point is incremented by one.  This continues
	;until either the modulus equals to zero, or the decimal limit +1 is reached.  Then it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDDivide64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		mov		_t3, 0
		mov		_t4, 0
		mov		_var3, 0
		mov		_var4, 0
		mov		_var1, 0
		mov		_dp, 0
		mov		_t1, rcx
		mov		_t2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._sign
		xor		rax, [r12]. ABD._sign
		mov		_sgn, rax
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Mod64
		mov		_var1, rax
		mov		_var3, 1
		mov		rcx, [r11]. ABD._num
		mov		rdx, rax
		Sub64
		mov		_var2, rax
		mov		_var4, 1
		mov		rcx, _var2
		mov		rdx, [r12]. ABD._num
		Divide64
		mov		_retval, rax
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		_t3, rax
		mov		_t4, 1

MainLoop:

		mov		rcx, _var1
		cmp		qword ptr[rcx], 1
		ja		Continue
		cmp		qword ptr[rcx+8], 0
		je		GoOut

Continue:

		mov		rdx, _t3
		Multiply64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, [r12]. ABD._num
		Mod64
		mov		rcx, _var1
		mov		_var1, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, _var1
		Sub64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, [r12]. ABD._num
		Divide64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _retval
		mov		rdx, _t3
		Multiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64
		mov		rcx, _retval
		mov		rdx, _var2
		Add64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64
		inc		_dp
		mov		rax, _decimallimit
		cmp		rax, _dp
		ja		MainLoop
		mov		rcx, _var1
		mov		rdx, _t3
		Multiply64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, [r12]. ABD._num
		Mod64
		mov		rcx, _var1
		mov		_var1, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, _var1
		Sub64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _var2
		mov		rdx, [r12]. ABD._num
		Divide64
		mov		rcx, _var2
		mov		_var2, rax
		call	Free64
		mov		rcx, _var2
		cmp		qword ptr[rcx+8], 5
		jb		KeepIt
		mov		rdx, _t3
		mov		qword ptr[rdx+8], 1
		mov		rcx, _retval
		Add64
		mov		rcx, _retval
		mov		_retval, rax
		call	Free64

KeepIt:

		cmp		_t4, 0
		je		CheckMod
		mov		rcx, _t3
		call	Free64

CheckMod:

		cmp		_var3, 0
		je		CheckTemp
		mov		rcx, _var1
		call	Free64

CheckTemp:

		cmp		_var4, 0
		je		CheckV1
		mov		rcx, _var2
		call	Free64

CheckV1:

		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		call	DestroyBD

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		GoOut
		call	DestroyBD
		jmp		GoOut
		
MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rcx, _retval
		call	CreateBDTake
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDDivide64 ENDP
	;;
	;@brief Modular reduction of a  ABD object.
	;
	;This function starts by aligning the decimal points, then sends the arrays to the Mod function in Asm64, creates a  ABD object with the return value and sets the sign equal to _val1.  Then it sends the result
	;to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDMod64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t3, rax
		cmp		rax, 65537
		cmovne	rax, _zero
		mov		_t3, rax
		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rcx
		mov		_t2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABD._decimalplace
		mov		_dp, rax
		mov		rax, [r11]. ABD._sign
		mov		_sgn, rax
		mov		rcx, [r11]. ABD._num
		mov		rdx, [r12]. ABD._num
		Mod64
		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd
		mov		_jfunc, rax
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		call	DestroyBD

Checkv2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		OutNow
		call	DestroyBD

OutNow:

		mov		rax, _jfunc

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

ReturnV1:

		mov		r11, _v1
		mov		rcx, [r11]. ABD._num
		mov		rax, qword ptr[rcx]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, rdi
		call	CreateBDTake
		mov		r11, _v1
		mov		rcx, [r11]. ABD._sign
		mov		[rax]. ABD._sign, rcx
		mov		rcx, [r11]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rcx
		mov		rcx, _var9
		mov		[rax]. ABD._adj, rcx
		mov		_jfunc, rax
		jmp		OutNow

	BDMod64 ENDP
	;;
	;@brief Special function used for exponentiation.
	;
	;This function is just used for the BDPow function.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDWorkingMultiply64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		mov		rax, [rcx]. ABD._sign
		xor		rax, [rdx]. ABD._sign
		mov		_sgn, rax
		mov		rax, [rcx]. ABD._decimalplace
		add		rax, [rdx]. ABD._decimalplace
		mov		_dp, rax
		mov		rcx, [rcx]. ABD._num
		mov		rdx, [rdx]. ABD._num
		Multiply64
		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rdx, _dp
		mov		[rax]. ABD._decimalplace, rdx
		mov		rdx, _var9
		mov		[rax]. ABD._adj, rdx
		mov		rcx, rax
		DWorkingFixTheEnd

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDWorkingMultiply64 ENDP
	;;
	;@brief Special function used for exponentiation.
	;
	;This function is just used for the BDPow function.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDWorkingSquare64 PROC _val1: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		r11, rcx
		mov		rcx, [rcx]. ABD._num
		Square64
		mov		rcx, rax
		mov		_jfunc, rax
		call	CreateBDTake
		mov		[rax]. ABD._sign, 0
		mov		r8, [r11]. ABD._decimalplace
		add		r8, r8
		mov		[rax]. ABD._decimalplace, r8
		mov		rcx, [r11]. ABD._adj
		mov		[rax]. ABD._adj, rcx
		mov		rcx, _jfunc
		mov		_jfunc, rax
		call	Free64
		mov		rax, _jfunc
		mov		rcx, rax
		DWorkingFixTheEnd

OutOfHere1:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDWorkingSquare64 ENDP
	;;
	;@brief Special function used for modular exponentiation.
	;
	;This function is just used in the ModPow function.
	;@returns a  ABD object.
	;@param rcx - first  ABD object.
	;@param rdx - second  ABD object.
	;;
	BDWorkingMod64 PROC _val1: ABD, _val2: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, rcx
		mov		_t2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		DAlignDecimalPlaces
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, [rcx]. ABD._decimalplace
		mov		_dp, r11
		mov		rax, [rcx]. ABD._sign
		mov		_sgn, rax
		mov		rcx, [rcx]. ABD._num
		mov		rdx, [rdx]. ABD._num
		Mod64
		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DWorkingFixTheEnd
		mov		_jfunc, rax
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		call	DestroyBD

Checkv2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		OutNow
		call	DestroyBD

OutNow:

		mov		rax, _jfunc

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

ReturnV1:

		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		mov		rax, qword ptr[rcx]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, rdi
		call	CreateBDTake
		mov		r11, _v1
		mov		rcx, [r11]. ABD._sign
		mov		[rax]. ABD._sign, rcx
		mov		rcx, [r11]. ABD._decimalplace
		mov		[rax]. ABD._decimalplace, rcx
		mov		rcx, _var9
		mov		[rax]. ABD._adj, rcx
		mov		_jfunc, rax
		jmp		OutNow

	BDWorkingMod64 ENDP
	;;
	;@brief Exponentiation of a  ABD object 2  ABD objects.
	;
	;This function starts by checking the decimal point of the exponent, if it does not equal zero, it sends the values to the power function in AdvancedAlgs64.  If it is zero then it sends the
	;values to the Pow function in Asm64 (along with the decimal point), then creates a  ABD object with the return value, checks the dign of the base, if it is 1 it checks to see if the exponent is even
	;or odd and sets the sign accordingly.  Then it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - Base  ABD object.
	;@param rdx - Exponent  ABD object.
	;;
	BDPow64 PROC _base: ABD, _exp: ABD

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		mov		_var9, rax
		cmp		[rdx]. ABD._decimalplace, 0
		ja		UseExp
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r12, rdx
		mov		r12, [rdx]. ABD._num
		mov		rax, [rcx]. ABD._decimalplace
		mov		_dp, rax
		mov		rax, [rcx]. ABD._sign
		mov		_sgn, rax
		cmp		rax, 0
		je		LeaveIt
		mov		rax, qword ptr[r12+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		mov		rdx, _v2
		mov		rdx, [rdx]. ABD._num
		mov		rax, _dp
		Pow64
		mov		_dp, rcx
		mov		rcx, rax
		call	CreateBDTake
		mov		rbx, _sgn
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _dp
		mov		[rax]. ABD._decimalplace, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		RET
	
UseExp:

		mov		rcx, _v1
		mov		rdx, _v2
		mov		r8, _defaultcyclecount
		call	BDPower64
		mov		rcx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		rcx, rax
		DFixTheEnd

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		RET

	BDPow64	ENDP
	;;
	;@brief Modular exponentiation of a  ABD object.
	;
	;If the decimal point on the exponent is greater than 0, this function will send the base and exponent to the Power function in AdvancedAlgs64, then compute the modular reduction of the result.
	;  Otherwise the function uses the standard repeated square and multiply algorithm.  Then it sends the result to the FixTheEnd alg.
	;@returns a  ABD object.
	;@param rcx - Base  ABD object.
	;@param rdx - Exponent  ABD object.
	;@param r8 - Modulus  ABD object.
	;;
	BDModPow64 PROC _base: ABD, _exp: ABD, _mod: ABD	

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_v3, r8
		mov		rax, [rcx]. ABD._adj
		cmp		rax, [rdx]. ABD._adj
		cmovb	rax, [rdx]. ABD._adj
		cmp		rax, [r8]. ABD._adj
		cmovb	rax, [r8]. ABD._adj
		mov		_var9, rax
		cmp		[r8]. ABD._decimalplace, 0
		ja		DoMod
		cmp		[rcx]. ABD._decimalplace, 0
		ja		DoMod
		mov		r12, rdx
		mov		rax, [rcx]. ABD._decimalplace
		mov		_t1, rax
		mov		_t2, rax
		mov		rax, [rcx]. ABD._sign
		mov		_sgn, rax
		cmp		rax, 0
		je		LeaveIt
		mov		rax, [rdx]. ABD._decimalplace
		mov		r12, rdx
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		mov		r11, [rcx]. ABD._num
		mov		r12, _v2
		mov		r12, [r12]. ABD._num
		mov		r13, [r8]. ABD._num
		lea		r10, _bits
		mov		rdi, qword ptr[r13]
		bsr		rax, qword ptr[r13+rdi*8]
		mov		rcx, 63
		sub		rcx, rax
		mov		rax, qword ptr[r11]
		cmp		rax, qword ptr[r13]
		cmovb	rax, qword ptr[r13]
		add		rax, rax
		add		rax, 10
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		r9, rax
		cmp		r9, 0
		je		GoOut
		mov		rdi, r9
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, r9
		call	CreateBDTake
		mov		r8, _v3
		mov		r11, _v1
		mov		rbx, [r11]. ABD._sign
		mov		[rax]. ABD._sign, rbx
		mov		rbx, _var9
		mov		[rax]. ABD._adj, rbx
		mov		r11, [r11]. ABD._num
		mov		r12, _v2
		mov		r12, [r12]. ABD._num
		mov		r9, rax
		mov		_retval, rax
		mov		rax, _t1
		mov		[r9]. ABD._decimalplace, rax
		cmp		qword ptr[r12], 0
		je		ReturnOneMod
		cmp		qword ptr[r12], 1
		ja		ContinueCheck
		cmp		qword ptr[r12+8], 0
		je		ReturnOneMod
		cmp		qword ptr[r12+8], 1
		je		ReturnV1Mod

ContinueCheck:

		cmp		qword ptr[r11], 0
		je		ReturnZero
		cmp		qword ptr[r11], 1
		ja		ContinueCheck2
		cmp		qword ptr[r11+8], 0
		je		ReturnZero
		cmp		qword ptr[r11+8], 1
		je		ReturnOneMod

ContinueCheck2:

		cmp		qword ptr[r13], 0
		je		GoOut
		cmp		qword ptr[r13], 1
		ja		ContinueOn
		cmp		qword ptr[r13+8], 0
		je		GoOut
		cmp		qword ptr[r13+8], 1
		je		ReturnZero

ContinueOn:

		mov		r14, qword ptr[r12]
		cmp		r14, 0
		je		GoOut
		bsr		r15, qword ptr[r12+r14*8]
		mov		_t4, r15
		cmp		r15, 0
		ja		DecAndGo
		mov		r15, 64
		mov		_t4, r15
		cmp		r14, 1
		jna		GoOut
		sub		r14, 1

DecAndGo:

		sub		r15, 1
		mov		_t4, r15

MainLoop:

		mov		rcx, _retval
		mov		rdx, _retval
		mov		rax, 65537
		DWorkingMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _v3
		mov		rax, 65537
		DWorkingMod64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		_t3, 1
		mov		r15, _t4
		lea		r10, _bits
		mov		rax, qword ptr[r12+r14*8]
		test	rax, qword ptr[r10+r15*8]
		jz		NoMult
		mov		rcx, _retval
		mov		rdx, _v1
		mov		rax, 65537
		DWorkingMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _v3
		mov		rax, 65537
		DWorkingMod64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

NoMult:

		cmp		r15, 0
		jna		NextByte
		dec		r15
		mov		_t4, r15
		jmp		MainLoop


NextByte:

		mov		r15, 63
		mov		_t4, r15
		cmp		r14, 1
		jna		GoOut
		dec		r14
		jmp		MainLoop

DoMod:

		mov		rcx, _v1
		mov		rdx, _v2
		DPow64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _v3
		DMod64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
	
GoOut:

		mov		rcx, _retval
		DFixTheEnd

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero

		RET	

ReturnOneMod:

		mov		r9, _retval
		mov		rbx, [r9]. ABD._num
		mov		qword ptr[rbx], 1
		mov		qword ptr[rbx+8], 1
		mov		[r9]. ABD._decimalplace, 0
		mov		[r9]. ABD._sign, 0
		mov		rcx, _retval
		mov		rdx, _v3
		DMod64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		jmp		GoOut

ReturnV1Mod:

		mov		rcx, _retval
		mov		rdx, _v3
		DMod64
		cmp		rax, 0
		je		GoOut
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		jmp		GoOut

ReturnZero:

		mov		r9, _retval
		mov		rbx, [r9]. ABD._num
		mov		qword ptr[rbx], 1
		mov		qword ptr[rbx+8], 1
		mov		[r9]. ABD._decimalplace, 0
		jmp		GoOut

	BDModPow64	ENDP
	;;
	;@brief To string function for  ABD objects.
	;
	;This function starts by sending the array to the ToString routine in Asm64, then it compares the decimal point to the number of digits returned from the algorithm.  If the decimal point is greater
	;then the number of digits, it inserts the correct number of zeros in front of the char array returned and adds a 0. in front.  If it is less than then it inserts the decimal point in the corrrect
	;position then checks if the sign==1, if so it appends a - to the front of the char array.
	;@returns a char array.
	;@param rcx -  ABD object to calc the string of.
	;@param rdx - qword representing the radix used for the calculation.
	;;
	BDToString64 PROC _val: ABD, _trad:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_t1, 0
		mov		_dp, 0
		mov		_t2, 0
		mov		_t3, 0
		mov		_t4, 0
		mov		_var2, 0
		mov		_var3, 0
		mov		_var4, 0
		mov		rcx, [rcx]. ABD._num
		mov		rdx, _v2
		ToString64
		mov		_v2, rax
		mov		_retval, rax
		mov		_t5, rcx
		mov		_var4, rcx
		mov		_var1, rcx
		mov		rcx, _v1
		cmp		[rcx]. ABD._sign, 0
		je		CheckDec
		mov		_t3, 1
		inc		_var1

CheckDec:

		mov		rdx, _v1
		mov		rax, [rdx]. ABD._decimalplace
		mov		_dp, rax
		cmp		rax, 0
		je		NoDec
		mov		_t4, 1

NoDec:

		inc		rax
		cmp		rax, _t5
		jne		MoveAhead
		mov		_var3, 1

MoveAhead:

		dec		rax
		cmp		rax, _t5
		jb		NoZeros
		sub		rax, _t5
		add		_var1, rax
		mov		_var2, rax
		mov		rcx, _var1
		add		rcx, 10
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		r8, rax
		mov		_retval, rax
		mov		rdi, rax
		mov		rcx, _var1
		add		rcx, 5
		mov		rax, 0
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _retval
		cmp		_t3, 0
		je		NoNeg1
		mov		byte ptr[rdi], '-'
		inc		rdi
		mov		_var4, 1

NoNeg1:

		mov		byte ptr[rdi], '0'
		inc		rdi
		mov		byte ptr[rdi], '.'
		inc		rdi
		add		_var4, 2
		mov		rcx, _var2
		inc		rcx
		add		_var4, rcx
		mov		rax, '0'
		pushf
		cld
rep		stosb
		mov		rsi, _v2
		mov		rcx, _t5
		add		_var4, rcx
rep		movsb
		popf
		jmp		ROut

NoZeros:

		cmp		_t3, 0
		ja		Isneg
		cmp		_t4, 0
		ja		IsNeg
		mov		rax, _v2
		mov		_retval, rax
		jmp		GoOut

IsNeg:

		mov		rcx, _var1
		add		rcx, 10
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_retval, rax
		mov		r8, rax
		mov		rdi, rax
		mov		rax, 0
		mov		rcx, _var1
		add		rcx, 5
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _retval
		cmp		_t3, 0
		je		NoNeg
		mov		byte ptr[rdi], '-'
		inc		rdi
		mov		_var4, 1

NoNeg:

		cmp		_var3, 0
		je		NoZero
		mov		byte ptr[rdi], '0'
		inc		rdi

NoZero:

		mov		rsi, _v2
		add		rdi, _t4
		add		rdi, _t5
		add		rsi, _t5
		cmp		_dp, 0
		je		NoDec2
		mov		rcx, _dp
		inc		rcx
		inc		rcx
		add		_var4, rcx
		pushf
		std
rep		movsb
		popf
		mov		byte ptr[rdi], '.'
		dec		rdi
		mov		rcx, _t5
		sub		rcx, _dp
		dec		rcx
		add		_var4, rcx
		inc		_var4
		pushf
		std
rep		movsb
		popf
		jmp		ROut

NoDec2:

		dec		rsi
		dec		rdi
		mov		rcx, _t5
		mov		_var4, rcx
		pushf
		std
rep		movsb
		popf

ROut:

		mov		rcx, _v2
		call	Free64
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rax, _retval
		mov		r11, _retval
		mov		rcx, _var4
		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDToString64 ENDP
















































	;;
	;@brief Creates a  ABF object equal to zero.
	;@returns a  ABF object.
	;;
	CreateBF PROC

		ENTERFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._numerator, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._denominator, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, _retval

		EXITFUNC _zero, _one, _one, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBF endp
	;;
	;@brief Creates a  ABF object equal to the array passed in.
	;@returns a  ABF object.
	;@param rcx - array to be copied to the numerator.
	;;
	CreateBFCopy PROC _num:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rax, _v1
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._numerator, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._denominator, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFCopy endp
	;;
	;@brief Creates a  ABF object equal to the arrays passed in.
	;@returns a  ABF object.
	;@param rcx - Array to copy to the numerator.
	;@param rdx - array to copy to the denominator.
	;;
	CreateBFCopyBoth PROC _num:PTR qword, _den:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rax, _v1
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._numerator, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _v2
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._denominator, rax
		mov		rdi, rax
		mov		rsi, _v2
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFCopyBoth endp
	;;
	;@brief Creates a  ABF object taking possession of the array passed in.
	;@returns a  ABF object.
	;@param rcx - array to take possession of for the numerator.
	;;
	CreateBFTake PROC _num:PTR qword

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, _v1
		mov		[rax]. ABF._numerator, rcx
		mov		rcx, 24
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._denominator, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFTake endp
	;;
	;@brief Creates a  ABF object taking possession of the arrays passed in.
	;@returns a  ABF object.
	;@param rcx - Array to take possession of for the numerator.
	;@param rdx - Array to take possession of for the denominator.
	;;
	CreateBFTakeBoth PROC _num:PTR qword, _den:PTR qword

		ENTERFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rcx, _v1
		mov		[rax]. ABF._numerator, rcx
		mov		rcx, _v2
		mov		[rax]. ABF._denominator, rcx

		EXITFUNC _zero, _one, _zero, _zero, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFTakeBoth endp
	;;
	;@brief Creates a  ABF object copying the values of another  ABF object.
	;@returns a  ABF object.
	;@param rcx -  ABF object to copy.
	;;
	CreateBFFromBF PROC _val: ABF

		ENTERFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABF._sign, 0
		mov		rax, _v1
		mov		rax, [rax]. ABF._numerator
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._numerator, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rsi, [rsi]. ABF._numerator
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _v1
		mov		rax, [rax]. ABF._denominator
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		rcx, _retval
		mov		[rcx]. ABF._denominator, rax
		mov		rdi, rax
		mov		rsi, _v1
		mov		rsi, [rsi]. ABF._denominator
		mov		rcx, qword ptr[rsi]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rax, _retval

		EXITFUNC _zero, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFFromBF endp
	;;
	;@brief Destroys a  ABF object.
	;@param rcx -  ABF object to delete.
	;;
	DestroyBF PROC _val: ABF

		ENTERFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, [rcx]. ABF._numerator
		call	Free64
		mov		rcx, _v1
		mov		rcx, [rcx]. ABF._denominator
		call	Free64
		mov		rcx, _v1
		call	free

		EXITFUNC _one, _one, _zero, _one, _one, _one, _one, _one, _one, _one, _zero, _zero, _zero, _zero, _zero

		RET

	DestroyBF endp
	;;
	;@brief Takes possession of the array passed in as the new numerator.
	;
	;This function deletes the current numerator and replaces it with the array passed in (by taking possession of the object).
	;@returns void.
	;@param rcx -  ABF object.
	;@param rdx -  An array.
	;;
	BFTakeNewNumerator PROC _val:ABF, _nn:PTR qword
		
		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx].ABF._numerator
		mov		[rcx].ABF._numerator, rdx
		mov		rcx, rax
		call	Free64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFTakeNewNumerator endp
	;;
	;@brief Takes possession of the array passed in as the new denominator.
	;
	;This function deletes the current denominator and replaces it with the array passed in (by taking possession of the object).
	;@returns void.
	;@param rcx -  ABF object.
	;@param rdx -  An array.
	;;
	BFTakeNewDenominator PROC _val:ABF, _nn:PTR qword
		
		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, [rcx].ABF._denominator
		mov		[rcx].ABF._denominator, rdx
		mov		rcx, rax
		call	Free64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFTakeNewDenominator endp
	;;
	;@brief Makes the denominators of 2  ABF objects equal to each other.
	;
	;This function sets the denominators equal to each other by cross multiplication for certain functions below.
	;@returns 2  ABF objects one in rcx representing the value passed in rcx and one in rdx representing the value passed in rdx.
	;@param rcx -  ABF object.
	;@param rdx -  ABF object.
	;;
	BFCommonDenominator64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rcx, [r11]. ABF._denominator
		mov		rdx, [r12]. ABF._denominator
		mov		rdi, qword ptr[rcx]
		cmp		rdi, qword ptr[rdx]
		jne		NotEq

CheckItLoop:

		mov		rax, qword ptr[rcx+rdi*8]
		cmp		rax, qword ptr[rdx+rdi*8]
		jne		NotEq
		sub		rdi, 1
		jnz		CheckItLoop
		mov		rcx, r11
		mov		rdx, r12
		jmp		OutOfHere

NotEq:

		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._denominator
		Multiply64
		mov		_t1, rax
		mov		rcx, [r12]. ABF._numerator
		mov		rdx, [r11]. ABF._denominator
		Multiply64
		mov		_t2, rax
		mov		rcx, [r11]. ABF._denominator
		mov		rdx, [r12]. ABF._denominator
		Multiply64
		mov		_t3, rax
		mov		rcx, _t1
		mov		rdx, _t3
		call	CreateBFTakeBoth
		mov		r11, _v1
		mov		rcx, [r11]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		_t4, rax
		mov		rcx, _t2
		mov		rdx, [rax]. ABF._denominator
		call	CreateBFCopyBoth
		mov		rcx, [r12]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		_t5, rax
		mov		rcx, _t2
		call	Free64
		mov		rcx, _t4
		mov		rdx, _t5

OutOfHere:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFCommonDenominator64 ENDP
	;;
	;@brief Reduces the result of a calculation by dividing both the numerator and the denominator with the GCD of the two.
	;
	;This function calculates the GCD of the numerator and the denominator, if it is greater than one it divides both by the gcd.  All calculations return a pure fraction, it does not split it into
	; an integer and fraction.
	;@returns a  ABF object.
	;@param rcx -  ABF object to reduce.
	;;
	BFReduce64 PROC _val1: ABF

		ENTERFUNC _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_t1, 0
		mov		_t2, 0
		mov		_retval, 0
		mov		r13, rcx
		mov		r11, [r13]. ABF._denominator
		mov		rax, rcx
		cmp		qword ptr[r11], 1
		ja		CheckIt
		cmp		qword ptr[r11+8], 1
		jna		OutOfHere

CheckIt:

		mov		r11, [r13]. ABF._numerator
		mov		rax, qword ptr[r11]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_t1, rax
		mov		rdi, rax
		mov		rsi, r11
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		mov		r11, _t1
		mov		r12, [r13]. ABF._denominator
		mov		rax, qword ptr[r12]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		mov		_t2, rax
		mov		rdi, rax
		mov		rsi, r12
		mov		rcx, qword ptr[r12]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		r12, _t2
		mov		rdi, qword ptr[r11]
		cmp		rdi, qword ptr[r12]
		ja		MainLoop
		jb		SwitchEm
		mov		rsi, r11
		mov		rdi, r12
		mov		rcx, qword ptr[r11]
		mov		rax, rcx
		mul		_eight
		add		rdi, rax
		add		rsi, rax
		pushf
		std
repe	cmpsq
		ja		PrepMainLoop
		jb		PrepSwitchEm
		popf
		jmp		MainLoop

PrepMainLoop:

		popf
		jmp		MainLoop

PrepSwitchEm:

		popf

SwitchEm:

		mov		rax, r11
		mov		r11, r12
		mov		r12, rax
		mov		_t4, r11

MainLoop:

		cmp		qword ptr[r12], 1
		ja		CarryOn
		cmp		qword ptr[r12+8], 0
		je		CheckDiv

CarryOn:

		mov		rcx, r11
		mov		rdx, r12
		Mod64
		mov		_retval, rax
		mov		rcx, r11
		call	Free64
		mov		r11, r12
		mov		r12, _retval
		mov		_t4, r11
		jmp		MainLoop

CheckDiv:

		mov		rcx, r12
		call	Free64
		cmp		qword ptr[r11], 1
		ja		DoDiv
		cmp		qword ptr[r11+8], 1
		jna		GoOut

DoDiv:

		mov		rcx, [r13]. ABF._numerator
		mov		rdx, r11
		Divide64
		mov		_t3, rax
		mov		rcx, r13
		mov		rdx, rax
		call	BFTakeNewNumerator
		mov		rcx, [r13]. ABF._denominator
		mov		rdx, _t4
		Divide64
		mov		rdx, rax
		mov		rcx, r13
		call	BFTakeNewDenominator

GoOut:

		mov		rcx, _t4
		call	Free64

OutOfHere:

		mov		rax, r13

		EXITFUNC _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero

		RET

	BFReduce64 ENDP
	;;
	;@brief Checks for equality between two  ABF objects.
	;
	;This alg sets the denominator equal then checks for equality.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFE64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t1, rcx
		mov		_t2, rdx
		FCommonDenominator64
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABF._sign
		cmp		rax, [r12]. ABF._sign
		je		SignsEQ
		mov		rax, 0
		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV22
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

CheckV22:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		OutOfHere2
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

OutOfHere2:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

SignsEQ:

		mov		r11, [rcx]. ABF._numerator
		mov		r12, [rdx]. ABF._numerator
		mov		rcx, qword ptr[r11]
		cmp		rcx, qword ptr[r12]
		jne		eNotEqual

eLoopIt:

		mov		rax, qword ptr[r11+rcx*8]
		cmp		rax, qword ptr[r12+rcx*8]
		jne		eNotEqual
		loope	eLoopIt
		mov		rax, 1
		jmp		eGoOut
		
eNotEqual:

		mov		rax, 0

eGoOut:

		mov		rcx, _v1
		cmp		rcx, _t1
		je		CheckV2
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t2
		je		OutOfHere
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

OutOfHere:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFE64 ENDP
	;;
	;@brief Checks for inequality between two  ABF objects.
	;
	;This alg sends the values to the BFE function then nots the answer.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFNE64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		FE64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BFNE64 ENDP	
	;;
	;@brief Checks if the first value is less than the second value.
	;
	;This function sets the denominators equal then checks if the first value is less than the second value.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFLT64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t2, rcx
		mov		_t3, rdx
		FCommonDenominator64
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABF._sign
		cmp		rax, [r12]. ABF._sign
		ja		ltLessThan
		jb		ltNotLessThan
		mov		_t1, rax
		mov		r11, [rcx]. ABF._numerator
		mov		r12, [rdx]. ABF._numerator
		mov		rcx, qword ptr[r11]
		cmp		rcx, qword ptr[r12]
		ja		ltNotLessThan
		jb		ltLessThan

ltLoopIt:

		mov		rax, qword ptr[r11+rcx*8]
		cmp		rax, qword ptr[r12+rcx*8]
		ja		ltNotLessThan
		jb		ltLessThan
		loope	ltLoopIt
		mov		rax, 0
		jmp		ltGoOut

ltNotLessThan:

		mov		rax, 0
		jmp		ltGoOut

ltLessThan:

		mov		rax, 1

ltGoOut:

		cmp		_t1, 0
		je		LeaveIt
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one

LeaveIt:

		mov		rcx, _v1
		cmp		rcx, _t2
		je		CheckV2
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

CheckV2:

		mov		rcx, _v2
		cmp		rcx, _t3
		je		OutOfHere
		mov		_retval, rax
		call	DestroyBD
		mov		rax, _retval

OutOfHere:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET
			
	BFLT64 ENDP	
	;;
	;@brief Checks if the first value is less than or equal to the second value.
	;
	;This alg switches the values, calls the less than alg, then nots the return value.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFLTOE64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABF._sign
		cmp		rax, [r12]. ABF._sign
		ja		ItsLessThan
		jb		ItsNotLessThan
		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		FLT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		jmp		GoOut

ItsLessThan:

		mov		rax, 1
		jmp		GoOut

ItsNotLessThan:

		mov		rax, 0

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET
		
	BFLTOE64 ENDP
	;;
	;@brief Checks if the first value is greater than the second value.
	;
	;This alg switches the values, then calls the less than alg.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFGT64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, rcx
		mov		rcx, rdx
		mov		rdx, rax
		FLT64		

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BFGT64 ENDP
	;;
	;@brief Checks if the first value is greater than or equal to the second value.
	;
	;This alg calls the less than alg then nots the return value.
	;@returns a bool.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFGTOE64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		FLT64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
gtoeAllreadySet:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BFGTOE64 ENDP
	;;
	;@brief Adds 2  ABF objects.
	;
	;This alg sets the denominators equal, then checks the signs, if they are equal it sends the numerators to the Add function in Asm64, if not, it compares the two numerators, sets the larger array to _val1,
	;and the other to _val2 then sends it to the subtract function in Asm64.  It creates a  ABF object from the return value and sets the sign accordingly.  Finally it sends the result to the reduce alg.
	;@returns a  ABF object.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFAdd64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t1, rcx
		mov		_t2, rdx
		mov		r11, rcx
		mov		r12, rdx
		FCommonDenominator64
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rbx, [r11]. ABF._sign
		mov		_sgn, rbx
		cmp		rbx, [r12]. ABF._sign
		jne		NotEqual
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		Add64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, [r12]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rdx, _sgn
		mov		[rax]. ABF._sign, rdx
		mov		rcx, _t3
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere1
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere1:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

NotEqual:

		mov		r8, [r11]. ABF._numerator
		mov		r9, [r12]. ABF._numerator
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		ja		Larger
		jb		Smaller

CheckItLoop:

		mov		rax, qword ptr[r8+rdi*8]
		cmp		rax, qword ptr[r9+rdi*8]
		ja		Larger
		jb		Smaller
		sub		rdi, 1
		jnz		CheckItLoop

Larger:

		mov		rbx, [r11]. ABF._sign
		mov		_sgn, rbx
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		Sub64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		rcx, _t3
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere2
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere2:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Smaller:

		mov		rbx, [r12]. ABF._sign
		mov		_sgn, rbx
		mov		rcx, [r12]. ABF._numerator
		mov		rdx, [r11]. ABF._numerator
		Sub64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		rcx, _t3
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere3
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere3:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFAdd64 ENDP 
	;;
	;@brief Subtracts 2  ABF objects.
	;
	;This function sets the denominators equal to each other then compares the signs.  If they DO NOT equal, it sends the numerators to the add function in Asm64.  if they do equal, it compares the 2 numerators
	;sets the larger to _val1, the other to _val2 and sends the numerators to the subtract routine in Asm64.  It then creates a  ABF object from the return value and sets the sign accordingly.
	;finally, it sends the result to the redulce routine.
	;@returns a  ABF object.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFSub64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		FCommonDenominator64
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rbx, [r11]. ABF._sign
		mov		_sgn, rbx
		cmp		rbx, [r12]. ABF._sign
		je		TheyEqual
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		Add64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, [r12]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rdx, _sgn
		mov		[rax]. ABF._sign, rdx
		mov		rcx, _t2
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere1
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere1:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

TheyEqual:

		mov		r8, [r11]. ABF._numerator
		mov		r9, [r12]. ABF._numerator
		mov		rdi, qword ptr[r8]
		cmp		rdi, qword ptr[r9]
		ja		Larger
		jb		Smaller

CheckIt:

		mov		rax, qword ptr[r8+rdi*8]
		cmp		rax, qword ptr[r9+rdi*8]
		ja		Larger
		jb		Smaller
		sub		rdi, 1
		jnz		CheckIt

Larger:

		mov		rbx, [r11]. ABF._sign
		mov		_sgn, rbx
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		Sub64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _v2
		mov		rdx, [r12]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		rcx, _t2
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere2
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere2:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

Smaller:

		cmp		[r12]. ABF._sign, 0
		cmove	rbx, _one
		cmovne	rbx,_zero
		mov		_sgn, rbx
		mov		rax, [r11]. ABF._numerator
		mov		rcx, [r12]. ABF._numerator
		mov		rdx, rax
		Sub64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _v2
		mov		rdx, [r12]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		rcx, _t2
		call	Free64
		mov		rcx, _v1
		cmp		rcx, _t1
		je		OutOfHere3
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

OutOfHere3:

		mov		rcx, _retval
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _one, _zero, _one, _one, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFSub64 ENDP
	;;
	;@brief Multiplies 2  ABF objects.
	;
	;This alg sends the two numerators to the multiply routine in Asm64, then sends the two denominators to the multiply routine in Asm64 and creates a  ABF object from the results. It then sets
	;the sign to the exclusive or of the two signs.  It then sends the result
	;to the Reduce alg.
	;@returns a  ABF object.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFMultiply64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [rcx]. ABF._sign
		xor		rax, [rdx]. ABF._sign
		mov		_sgn, rax
		mov		rcx, [rcx]. ABF._numerator
		mov		rdx, [rdx]. ABF._numerator
		Multiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABF._denominator
		mov		rdx, [r12]. ABF._denominator
		Multiply64
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBFTakeBoth
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		rcx, rax
		FReduce64

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFMultiply64 ENDP
	;;
	;@brief Squares a  ABF object.
	;
	;This routine sends the numerator to the square routine in then sends the denominator to the square routine in Asm64 and creates a  ABF object from the results and sets the sign to 0.
	;It then sends the result to the Reduce alg.
	;@returns a  ABF object.
	;@param rcx -  ABF object to square.
	;;
	BFSquare64 PROC _val1: ABF

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		r11, rcx
		mov		rcx, [rcx]. ABF._numerator
		Square64
		mov		_t1, rax
		mov		rcx, [r11]. ABF._denominator
		Square64
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBFTakeBoth
		mov		[rax]. ABF._sign, 0
		mov		rcx, rax
		FReduce64

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BFSquare64 ENDP
	;;
	;@brief Divides one  ABF object from another.
	;
	;This alg sends the numerator from _val1 and the denominator from _val2 to the multiply routine in the Asm64 file (to be used for the result numerator), then it sends the denominator from _val1
	;and the numerator from _val2 to the multiply alg in Asm64 (to be used for the result denominator) then creates a  ABF object from the results and sets the sign to the exclusive or of the 2 signs.
	;  Finally it sends the result to the Reduce alg.
	;@returns a  ABF object.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFDivide64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r11]. ABF._sign
		xor		rax, [r12]. ABF._sign
		mov		_sgn, rax
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._denominator
		Multiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABF._denominator
		mov		rdx, [r12]. ABF._numerator
		Multiply64
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBFTakeBoth
		mov		rcx, rax
		FReduce64
		mov		rcx, _sgn
		mov		[rax]. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFDivide64 ENDP
	;;
	;@brief Modular reduction of a  ABF object.
	;
	;This alg sets the denominators equal to each other, then sends the numerators to the Mod alg in Asm64 and creates a  ABF object from the result and sets the sign equal to the sign in _val1.
	;  Finally it sends the result to the Reduce routine.
	;@returns a  ABF object.
	;@param rcx - first  ABF object.
	;@param rdx - second  ABF object.
	;;
	BFMod64 PROC _val1: ABF, _val2: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_t1, rcx
		mov		_t2, rdx
		FCommonDenominator64
		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		Mod64
		mov		_t3, rax
		mov		rcx, rax
		mov		rdx, [r12]. ABF._denominator
		call	CreateBFCopyBoth
		mov		_retval, rax
		mov		rcx, _t3
		call	Free64
		mov		rax, _retval
		mov		r11, _v1
		mov		rcx, [r11]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		rcx, _v1
		cmp		rcx, _t1
		je		GoOut
		call	DestroyBD
		mov		rcx, _v2
		call	DestroyBD
		mov		rax, _retval

GoOut:

		mov		rcx, rax
		FReduce64

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFMod64 ENDP
	;;
	;@brief Exponentiation of a  ABF object.
	;
	;This alg checks the denominator, if its greater than one it sends the values to the Power function.  Otherwise it uses the repeated square and multiply alg to calculate the result.  Then it sends
	;the result to the Reduce alg.
	;@returns a  ABF object.
	;@param rcx - Base  ABF object.
	;@param rdx - Exponent  ABF object.
	;;
	BFPow64 PROC _base: ABF, _exp: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _one, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		r11, rcx
		mov		r12, rdx
		mov		rax, [r12]. ABF._numerator
		cmp		qword ptr[rax], 1
		ja		MoveOn
		cmp		qword ptr[rax+8], 0
		je		ReturnOne
		cmp		qword ptr[rax+8], 1
		je		ReturnVal1
		mov		rax, [r11]. ABF._sign
		mov		_sgn, rax
		cmp		rax, 0
		je		LeaveIt

MoveOn:

		mov		rax, [r12]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		lea		r13, _bits
		mov		rcx, [r12]. ABF._numerator
		mov		rax, qword ptr[rcx]
		mov		r14, rax
		bsr		r15, qword ptr[rcx+rax*8]
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		r8, rax
		mov		_retval, rax
		cmp		r15, 0
		je		NextByte
		dec		r15

MainLoop:

		mov		rcx, _retval
		mov		rdx, _retval
		FMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		r9, [r12]. ABF._numerator
		mov		rax, qword ptr[r13+r15*8]
		test	qword ptr[r9+r14*8], rax
		jz		PassMult
		mov		rcx, _v1
		mov		rdx, _retval
		call	BFMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

PassMult:

		cmp		r15, 0
		je		NextByte
		dec		r15
		jmp		MainLoop

NextByte:

		cmp		r14, 1
		jna		GoOut
		mov		r15, 63
		dec		r14
		jmp		MainLoop

GoOut:

		mov		rcx, _retval
		FReduce64

OutOfHere:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _one, _one, _one, _zero

		RET

ReturnOne:

		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rcx, rax
		call	CreateBFTake
		mov		rcx, [rax]. ABF._denominator
		mov		qword ptr[rcx+8], 1
		mov		qword ptr[rcx], 1
		jmp		OutOfHere

ReturnVal1:

		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r11]. ABF._denominator
		call	CreateBFCopyBoth
		mov		r11, _v1
		mov		rcx, [r11]. ABF._sign
		mov		[rax]. ABF._sign, rcx
		mov		_retval, rax
		jmp		GoOut
	
	BFPow64	ENDP
	;;
	;@brief Modular exponentiation of a  ABF object 2  ABF objects.
	;
	;In this alg, if the exponent denominator is greater than one, it sends the base and exponent to the power function then computes the modular reduction.  Otherwise it uses the repeated square and
	;multiply algorithm to calculate the result, then sends the result to the Reduce routine.
	;@returns a  ABF object.
	;@param rcx - Base  ABF object.
	;@param rdx - Exponent  ABF object.
	;@param r8 - Modulus  ABF object.
	;;
	BFModPow64 PROC _base: ABF, _exp: ABF, _mod: ABF	

		ENTERFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_v3, r8
		mov		r11, rcx
		mov		r12, rdx
		mov		r13, r8
		mov		rax, [r11]. ABF._sign
		mov		_sgn, rax
		cmp		rax, 0
		je		LeaveIt
		mov		rax, [r12]. ABF._numerator
		mov		rax, qword ptr[rax+8]
		xor		rdx, rdx
		div		_two
		cmp		rdx, 0
		jne		LeaveIt
		mov		_sgn, 0

LeaveIt:

		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r12]. ABF._numerator
		mov		r8, [r13]. ABF._numerator
		ModPow64
		mov		rcx, rax
		call	CreateBFTake
		mov		rbx, _sgn
		mov		[rax]. ABF._sign, rbx
		mov		_retval, rax
		mov		rcx, rax
		call	BFReduce64

		EXITFUNC _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _one, _one, _one, _zero, _zero, _zero

		RET	

	BFModPow64	ENDP
	;;
	;@brief To string function for  ABF objects.
	;
	;This alg sends the numerator to the ToString routine in Asm64 then appends a / to the end and sends the denominator to the ToString routine.  finally if the sign equals one it appends a -
	; to the front of the char array.
	;@returns a char array.
	;@param rcx -  ABF object.
	;@param rdx - qword Radix to use.
	;;
	BFToString64 PROC _val: ABF, _trad:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, [rcx]. ABF._numerator
		mov		rdx, _v2
		ToString64
		mov		_t1, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABF._denominator
		mov		rdx, _v2
		ToString64
		mov		_t2, rax
		mov		rcx, _v1
		mov		rax, [rcx]. ABF._numerator
		mov		rax, qword ptr[rax]
		mov		rdx, [rcx]. ABF._denominator
		add		rax, qword ptr[rdx]
		mul		_twenty
		mul		_eight
		add		rax, 5
		mov		rcx, rax
		call	GetMem64
		mov		_v2, rax
		mov		rcx, _v1
		mov		rax, [rcx]. ABF._numerator
		mov		rax, qword ptr[rax]
		mov		rdx, [rcx]. ABF._denominator
		add		rax, qword ptr[rdx]
		mul		_twenty
		add		rax, 5
		mov		rcx, rax
		xor		rax, rax
		mov		rax, 0
		mov		rdi, _v2
		mov		r8, _v2
rep		stosb
		mov		rax, _v2
		mov		r8, _t1
		mov		rsi, 0
		mov		rcx, _v1
		mov		rdi, 0
		mov		rsi, 0
		xor		rcx, rcx
		cmp		[rcx]. ABF._sign, 0
		je		CopyNumLoop
		mov		byte ptr[rax], '-'
		mov		rdi, 1

CopyNumLoop:

		cmp		byte ptr[r8+rsi], 0
		je		NOut
		mov		cl, byte ptr[r8+rsi]
		mov		byte ptr[rax+rdi], cl
		inc		rdi
		inc		rsi
		jmp		CopyNumLoop

NOut:

		mov		rdx, _v1
		mov		rdx, [rdx]. ABF._denominator
		cmp		qword ptr[rdx], 1
		ja		DoDen
		cmp		qword ptr[rdx+8], 1
		jna		DOut

DoDen:

		mov		r8, _t2
		mov		rsi, 0
		mov		byte ptr[rax+rdi], '/'
		inc		rdi

CopyDenLoop:

		cmp		byte ptr[r8+rsi], 0
		je		DOut
		mov		cl, byte ptr[r8+rsi]
		mov		byte ptr[rax+rdi], cl
		inc		rdi
		inc		rsi
		jmp		CopyDenLoop

DOut:

		mov		byte ptr[rax+rdi], 0
		mov		rcx, _retval
		call	Free64

GoOut:

		mov		rax, _v2

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFToString64 ENDP





































	;;
	;@brief Creates a  ABDI object equal to zero.
	;@returns a  ABDI object.
	;;
	CreateBDI PROC

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDI endp
	;;
	;@brief Creates a  ABDI object by copying another  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - A  ABDI object.
	;;
	CreateBDIFromBDI PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		call	CreateBDFromBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		call	CreateBDFromBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDIFromBDI endp
	;;
	;@brief Creates a  ABDI object By copying a  ABD object.
	;@returns a  ABDI object.
	;@param rcx - The  ABD object to copy to the real part.
	;;
	CreateBDICopyBD PROC _r: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDFromBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDICopyBD endp
	;;
	;@brief Creates a  ABDI object by copying an array.
	;@returns a  ABDI object.
	;@param rcx - The array to copy to the real part.
	;;
	CreateBDICopyArray PROC _r:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDCopy
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDICopyArray endp
	;;
	;@brief Creates a  ABDI object by copying 2  ABD objects.
	;@returns a  ABDI object.
	;@param rcx - The  ABD object to copy to the real part.
	;@param rdx - The  ABD object to copy to the imaginary part.
	;;
	CreateBDICopyBothBD PROC _r: ABD, _i: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDFromBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		mov		rcx, _v2
		call	CreateBDFromBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDICopyBothBD endp
	;;
	;@brief Creates a  ABDI object by copying 2 arrays.
	;@returns a  ABDI object.
	;@param rcx - The array to copy to the real part.
	;@param rdx - The array to copy to the imaginary part.
	;;
	CreateBDICopyBothArray PROC _r:PTR qword, _i:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDCopy
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		mov		rcx, _v2
		call	CreateBDCopy
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDICopyBothArray endp
	;;
	;@brief Creates a  ABDI object by taking possession of a  ABD object.
	;@returns a  ABDI object.
	;@param rcx - The  ABD object to take possession for the real part.
	;;
	CreateBDITakeBD PROC _r: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		mov		[rax]. ABDI._real, rcx
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDITakeBD endp
	;;
	;@brief Creates a  ABDI object by taking possession of an array.
	;@returns a  ABDI object.
	;@param rcx - The array to take possession of for the real part.
	;;
	CreateBDITakeArray PROC _r:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDTake
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		call	CreateBD
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDITakeArray endp
	;;
	;@brief Creates a  ABDI object by taking possession of 2  ABD objects.
	;@returns a  ABDI object.
	;@param rcx - The  ABD object to take possession of for the real part.
	;@param rdx - The  ABD object to take possession of for the imaginary part.
	;;
	CreateBDITakeBothBD PROC _r: ABD, _i: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		mov		[rax]. ABDI._real, rcx
		mov		rcx, _v2
		mov		[rax]. ABDI._im, rcx
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDITakeBothBD endp
	;;
	;@brief Creates a  ABDI object by taking possession of 2 arrays.
	;@returns a  ABDI object.
	;@param rcx - The array to take possession of for the real part.
	;@param rdx - The array to take possession of for the imaginary part.
	;;
	CreateBDITakeBothArray PROC _r:PTR qword, _i:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0
		mov		rcx, _v1
		call	CreateBDTake
		mov		rcx, _retval
		mov		[rcx]. ABDI._real, rax
		mov		rcx, _v2
		call	CreateBDTake
		mov		rcx, _retval
		mov		[rcx]. ABDI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBDITakeBothArray endp
	;;
	;@brief Destroys a  ABDI object.
	;@param rcx - The  ABDI object to destroy.
	;;
	DestroyBDI PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, [rcx]. ABDI._real
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		call	DestroyBD
		mov		rcx, _v1
		call	free

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	DestroyBDI endp
	;;
	;@brief Returns the conjugate of a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to conjugate.
	;;
	BDIConjugate PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		rdx, [rcx]. ABDI._im
		mov		rcx, [rcx]. ABDI._real
		call	CreateBDICopyBothBD
		mov		_retval, rax
		mov		rdx, [rax]. ABDI._im
		cmp		[rdx]. ABD._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rdx]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDIConjugate ENDP
	;;
	;@brief Converts a  ABDI object to polar form.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDIToPolar PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABDI._polarform, 1
		jne		DoCalc
		call	CreateBDIFromBDI
		mov		_retval, rax
		jmp		ExitNow

DoCalc:

		mov		_v1, rcx
		mov		rdx, [rcx]. ABDI._real
		mov		rcx, rdx
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABDI._im
		mov		rcx, rdx
		call	BDMultiply64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BDSqrt64
		mov		_retval, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _bdzero
		call	BDE64
		cmp		rax, 1
		je		Equal
		mov		rcx, _v1
		mov		rdx, [rcx]. ABDI._real
		mov		rcx, [rcx]. ABDI._im
		call	BDDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BDRadATan64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _bdzero
		call	BDGT64
		cmp		rax, 1
		je		GoOut
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _bdzero
		call	BDLT64
		cmp		rax, 1
		je		DoSub
		mov		rcx, _t1
		mov		rdx, _bdpi
		call	BDAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		jmp		GoOut

DoSub:

		mov		rcx, _t1
		mov		rdx, _bdpi
		call	BDSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		jmp		GoOut

Equal:

		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _bdzero
		call	BDE64
		cmp		rax, 1
		je		ErrorOut
		mov		rcx, _bdpi
		call	CreateBDFromBD
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _bdzero
		call	BDLT64
		cmp		rax, 0
		je		GoOut
		mov		rax, _t1
		mov		[rax]. ABD._sign, 1

GoOut:

		mov		rcx, _retval
		mov		rdx, _t1
		call	CreateBDITakeBothBD
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 1
		mov		rcx, _t2
		call	DestroyBD

ExitNow:

		mov		rax, _retval

ErrorOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDIToPolar ENDP
	;;
	;@brief Converts a  ABDI object to rectangular form.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDIFromPolar PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABDI._polarform, 1
		jne		DoCalc
		call	CreateBDIFromBDI
		mov		_retval, rax
		jmp		ExitNow

DoCalc:

		mov		_v1, rcx
		mov		rcx, [rcx]. ABDI._im
		call	BDRadCos64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABDI._real
		DMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		call	BDRadSin64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABDI._real
		DMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBDITakeBothBD
		mov		_retval, rax
		mov		[rax]. ABDI._polarform, 0

ExitNow:

		mov		rax, _retval

ErrorOut:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDIFromPolar ENDP
	;;
	;@brief Checks for equality between 2  ABDI objects.
	;@returns a bool.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDIE64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		cmp		[rcx]. ABDI._polarform, 0
		je		CheckNext
		cmp		[rdx]. ABDI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BDIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABDI._polarform, 0
		je		DoIt
		call	BDIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r12]. ABDI._real
		call	BDE64
		cmp		rax, 0
		je		GoOut
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r12]. ABDI._im
		call	BDE64
		jmp		GoOut

GoOut:

		mov		_retval, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBDI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBDI

V2Good:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDIE64 ENDP
	;;
	;@brief Checks for inequality between 2  ABDI objects.
	;@returns a bool.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDINE64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		call	BDIE64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BDINE64 ENDP	
	;;
	;@brief Adds 2  ABDI objects.
	;@returns a  ABDI object.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDIAdd64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		mov		_var3, 0
		cmp		[rcx]. ABDI._polarform, 0
		je		CheckNext
		mov		_var3, 1
		cmp		[rdx]. ABDI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BDIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABDI._polarform, 0
		je		DoIt
		mov		_var3, 1
		call	BDIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r12]. ABDI._real
		call	BDAdd64
		mov		_retval, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r12]. ABDI._im
		call	BDAdd64
		mov		_t3, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBDI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBDI

V2Good:

		mov		rcx, _retval
		mov		rdx, _t3
		call	CreateBDITakeBothBD
		mov		rcx, _var3
		mov		[rax]. ABDI._polarform, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDIAdd64 ENDP 
	;;
	;@brief Subtracts 2  ABDI objects.
	;@returns a  ABDI object.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDISub64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		mov		_var3, 0
		cmp		[rcx]. ABDI._polarform, 0
		je		CheckNext
		mov		_var3, 1
		cmp		[rdx]. ABDI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BDIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABDI._polarform, 0
		je		DoIt
		mov		_var3, 1
		call	BDIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r12]. ABDI._real
		call	BDSub64
		mov		_retval, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r12]. ABDI._im
		call	BDSub64
		mov		_t3, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBDI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBDI

V2Good:

		mov		rcx, _retval
		mov		rdx, _t3
		call	CreateBDITakeBothBD
		mov		rcx, _var3
		mov		[rax]. ABDI._polarform, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDISub64 ENDP
	;;
	;@brief Multiplies 2  ABDI objects.
	;@returns a  ABDI object.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDIMultiply64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		[rcx]. ABDI._polarform, 0
		je		CheckNext
		mov		_var1, rcx
		cmp		[rdx]. ABDI._polarform, 1
		je		DoPolar
		mov		rcx, rdx
		call	BDIToPolar
		mov		_var2, rdx
		jmp		DoPolar

CheckNext:

		cmp		[rdx]. ABDI._polarform, 0
		je		DoCartesian
		mov		_var2, rdx
		call	BDIToPolar
		mov		_var1, rax
		jmp		DoPolar

DoCartesian:

		mov		r11, _v1
		mov		r12, _v2
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r12]. ABDI._im
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r12]. ABDI._real
		call	BDMultiply64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		_var1, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r11, _v1
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r12]. ABDI._real
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r12]. ABDI._im
		call	BDMultiply64
		mov		_t2, rax
		cmp		[rax]. ABD._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABD._sign, rcx
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		_retval, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var1
		call	CreateBDITakeBothBD

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoPolar:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _var2
		mov		rdx, [rdx]. ABDI._real
		DMultiply64
		mov		_t1, rax
		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _var2
		mov		rdx, [rdx]. ABDI._im
		DAdd64
		mov		_t2, rax
		mov		rcx, _v1
		cmp		[rcx]. ABDI._polarform, 1
		je		V1Good
		mov		rcx, _var1
		call	DestroyBDI

V1Good:

		mov		rcx, _v2
		cmp		[rcx]. ABDI._polarform, 1
		je		V2Good
		mov		rcx, _var2
		call	DestroyBDI

V2Good:

		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBDITakeBothBD
		mov		[rax]. ABDI._polarform, 1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDIMultiply64 ENDP
	;;
	;@brief Squares a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABDI object to square.
	;;
	BDISquare64 PROC _val1: ABDI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		cmp		[rcx]. ABDI._polarform, 1
		je		DoPolar
		mov		r11, _v1
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r11]. ABDI._im
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r11]. ABDI._real
		call	BDMultiply64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		_var1, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r11, _v1
		mov		rcx, [r11]. ABDI._real
		mov		rdx, [r11]. ABDI._real
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABDI._im
		mov		rdx, [r11]. ABDI._im
		call	BDMultiply64
		mov		_t2, rax
		cmp		[rax]. ABD._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABD._sign, rcx
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		_retval, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _retval
		mov		rdx, _var1
		call	CreateBDITakeBothBD

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoPolar:

		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, rcx
		DMultiply64
		mov		_t1, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, rcx
		DAdd64
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBDITakeBothBD
		mov		[rax]. ABDI._polarform, 1

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BDISquare64 ENDP
	;;
	;@brief Divides 2  ABDI objects.
	;@returns a  ABDI object.
	;@param rcx - The first  ABDI object.
	;@param rdx - The second  ABDI object.
	;;
	BDIDivide64 PROC _val1: ABDI, _val2: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		[rcx]. ABDI._polarform, 0
		je		CheckNext
		cmp		[rdx]. ABDI._polarform, 1
		je		DoPolar
		mov		_var1, rcx
		mov		rcx, rdx
		call	BDIToPolar
		mov		_var2, rax
		jmp		DoPolar

CheckNext:

		cmp		[rdx]. ABDI._polarform, 0
		je		DoCartesian
		mov		_var2, rdx
		call	BDIToPolar
		mov		_var1, rax
		jmp		DoPolar

DoCartesian:

		mov		rax, _decimallimit
		mov		_var7, rax
		mul		_three
		mov		_decimallimit, rax
		mov		rdx, [rcx]. ABDI._real
		mov		rcx, rdx
		call	BDMultiply64
		mov		_t1, rax
		mov		rcx, _v2
		mov		rdx, [rcx]. ABDI._im
		mov		rcx, rdx
		call	BDMultiply64
		mov		_t2, rax
		mov		rdx, rax
		mov		rcx, _t1
		call	BDAdd64
		mov		_var6, rax
		mov		rcx, _v2
		call	BDIConjugate
		mov		_var3, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _var3
		mov		rdx, [rdx]. ABDI._real
		call	BDMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _var3
		mov		rdx, [rdx]. ABDI._im
		call	BDMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		cmp		[rax]. ABD._sign, 0
		cmove	rdx, _one
		cmovne	rdx, _zero
		mov		[rax]. ABD._sign, rdx
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		call	BDAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var6
		call	BDDivide64
		mov		_var1, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _var3
		mov		rdx, [rdx]. ABDI._im
		call	BDMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _var3
		mov		rdx, [rdx]. ABDI._real
		call	BDMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _t2
		call	BDAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _var6
		call	BDDivide64
		mov		_var2, rax
		mov		rdx, rax
		mov		rcx, _var1
		call	CreateBDITakeBothBD
		mov		_var1, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var6
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBDI
		mov		rcx, _var7
		mov		_decimallimit, rcx
		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._real
		call	BDFixTheEnd
		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._im
		call	BDFixTheEnd
		mov		rax, _var1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

DoPolar:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _var2
		mov		rdx, [rdx]. ABDI._real
		DDivide64
		mov		_t1, rax
		mov		rcx, _var1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _var2
		mov		rdx, [rdx]. ABDI._im
		DSub64
		mov		_t2, rax
		mov		rcx, _v1
		cmp		[rcx]. ABDI._polarform, 1
		je		V1Good
		mov		rcx, _var1
		call	DestroyBDI

V1Good:

		mov		rcx, _v2
		cmp		[rcx]. ABDI._polarform, 1
		je		V2Good
		mov		rcx, _var2
		call	DestroyBDI

V2Good:

		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBDITakeBothBD
		mov		[rax]. ABDI._polarform, 1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDIDivide64 ENDP
	;;
	;@brief Exponentiation of a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The base  ABDI object.
	;@param rdx - The exponent  ABDI object.
	;;
	BDIPow64 PROC _base: ABDI, _exp: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rax, _decimallimit
		mov		_var1, rax
		add		rax, 10
		mov		_decimallimit, rax
		call	CreateBDIFromBDI
		mov		_retval, rax
		mov		rcx, _v2
		mov		r12, [rcx]. ABDI._real
		mov		r12, [r12]. ABD._num
		cmp		qword ptr[r12], 0
		je		ReturnOne
		cmp		qword ptr[r12], 1
		ja		NextCheck
		cmp		qword ptr[r12+8], 0
		je		ReturnOne
		cmp		qword ptr[r12+8], 1
		je		GoOut

NextCheck:

		mov		rcx, _v2
		mov		rax, [rcx]. ABDI._im
		mov		rax, [rax]. ABD._num
		cmp		qword ptr[rax], 1
		ja		DoOtherWay
		cmp		qword ptr[rax+8], 0
		ja		DoOtherWay
		mov		r14, qword ptr[r12]
		bsr		r15, qword ptr[r12+r14*8]
		lea		r9, _bits
		cmp		r15, 0
		je		NextByte
		dec		r15
		mov		r8, _retval
		mov		r12, _v2
		mov		r12, [r12]. ABDI._real
		mov		r12, [r12]. ABD._num
		mov		r11, _v1

MainLoop:

		mov		rcx, _retval
		call	BDISquare64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBDI
		lea		r9, _bits
		mov		rax, qword ptr[r9+r15*8]
		test	qword ptr[r12+r14*8], rax
		jz		PassMult
		mov		rcx, _v1
		mov		rdx, _retval
		call	BDIMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBDI

PassMult:

		cmp		r15, 0
		je		NextByte
		dec		r15
		jmp		MainLoop

NextByte:

		cmp		r14, 1
		jna		GoOut
		mov		r15, 63
		dec		r14
		jmp		MainLoop

ReturnZero:

		mov		r8, _retval
		mov		rax, [r8]. ABDI._real
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rax, [r8]. ABDI._real
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._sign, 0
		mov		rax, [r8]. ABDI._im
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rax, [r8]. ABDI._im
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		jmp		GoOut


ReturnOne:

		mov		r8, _retval
		mov		rax, [r8]. ABDI._real
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r8]. ABDI._real
		mov		[rax]. ABD._decimalplace, 0
		mov		[rax]. ABD._sign, 0
		mov		rax, [r8]. ABDI._im
		mov		rax, [rax]. ABD._num
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		rax, [r8]. ABDI._im
		mov		[rax]. ABD._sign, 0
		mov		[rax]. ABD._decimalplace, 0
		jmp		GoOut

GoOut:

		mov		rax, _var1
		mov		_decimallimit, rax
		mov		rcx, _retval
		mov		rcx, [rcx]. ABDI._real
		call	BDFixTheEnd
		mov		rcx, _retval
		mov		rcx, [rcx]. ABDI._im
		call	BDFixTheEnd
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		RET

DoOtherWay:

		jmp		ReturnZero
	
	BDIPow64	ENDP
	;;
	;@brief To string function for  ABDI objects.
	;@returns a char array.
	;@param rcx - The  ABDI object to convert to string.
	;@param rdx - The radix to use for the calculations (qword).
	;;
	BDIToString64 PROC _val: ABDI, _rad:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, [rcx]. ABDI._real
		mov		rdx, _v2
		call	BDToString64
		mov		_retval, rax
		mov		_var1, rcx
		mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im
		mov		rdx, _v2
		call	BDToString64
		mov		_var2, rax
		mov		_var3, rcx
		mov		rcx, _var1
		add		rcx, _var3
		add		rcx, 10
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_var4, rax
		mov		rdi, rax
		mov		rax, 0
		mov		rcx, _var1
		add		rcx, _var3
		add		rcx, 10
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _var4
		mov		byte ptr[rdi], '('
		inc		rdi
		mov		rsi, _retval
		mov		rcx, _var1
		pushf
		cld
rep		movsb
		popf
		mov		rax, _v1
		mov		rax, [rax]. ABDI._real
		cmp		[rax]. ABD._sign, 1
		je		NoPlus
		mov		byte ptr[rdi], '+'
		inc		rdi

NoPlus:

		mov		rsi, _var2
		mov		rcx, _var3
		pushf
		cld
rep		movsb
		popf
		mov		byte ptr[rdi], ')'
		mov		rcx, _retval
		call	Free64
		mov		rcx, _var2
		call	Free64
		mov		rax, _var4
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDIToString64 ENDP
































	;;
	;@brief Creates a  ABFI object equal to zero.
	;@returns a  ABFI object.
	;;
	CreateBFI PROC

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFI endp
	;;
	;@brief Creates a  ABFI object by copying another  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to copy.
	;;
	CreateBFIFromBFI PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real
		call	CreateBFFromBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		call	CreateBFFromBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFIFromBFI endp
	;;
	;@brief Creates a  ABFI object by copying a  ABF object.
	;@returns a  ABFI object.
	;@param rcx - The  ABF object to copy to the real part.
	;;
	CreateBFICopyBF PROC _r: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFFromBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFICopyBF endp
	;;
	;@brief Creates a  ABFI object by copying an array.
	;@returns a  ABFI object.
	;@param rcx - The array to copy to the real part.
	;;
	CreateBFICopyArray PROC _r:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFCopy
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFICopyArray endp
	;;
	;@brief Creates a  ABFI object by copying 2  ABF objects.
	;@returns a  ABFI object.
	;@param rcx - The  ABF object to copy to the real part.
	;@param rdx - The  ABF object to copy to the imaginary part.
	;;
	CreateBFICopyBothBF PROC _r: ABF, _i: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFFromBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		mov		rcx, _v2
		call	CreateBFFromBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFICopyBothBF endp
	;;
	;@brief Creates a  ABFI object by copying 2 arrays.
	;@returns a  ABFI object.
	;@param rcx - The array to copy to the real part.
	;@param rdx - The array to copy to the imaginary part.
	;;
	CreateBFICopyBothArray PROC _r:PTR qword, _i:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFCopy
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		mov		rcx, _v2
		call	CreateBFCopy
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFICopyBothArray endp
	;;
	;@brief Creates a  ABFI object by taking possession of a  ABF object.
	;@returns a  ABFI object.
	;@param rcx - The  ABF object to take possession of for the real part.
	;;
	CreateBFITakeBF PROC _r: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		mov		[rax]. ABFI._real, rcx
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFITakeBF endp
	;;
	;@brief Creates a  ABFI object by taking possession of an array.
	;@returns a  ABFI object.
	;@param rcx - The array to take possession of for the real part.
	;;
	CreateBFITakeArray PROC _r:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFTake
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		call	CreateBF
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFITakeArray endp
	;;
	;@brief Creates a  ABFI object by taking possession of 2  ABF objects.
	;@returns a  ABFI object.
	;@param rcx - The  ABF object to take possession of for the real part.
	;@param rdx - The  ABF object to take possession of for the imaginary part.
	;;
	CreateBFITakeBothBF PROC _r: ABF, _i: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		mov		[rax]. ABFI._real, rcx
		mov		rcx, _v2
		mov		[rax]. ABFI._im, rcx
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFITakeBothBF endp
	;;
	;@brief Creates a  ABFI object by taking possession of 2 arrays.
	;@returns a  ABFI object.
	;@param rcx - The array to take possession of for the real part.
	;@param rdx - The array to take possession of for the imaginary part.
	;;
	CreateBFITakeBothArray PROC _r:PTR qword, _i:PTR qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, 32
		call	malloc
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0
		mov		rcx, _v1
		call	CreateBFTake
		mov		rcx, _retval
		mov		[rcx]. ABFI._real, rax
		mov		rcx, _v2
		call	CreateBFTake
		mov		rcx, _retval
		mov		[rcx]. ABFI._im, rax
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	CreateBFITakeBothArray endp
	;;
	;@brief Destroys a  ABFI object.
	;@param rcx - The  ABFI object to destroy.
	;;
	DestroyBFI PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		mov		_v1, rcx
		mov		rcx, [rcx]. ABFI._real
		call	DestroyBF
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		call	DestroyBF
		mov		rcx, _v1
		call	free

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	DestroyBFI endp
	;;
	;@brief Returns the conjuagte of a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to conjugate.
	;;
	BFIConjugate64 PROC _val:qword

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		call	CreateBFIFromBFI
		mov		_retval, rax
		cmp		[rax]. ABFI._im. ABF._sign, 0
		cmove	rdx, _one
		cmovne	rdx, _zero
		mov		[rax]. ABFI._im. ABF._sign, rdx

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFIConjugate64 ENDP
	;;
	;@brief Converts a  ABFI object to polar form.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to convert.
	;;
	BFIToPolar PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABFI._polarform, 1
		jne		DoCalc
		call	CreateBFIFromBFI
		mov		_retval, rax
		jmp		ExitNow

DoCalc:

		mov		_v1, rcx
		mov		rdx, [rcx]. ABFI._real
		mov		rcx, rdx
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, _v1
		mov		rdx, [rcx]. ABFI._im
		mov		rcx, rdx
		call	BFMultiply64
		mov		_t2, rax
		cmp		[rax]. ABF._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABF._sign, rcx
		mov		rdx, rax
		mov		rcx, _t1
		call	BFAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BFSqrt64
		mov		_retval, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real
		mov		rdx, _bfzero
		call	BFE64
		cmp		rax, 1
		je		Equal
		mov		rcx, _v1
		mov		rdx, [rcx]. ABFI._real
		mov		rcx, [rcx]. ABFI._im
		call	BFDivide64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t1
		mov		rdx, _defaultcyclecount
		call	BFRadATan64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real
		mov		rdx, _bfzero
		call	BFGT64
		cmp		rax, 1
		je		GoOut
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _bfzero
		call	BFLT64
		cmp		rax, 1
		je		DoSub
		mov		rcx, _t1
		mov		rdx, _bfpi
		call	BFAdd64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		jmp		GoOut

DoSub:

		mov		rcx, _t1
		mov		rdx, _bfpi
		call	BFSub64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		jmp		GoOut

Equal:

		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _bfzero
		call	BFE64
		cmp		rax, 1
		je		ErrorOut
		mov		rcx, _bfpi
		call	CreateBFITakeBothBF
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _bfzero
		call	BFLT64
		cmp		rax, 0
		je		GoOut
		mov		rax, _t1
		mov		[rax]. ABF._sign, 1

GoOut:

		mov		rcx, _retval
		mov		rdx, _t1
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 1
		mov		rcx, _t2
		call	DestroyBD

ExitNow:

		mov		rax, _retval

ErrorOut:

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFIToPolar ENDP
	;;
	;@brief Converts a  ABFI object to rectangular form.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object convert.
	;;
	BFIFromPolar PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		cmp		[rcx]. ABFI._polarform, 0
		jne		DoCalc
		call	CreateBFIFromBFI
		mov		_retval, rax
		jmp		ExitNow

DoCalc:

		mov		_v1, rcx
		mov		rcx, [rcx]. ABFI._im
		call	BFRadCos64
		mov		_t1, rax
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABDI._real
		FMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBFI
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		call	BFRadSin64
		mov		_t2, rax
		mov		rcx, rax
		mov		rdx, _v1
		mov		rdx, [rdx]. ABFI._real
		FMultiply64
		mov		rcx, _t2
		mov		_t2, rax
		call	DestroyBFI
		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		[rax]. ABFI._polarform, 0

ExitNow:

		mov		rax, _retval

ErrorOut:

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFIFromPolar ENDP
	;;
	;@brief Checks for equality between 2  ABFI objects.
	;@returns a bool.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFIE64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		cmp		[rcx]. ABFI._polarform, 0
		je		CheckNext
		cmp		[rdx]. ABFI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BFIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABFI._polarform, 0
		je		DoIt
		call	BFIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._real
		call	BFE64
		cmp		rax, 1
		jne		NotEq
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._im
		call	BFE64
		cmp		rax, 1
		jne		NotEq
		mov		_retval, 1
		jmp		GoOut

NotEq:

		mov		_retval, 0

GoOut:

		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBFI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBFI

V2Good:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFIE64 ENDP
	;;
	;@brief Checks for inequality between 2  ABFI objects.
	;@returns a bool.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFINE64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		call	BFIE64
		cmp		rax, 1
		cmove	rax, _zero
		cmovne	rax, _one
		
neAllreadySet:	

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET
		
	BFINE64 ENDP	
	;;
	;@brief Adds 2  ABFI objects.
	;@returns a  ABFI object.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFIAdd64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		mov		_var3, 0
		cmp		[rcx]. ABFI._polarform, 0
		je		CheckNext
		mov		_var3, 1
		cmp		[rdx]. ABFI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BFIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABFI._polarform, 0
		je		DoIt
		mov		_var3, 1
		call	BFIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._real
		call	BFAdd64
		mov		_retval, rax
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._im
		call	BFAdd64
		mov		_t1, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBFI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBFI

V2Good:

		mov		rdx, _t1
		mov		rcx, _retval
		call	CreateBFITakeBothBF
		mov		rcx, _var3
		mov		[rax]. ABFI._polarform, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFIAdd64 ENDP 
	;;
	;@brief Subtracts 2  ABFI objects.
	;@returns a  ABFI object.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFISub64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		_var1, rcx
		mov		_var2, rdx
		mov		_var3, 0
		cmp		[rcx]. ABFI._polarform, 0
		je		CheckNext
		mov		_var3, 1
		cmp		[rdx]. ABFI._polarform, 1
		je		DoIt
		mov		rcx, rdx
		call	BFIToPolar
		mov		_var2, rax
		jmp		DoIt

CheckNext:

		cmp		[rdx]. ABFI._polarform, 0
		je		DoIt
		mov		_var3, 1
		call	BFIToPolar
		mov		_var1, rax

DoIt:

		mov		r11, _var1
		mov		r12, _var2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._real
		call	BFSub64
		mov		_retval, rax
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._im
		call	BFSub64
		mov		_t1, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBFI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBFI

V2Good:

		mov		rdx, _t1
		mov		rcx, _retval
		call	CreateBFITakeBothBF
		mov		rcx, _var3
		mov		[rax]. ABFI._polarform, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFISub64 ENDP
	;;
	;@brief Multiplies 2  ABFI objects.
	;@returns a  ABFI object.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFIMultiply64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		[rcx]. ABFI._polarform, 0
		je		CheckNext
		mov		_var1, rcx
		mov		_var2, rdx
		cmp		[rdx]. ABFI._polarform, 1
		je		DoPolar
		mov		rcx, rdx
		call	BFIToPolar
		mov		_var2, rax
		jmp		DoPolar

CheckNext:

		cmp		[rdx]. ABFI._polarform, 0
		je		DoCartesian
		mov		_var2, rdx
		call	BFIToPolar
		mov		_var1, rax
		jmp		DoPolar

DoCartesian:

		mov		r11, _v1
		mov		r12, _v2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._real
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._im
		call	BFMultiply64
		mov		_t2, rax
		cmp		[rax]. ABF._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABF._sign, rcx
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_retval, rax
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._im
		call	BFMultiply64
		mov		rcx, _t1
		mov		_t1, rax
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r11, _v1
		mov		r12, _v2
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._real
		call	BFMultiply64
		mov		_t2, rax
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_var1, rax
		mov		rcx, _retval
		mov		rdx, rax
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoPolar:

		mov		rcx, _var1
		mov		rcx, [rcx]. ABFI._real
		mov		rdx, _var2
		mov		rdx, [rdx]. ABFI._real
		FMultiply64
		mov		_t1, rax
		mov		rcx, _var1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _var2
		mov		rdx, [rdx]. ABFI._im
		FAdd64
		mov		_t2, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBFI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBFI

V2Good:

		mov		rcx, _t1
		mov		rdx, _t2
		call	CreateBFITakeBothBF
		mov		[rax]. ABFI._polarform, 1

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

	BFIMultiply64 ENDP
	;;
	;@brief Squares a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABFI object to square.
	;;
	BFISquare64 PROC _val1: ABFI

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		cmp		[rcx]. ABFI._polarform, 1
		je		DoPolar
		mov		r12, _v1
		mov		rcx, [r12]. ABFI._real
		mov		rdx, rcx
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r12]. ABFI._im
		mov		rdx, rcx
		call	BFMultiply64
		mov		_t2, rax
		cmp		[rax]. ABF._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABF._sign, rcx
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_retval, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r12, _v1
		mov		rcx, [r12]. ABFI._real
		mov		rdx, [r12]. ABFI._im
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r12]. ABFI._im
		mov		rdx, [r12]. ABFI._real
		call	BFMultiply64
		mov		_t2, rax
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_var1, rax
		mov		rcx, _retval
		mov		rdx, rax
		call	CreateBFITakeBothBF
		mov		_retval, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		RET

DoPolar:

		mov		rcx, [rcx]. ABFI._real
		mov		rdx, rcx
		FMultiply64
		mov		_t1, rax
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, rcx
		FAdd64
		mov		rdx, rax
		mov		rcx, _t1
		call	CreateBFITakeBothBF
		mov		[rax]. ABFI._polarform, 1

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero

		RET

	BFISquare64 ENDP
	;;
	;@brief Divides 2  ABFI objects.
	;@returns a  ABFI object.
	;@param rcx - The first  ABFI object.
	;@param rdx - The second  ABFI object.
	;;
	BFIDivide64 PROC _val1: ABFI, _val2: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		cmp		[rcx]. ABFI._polarform, 0
		je		CheckNext
		mov		_var1, rcx
		mov		_var2, rdx
		cmp		[rdx]. ABFI._polarform, 1
		je		DoPolar
		mov		rcx, rdx
		call	BFIToPolar
		mov		_var2, rax
		jmp		DoPolar

CheckNext:

		cmp		[rdx]. ABFI._polarform, 0
		je		DoCartesian
		mov		_var2, rdx
		call	BFIToPolar
		mov		_var1, rax
		jmp		DoPolar

DoCartesian:

		mov		r12, _v2
		mov		rcx, r12
		call	BFIConjugate64
		mov		_var2, rax
		mov		rcx, [r12]. ABFI._real
		mov		rdx, rcx
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r12]. ABFI._im
		mov		rdx, rcx
		call	BFMultiply64
		mov		_t2, rax
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_var3, rax
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r11, _v1
		mov		r12, _var2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._real
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._im
		call	BFMultiply64
		mov		_t2, rax
		cmp		[rax]. ABF._sign, 0
		cmove	rcx, _one
		cmovne	rcx, _zero
		mov		[rax]. ABF._sign, rcx
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_retval, rax
		mov		rcx, rax
		mov		rdx, _var3
		call	BFDivide64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		r11, _v1
		mov		r12, _var2
		mov		rcx, [r11]. ABFI._real
		mov		rdx, [r12]. ABFI._im
		call	BFMultiply64
		mov		_t1, rax
		mov		rcx, [r11]. ABFI._im
		mov		rdx, [r12]. ABFI._real
		call	BFMultiply64
		mov		_t2, rax
		mov		rcx, _t1
		mov		rdx, rax
		call	BFAdd64
		mov		_var1, rax
		mov		rcx, rax
		mov		rdx, _var3
		call	BFDivide64
		mov		rcx, _var1
		mov		_var1, rax
		call	DestroyBD
		mov		rcx, _t1
		call	DestroyBD
		mov		rcx, _t2
		call	DestroyBD
		mov		rcx, _var3
		call	DestroyBD
		mov		rcx, _var2
		call	DestroyBFI
		mov		rcx, _retval
		mov		rdx, _var1
		call	CreateBFITakeBothBF

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _zero, _zero, _zero

		RET

DoPolar:
		
		mov		rcx, _var1
		mov		rcx, [rcx]. ABFI._real
		mov		rdx, _var2
		mov		rdx, [rdx]. ABFI._real
		FDivide64
		mov		_t1, rax
		mov		rcx, _var1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _var2
		mov		rdx, [rdx]. ABFI._im
		FSub64
		mov		_t2, rax
		mov		rcx, _var1
		cmp		rcx, _v1
		je		V1Good
		call	DestroyBFI

V1Good:

		mov		rcx, _var2
		cmp		rcx, _v2
		je		V2Good
		call	DestroyBFI

V2Good:

		mov		rdx, _t2
		mov		rcx, _t1
		call	CreateBFITakeBothBF
		mov		[rax]. ABFI._polarform, 1

	BFIDivide64 ENDP
	;;
	;@brief Exponentiation of a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The Base  ABFI object.
	;@param rdx - The Exponent  ABFI object.
	;;
	BFIPow64 PROC _base: ABFI, _exp: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		call	CreateBFIFromBFI
		mov		_retval, rax
		mov		rcx, _v2
		mov		rax, [rcx]. ABFI._im
		mov		rax, [rax]. ABF._numerator
		cmp		qword ptr[rax], 1
		ja		DoOtherWay
		cmp		qword ptr[rax+8], 0
		ja		DoOtherWay
		mov		rax, [rcx]. ABFI._im
		mov		rax, [rax]. ABF._denominator
		cmp		qword ptr[rax], 1
		ja		DoOtherWay
		cmp		qword ptr[rax+8], 0
		ja		DoOtherWay
		mov		rax, [rcx]. ABFI._real
		mov		rax, [rax]. ABF._denominator
		cmp		qword ptr[rax],0
		je		ReturnOne
		cmp		qword ptr[rax], 1
		ja		DoDec
		cmp		qword ptr[rax+8], 1
		ja		DoDec
		mov		rax, [rcx]. ABFI._real
		mov		rax, [rax]. ABF._numerator
		cmp		qword ptr[rax], 0
		je		ReturnOne
		cmp		qword ptr[rax], 1
		ja		NextCheck
		cmp		qword ptr[rax+8], 0
		je		ReturnOne
		cmp		qword ptr[rax+8], 1
		je		GoOut

NextCheck:

		mov		r8, _retval
		mov		r12, _v2
		mov		r12, [r12]. ABFI._real
		mov		r12, [r12]. ABF._numerator
		mov		r11, _v1
		mov		r14, qword ptr[rax]
		bsr		r15, qword ptr[rax+r14*8]
		lea		r9, _bits
		cmp		r15, 0
		je		NextByte
		dec		r15

MainLoop:

		mov		rcx, _retval
		call	BFISquare64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD
		lea		r9, _bits
		mov		rax, qword ptr[r9+r15*8]
		test	qword ptr[r12+r14*8], rax
		jz		PassMult
		mov		rcx, _v1
		mov		rdx, _retval
		call	BFIMultiply64
		mov		rcx, _retval
		mov		_retval, rax
		call	DestroyBD

PassMult:

		cmp		r15, 0
		je		NextByte
		dec		r15
		jmp		MainLoop

NextByte:

		cmp		r14, 1
		jna		GoOut
		mov		r15, 63
		dec		r14
		jmp		MainLoop

ReturnZero:

		mov		r8, _retval
		mov		rax, [r8]. ABFI._real
		mov		rax, [rax]. ABF._denominator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r8]. ABFI._real
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		[r8]. ABFI._real. ABF._sign, 0
		mov		rax, [r8]. ABFI._im
		mov		rax, [rax]. ABF._denominator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r8]. ABFI._im
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		[r8]. ABFI._im. ABF._sign, 0
		jmp		GoOut

ReturnOne:

		mov		r8, _retval
		mov		rax, [r8]. ABFI._real
		mov		rax, [rax]. ABF._denominator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r8]. ABFI._real
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		[r8]. ABFI._real. ABF._sign, 0
		mov		rax, [r8]. ABFI._im
		mov		rax, [rax]. ABF._denominator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r8]. ABFI._im
		mov		rax, [rax]. ABF._numerator
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 0
		mov		[r8]. ABFI._im. ABF._sign, 0
		jmp		GoOut

GoOut:

		mov		rax, _retval

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _one, _one, _zero, _one, _one, _zero, _one, _one, _zero

		RET

DoOtherWay:

		jmp		ReturnZero

DoDec:

		jmp		ReturnZero	
	
	BFIPow64	ENDP
	;;
	;@brief To string function for  ABFI objects.
	;@returns a char array.
	;@param rcx - The  ABFI object to convert to string.
	;@param rdx - The radix to use for the calculation.
	;;
	BFIToString64 PROC _val: ABFI, _rad:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_v1, rcx
		mov		_v2, rdx
		mov		rcx, [rcx]. ABFI._real
		mov		rdx, _v2
		call	BFToString64
		mov		_retval, rax
		mov		_var1, rcx
		mov		_var5, rcx
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		mov		rdx, _v2
		call	BFToString64
		mov		_var2, rax
		mov		_var3, rcx
		add		_var5, rcx
		mov		rcx, _v1
		mov		rax, [rcx]. ABFI._real
		mov		rax, [rax]. ABF._sign
		mov		rcx, [rcx]. ABFI._im
		add		rax, [rcx]. ABF._sign
		add		_var5, rax
		mov		rcx, _var5
		add		rcx, 10
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		_var4, rax
		mov		rdi, rax
		mov		rcx, _var5
		add		rcx, 8
		mov		rax, 0
		pushf
		cld
rep		stosb
		popf
		mov		rdi, _var4
		mov		byte ptr[rdi], '('
		inc		rdi
		mov		rsi, _retval
		mov		rcx, _var1
		pushf
		cld
rep		movsb
		popf
		mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im
		cmp		[rcx]. ABF._sign, 1
		je		NoPlus
		mov		byte ptr[rdi], '+'
		inc		rdi
		
NoPlus:	

		mov		rsi, _var2
		mov		rcx, _var3
		pushf
		cld
rep		movsb
		popf
		mov		byte ptr[rdi], 'i'
		inc		rdi
		mov		byte ptr[rdi], ')'
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rcx, _retval
		call	Free64
		mov		rcx, _var2
		call	Free64
		mov		rax, _var4

		EXITFUNC _zero, _zero, _zero, _zero, _one, _one, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFIToString64 ENDP
	
	
	
















	;;
	;@brief Changes the decimal limit used by  ABD objects.
	;@param rcx - qword the value of the new decimallimit.
	;;
	BDChangeDecimalLimit PROC _newval:qword

		ENTERFUNC _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		_decimallimit, rcx
		mov		rax, rcx
		xor		rdx, rdx
		div		_ten
		add		rax, rcx
		mov		_workingdecimallimit, rax

		EXITFUNC _one, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDChangeDecimalLimit  ENDP
	;;
	;@brief Returns the current decimallimit.
	;@returns a qword.
	;;
	BDGetDecimalLimit PROC

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, _decimallimit

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDGetDecimalLimit  ENDP
	;;
	;@brief Sets how the  ABF class handles rounding.
	;@param rcx - qword representing the value of how to handle rounding.
	;;
	BDChangeHowToHandleRounding PROC  _newval:qword

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		cmp		rcx, 5
		ja		KeepThis
		mov		_howtohandlerounding, rcx

KeepThis:

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDChangeHowToHandleRounding  ENDP
	;;
	;@brief Gets the value for how to handle rounding.
	;@returns a qword.
	;;
	BDGetHowToHandleRounding PROC

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rax, _howtohandlerounding

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDGetHowToHandleRounding   ENDP
	;;
	;@brief Converts a  ABI object to a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABI object to convert.
	;;
	BItoBD PROC _val: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero
		
		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABI._num
		call	CreateBDCopy
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1	
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BItoBD ENDP
	;;
	;@brief Converts a  ABI object to a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABI object to convert.
	;;
	BItoBF PROC _val: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]
		mov		rcx, [rcx]. ABI._num
		call	CreateBFCopy
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BItoBF ENDP
	;;
	;@brief Converts a  ABI object to a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABI object to convert.
	;;
	BItoBDI PROC _val: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBDI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABDI._real. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BItoBDI ENDP
	;;
	;@brief Converts a  ABI object to a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABI object to convert.
	;;
	BItoBFI PROC _val: ABI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBFI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABI._sign
		mov		[rax]. ABFI._real. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BItoBFI ENDP
	;;
	;@brief Converts a  ABD object to a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABD object to convert.
	;;
	BDtoBI PROC _val1: ABD

		ENTERFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		rax, rbp							;make enough room for ;vals
		sub		rsp, 128
		mov		rbp, rsp
		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		qword ptr[rbp+40], 0				;mov		_t, 0
		mov		qword ptr[rbp+48], 0				;mov		_mod, 0
		mov		qword ptr[rbp+56], 0				;mov		_div, 0
		mov		qword ptr[rbp+96], 0				;mov		_val, 0
		mov		r11, qword ptr[rbp+16]				;mov		rcx, _v1
		cmp		[r11]. ABD._decimalplace, 0
		je		JustReturnNum
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rbp+40], rax				;mov		_t, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		rcx, rax
		mov		rdx, [r11]. ABD._decimalplace
		dec		rdx
		Pow64
		mov		qword ptr[rbp+56], rax				;mov		_div, rax
		mov		rdx, rax
		mov		rcx, [r11]. ABD._num
		Mod64
		mov		qword ptr[rbp+48], rax				;mov		_mod, rax
		mov		rdx, rax
		mov		rcx, [r11]. ABD._num
		Sub64
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		mov		rdx, qword ptr[rbp+56]				;mov		rdx, _div
		mov		rcx, rax
		Divide64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		call	Free64
		mov		rdx, qword ptr[rbp+56]				;mov		rdx, _div
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 10
		mov		rcx, _v1
		Mod64
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _mod
		mov		qword ptr[rbp+48], rax				;mov		_mod, rax
		call	Free64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _mod
		Sub64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		call	Free64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		rdx, qword ptr[rbp+56]				;mov		rdx, _div
		Divide64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		call	Free64
		mov		rdx, qword ptr[rbp+56]				;mov		rdx, _div
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _mod
		mov		qword ptr[rdx+8], 1
		cmp		qword ptr[rcx+8], 5
		jb		DontInc
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		Add64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		call	Free64

DontInc:

		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _div
		call	Free64
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _mod
		call	Free64
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _t
		call	Free64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		call	CreateBITake
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABD._sign
		mov		[rax]. ABI._sign, rcx
		jmp		GoOut

JustReturnNum:

		mov		rcx, [r11]. ABD._num
		call	CreateBICopy
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABD._sign
		mov		[rax]. ABI._sign, rcx
		jmp		GoOut

MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		mov		rdx, qword ptr[rbp+8]				;reset the pointers to their original positions
		mov		rsp, rbp
		add		rsp, 128
		mov		rbp, rdx

		EXITFUNC _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDtoBI ENDP
	;;
	;@brief Converts a  ABD object to a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABD object to convert.
	;;
	BDtoBF PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		qword ptr[rbp+40], 0				;mov		_num, 0
		mov		qword ptr[rbp+48], 0				;mov		_den, 0
		mov		r11, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rbp+48], rax				;mov		_den, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		mov		rax, [r11]. ABD._num
		mov		rax, qword ptr[rax]
		add		rax, 5
		mul		_eight
		mov		rcx, rax
		call	GetMem64
		cmp		rax, 0
		je		MallocError
		mov		qword ptr[rbp+40], rax				;mov		_num, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 1
		cmp		[r11]. ABD._decimalplace, 0
		je		JustCopyV1
		mov		rcx, [r11]. ABD._decimalplace
		mov		qword ptr[rax+8], rcx
		mov		rdx, rax
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _den
		mov		qword ptr[rcx+8], 10
		Pow64
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _den
		mov		qword ptr[rbp+48], rax				;mov		_den, rax
		call	Free64

JustCopyV1:

		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _num
		mov		r11, [r11]. ABD._num
		mov		rsi, r11
		mov		rdi, rcx
		mov		rcx, qword ptr[r11]
		inc		rcx
		pushf
		cld
rep		movsq
		popf
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _den
		call	CreateBFTakeBoth
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABD._sign
		mov		[rax]. ABF._sign, rcx
		jmp		GoOut
		
MallocError:

		mov		rcx, 32
		call	GetMem64
		cmp		rax, 0
		je		GoOut
		mov		r8, rax
		call	GetLastError
		mov		qword ptr[r8], 0
		mov		qword ptr[r8+8], 0
		mov		qword ptr[r8+16], rax

GoOut:

		EXITFUNC _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BDtoBF ENDP
	;;
	;@brief Converts a  ABD object to a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABD object to convert.
	;;
	BDtoBDI PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABD._num
		call	CreateBDICopyBD
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rdx, [rcx]. ABD._decimalplace
		mov		[rax]. ABDI._real. ABD._decimalplace, rdx
		mov		rcx, [rcx]. ABD._sign
		mov		[rax]. ABDI._real. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDtoBDI ENDP
	;;
	;@brief Converts a  ABD object to a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABD object to convert.
	;;
	BDtoBFI PROC _val: ABD

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBFI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABD._sign
		mov		[rax]. ABFI._real. ABF._sign, rcx
		mov		rdx, qword ptr[rbp+8]				;reset the pointers to their original positions
		mov		rsp, rbp
		add		rsp, 128
		mov		rbp, rdx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDtoBFI ENDP
	;;
	;@brief Converts a  ABF object to a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABF object to convert.
	;;
	BFtoBI PROC _val1: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		qword ptr[rbp+40], 0				;mov		_mod, 0
		mov		qword ptr[rbp+48], 0				;mov		_div, 0
		mov		r11, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r11]. ABF._denominator
		Mod64
		mov		qword ptr[rbp+56], rax				;mov		_tval, rax
		mov		rcx, [r11]. ABF._numerator
		mov		rdx, [r11]. ABF._denominator
		Divide64
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		mov		rcx, 32
		call	GetMem64
		mov		qword ptr[rbp+48], rax				;mov		_div, rax
		mov		qword ptr[rax], 1
		mov		qword ptr[rax+8], 10
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _mod
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _div
		Mod64
		mov		qword ptr[rbp+40], rax				;mov		_mod, rax
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		rdx, rax
		Sub64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		qword ptr[rbp+56], rax				;mov		_tval, rax
		call	Free64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _div
		Divide64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		qword ptr[rbp+56], rax				;mov		_tval, rax
		call	Free64

RoundingLoop:

		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		cmp		qword ptr[rcx], 1
		ja		CarryOn
		cmp		qword ptr[rcx+8], 0
		ja		CarryOn
		mov		rdx, qword ptr[rbp+40]				;mov		rdx, _mod
		cmp		qword ptr[rdx+8], 5
		jb		GoOut
		mov		qword ptr[rdx], 1
		mov		qword ptr[rdx+8], 1
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		Add64
		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		mov		qword ptr[rbp+96], rax				;mov		_val, rax
		call	Free64
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _mod
		call	Free64
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _div
		call	Free64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		call	Free64
		jmp		GoOut

CarryOn:

		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _div
		Mod64
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _mod
		mov		qword ptr[rbp+40], rax				;mov		_mod, rax
		call	Free64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		rdx, qword ptr[rbp+40]				;mov		rdx, _mod
		Sub64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		qword ptr[rbp+56], rax				;mov		_tval, rax
		call	Free64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		rdx, qword ptr[rbp+48]				;mov		rdx, _div
		Divide64
		mov		rcx, qword ptr[rbp+56]				;mov		rcx, _tval
		mov		qword ptr[rbp+56], rax				;mov		_tval, rax
		call	Free64
		jmp		RoundingLoop

GoOut:

		mov		rcx, qword ptr[rbp+96]				;mov		rcx, _val
		call	CreateBITake
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABF._sign
		mov		[rax]. ABI._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BFtoBI ENDP
	;;
	;@brief Converts a  ABF object to a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABF object to convert.
	;;
	BFtoBD PROC _val: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		r11, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [r11]. ABF._numerator
		call	CreateBDTake
		mov		qword ptr[rbp+40], rax				;mov		_num, rax
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABF._denominator
		call	CreateBDTake
		mov		qword ptr[rbp+48], rax				;mov		_den, rax
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _num
		mov		rdx, rax
		call	BDDivide64
		mov		qword ptr[rbp+56], rax				;mov		_r, rax
		mov		rcx, qword ptr[rbp+40]				;mov		rcx, _num
		call	DestroyBD
		mov		rcx, qword ptr[rbp+48]				;mov		rcx, _den
		call	DestroyBD
		mov		rax, qword ptr[rbp+56]				;mov		rax, _r
		mov		r11, qword ptr[rbp+16]				;mov		r11, _v1
		mov		rcx, [r11]. ABF._sign
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _one, _zero, _zero, _zero, _zero, _zero

		RET

	BFtoBD ENDP
	;;
	;@brief Converts a  ABF object to a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABF object to convert.
	;;
	BFtoBDI PROC _val: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBDI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABF._sign
		mov		[rax]. ABDI._real. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFtoBDI ENDP
	;;
	;@brief Converts a  ABF object to a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABF object to convert.
	;;
	BFtoBFI PROC _val: ABF

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBFI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABF._sign
		mov		[rax]. ABFI._real. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFtoBFI ENDP
	;;
	;@brief Converts a  ABDI object to a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDItoBI PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real. ABD._sign
		mov		[rax]. ABI._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDItoBI ENDP
	;;
	;@brief Converts a  ABDI object to a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDItoBD PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+8], rax
		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBD
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real. ABD._sign
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDItoBD ENDP
	;;
	;@brief Converts a  ABDI object to a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDItoBF PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBF
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real. ABD._sign
		mov		[rax]. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDItoBF ENDP
	;;
	;@brief Converts a  ABDI object to a  ABFI object.
	;@returns a  ABFI object.
	;@param rcx - The  ABDI object to convert.
	;;
	BDItoBFI PROC _val: ABDI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBFI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._real. ABD._sign
		mov		[rax]. ABFI._real. ABF._sign, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABDI._im. ABD._sign
		mov		[rax]. ABFI._im. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BDItoBFI ENDP
	;;
	;@brief Converts a  ABFI object to a  ABI object.
	;@returns a  ABI object.
	;@param rcx - The  ABFI object to convert.
	;;
	BFItoBI PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real. ABF._sign
		mov		[rax]. ABI._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFItoBI ENDP
	;;
	;@brief Converts a  ABFI object to a  ABD object.
	;@returns a  ABD object.
	;@param rcx - The  ABFI object to convert.
	;;
	BFItoBD PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBD
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real. ABF._sign
		mov		[rax]. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFItoBD ENDP
	;;
	;@brief Converts a  ABFI object to a  ABF object.
	;@returns a  ABF object.
	;@param rcx - The  ABFI object to convert.
	;;
	BFItoBF PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBF
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real. ABF._sign
		mov		[rax]. ABF._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFItoBF ENDP
	;;
	;@brief Converts a  ABFI object to a  ABDI object.
	;@returns a  ABDI object.
	;@param rcx - The  ABFI object to convert.
	;;
	BFItoBDI PROC _val: ABFI

		ENTERFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		mov		qword ptr[rbp+16], rcx				;mov		_v1, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		call	CreateBDI
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._real. ABF._sign
		mov		[rax]. ABDI._real. ABD._sign, rcx
		mov		rcx, qword ptr[rbp+16]				;mov		rcx, _v1
		mov		rcx, [rcx]. ABFI._im. ABF._sign
		mov		[rax]. ABDI._im. ABD._sign, rcx

		EXITFUNC _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero, _zero

		RET

	BFItoBDI ENDP



	
	; Custom Build Step, including a listing file placed in intermediate directory
; but without Source Browser information
; debug:
; ml -c -Zi "-Fl$(IntDir)\$(InputName).lst" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; release:
; ml -c "-Fl$(IntDir)\$(InputName).lst" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; outputs:
; $(IntDir)\$(InputName).obj

; Custom Build Step, including a listing file placed in intermediate directory
; and Source Browser information also placed in intermediate directory
; debug:
; ml -c -Zi "-Fl$(IntDir)\$(InputName).lst" "-FR$(IntDir)\$(InputName).sbr" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; release:
; ml -c "-Fl$(IntDir)\$(InputName).lst" "-FR$(IntDir)\$(InputName).sbr" "-Fo$(IntDir)\$(InputName).obj" "$(InputPath)"
; outputs:
; $(IntDir)\$(InputName).obj
; $(IntDir)\$(InputName).sbr
;define PORTABLE64BIT 1

	
	
	
	
	
		
	
	
	
	
	
	
	
	
	
		
	end

